@@ -10,8 +10,17 @@ enum rpmdbFlags {
     RPMDB_FLAG_JUSTCHECK	= (1 << 0),
     RPMDB_FLAG_REBUILD		= (1 << 1),
     RPMDB_FLAG_VERIFYONLY	= (1 << 2),
+    RPMDB_FLAG_SALVAGE		= (1 << 3),
 };
 
+typedef enum dbCtrlOp_e {
+    DB_CTRL_LOCK_RO		= 1,
+    DB_CTRL_UNLOCK_RO		= 2,
+    DB_CTRL_LOCK_RW		= 3,
+    DB_CTRL_UNLOCK_RW		= 4,
+    DB_CTRL_INDEXSYNC		= 5
+} dbCtrlOp;
+
 typedef struct dbiIndex_s * dbiIndex;
 typedef struct dbiCursor_s * dbiCursor;
 
@@ -20,8 +29,17 @@ struct dbConfig_s {
     int	db_cachesize;	/*!< (128Kb) */
     int	db_verbose;
     int	db_no_fsync;	/*!< no-op fsync for db */
+    int db_eflags;	/*!< obsolete */
 };
 
+struct dbiConfig_s {
+    int	dbi_oflags;		/*!< open flags */
+    int	dbi_no_dbsync;		/*!< don't call dbiSync */
+    int	dbi_lockdbfd;		/*!< do fcntl lock on db fd */
+};
+
+struct rpmdbOps_s;
+
 /** \ingroup rpmdb
  * Describes the collection of index databases used by rpm.
  */
@@ -32,7 +50,7 @@ struct rpmdb_s {
     int		db_flags;
     int		db_mode;	/*!< open mode */
     int		db_perms;	/*!< open permissions */
-    char	* db_descr;	/*!< db backend description (for error msgs) */
+    const char	* db_descr;	/*!< db backend description (for error msgs) */
     struct dbChk_s * db_checked;/*!< headerCheck()'ed package instances */
     rpmdb	db_next;
     int		db_opens;
@@ -42,8 +60,11 @@ struct rpmdb_s {
     dbiIndex 	* db_indexes;	/*!< Tag indices. */
     int		db_buildindex;	/*!< Index rebuild indicator */
 
+    const struct rpmdbOps_s * db_ops;	/*!< backend ops */
+
     /* dbenv and related parameters */
     void * db_dbenv;		/*!< Backend private handle */
+    void * db_cache;		/*!< Backend private cache handle */
     struct dbConfig_s cfg;
     int db_remove_env;
 
@@ -71,31 +92,56 @@ enum dbcFlags_e {
     DBC_WRITE	= (1 << 0),
 };
 
+enum dbcSearchType_e {
+    DBC_NORMAL_SEARCH   = 0,
+    DBC_PREFIX_SEARCH   = (1 << 0),
+};
+
 /** \ingroup dbi
  * Describes an index database (implemented on Berkeley db functionality).
  */
 struct dbiIndex_s {
-    const char * dbi_file;	/*!< file component of path */
-
-    int	dbi_oflags;		/*!< db->open flags */
-    int	dbi_no_dbsync;		/*!< don't call dbiSync */
-    int	dbi_lockdbfd;		/*!< do fcntl lock on db fd */
-    int	dbi_byteswapped;
-
     rpmdb dbi_rpmdb;		/*!< the parent rpm database */
     dbiIndexType dbi_type;	/*! Type of dbi (primary / index) */
+    const char * dbi_file;	/*!< file component of path */
+    int dbi_flags;
+    int	dbi_byteswapped;
+
+    struct dbiConfig_s cfg;
 
     void * dbi_db;		/*!< Backend private handle */
 };
 
+union _dbswap {
+    unsigned int ui;
+    unsigned char uc[4];
+};
+
+#define	_DBSWAP(_a) \
+\
+  { unsigned char _b, *_c = (_a).uc; \
+    _b = _c[3]; _c[3] = _c[0]; _c[0] = _b; \
+    _b = _c[2]; _c[2] = _c[1]; _c[1] = _b; \
+\
+  }
+
+typedef rpmRC (*idxfunc)(dbiIndex dbi, dbiCursor dbc,
+			const char *keyp, size_t keylen, dbiIndexItem rec);
+
 #ifdef __cplusplus
 extern "C" {
 #endif
 
+RPM_GNUC_INTERNAL
+rpmRC tag2index(dbiIndex dbi, rpmTagVal rpmtag, unsigned int hdrNum, Header h,
+		idxfunc idxupdate);
 
 RPM_GNUC_INTERNAL
 /* Globally enable/disable fsync in the backend */
-void dbSetFSync(void *dbenv, int enable);
+void dbSetFSync(rpmdb rdb, int enable);
+
+RPM_GNUC_INTERNAL
+int dbCtrl(rpmdb rdb, dbCtrlOp ctrl);
 
 /** \ingroup dbi
  * Return new configured index database handle instance.
@@ -114,15 +160,6 @@ dbiIndex dbiNew(rpmdb rdb, rpmDbiTagVal rpmtag);
 RPM_GNUC_INTERNAL
 dbiIndex dbiFree( dbiIndex dbi);
 
-/** \ingroup dbi
- * Format dbi open flags for debugging print.
- * @param dbflags		db open flags
- * @param print_dbenv_flags	format db env flags instead?
- * @return			formatted flags (malloced)
- */
-RPM_GNUC_INTERNAL
-char * prDbiOpenFlags(int dbflags, int print_dbenv_flags);
-
 /** \ingroup dbi
  * Actually open the database of the index.
  * @param db		rpm database
@@ -152,14 +189,6 @@ int dbiClose(dbiIndex dbi, unsigned int flags);
 RPM_GNUC_INTERNAL
 int dbiVerify(dbiIndex dbi, unsigned int flags);
 
-/** \ingroup dbi
- * Type of dbi (primary data / index)
- * @param dbi		index database handle
- * @return		type of dbi
- */
-RPM_GNUC_INTERNAL
-dbiIndexType dbiType(dbiIndex dbi);
-
 /** \ingroup dbi
  * Retrieve index control flags (new/existing, read-only etc)
  * @param dbi		index database handle
@@ -191,31 +220,78 @@ dbiCursor dbiCursorInit(dbiIndex dbi, unsigned int flags);
  * @return		NULL always
  */
 RPM_GNUC_INTERNAL
-dbiCursor dbiCursorFree(dbiCursor dbc);
+dbiCursor dbiCursorFree(dbiIndex dbi, dbiCursor dbc);
+
 
 RPM_GNUC_INTERNAL
-const void * dbiCursorKey(dbiCursor dbc, unsigned int *keylen);
+rpmRC pkgdbPut(dbiIndex dbi, dbiCursor dbc,  unsigned int *hdrNum,
+               unsigned char *hdrBlob, unsigned int hdrLen);
+RPM_GNUC_INTERNAL
+rpmRC pkgdbDel(dbiIndex dbi, dbiCursor dbc,  unsigned int hdrNum);
+RPM_GNUC_INTERNAL
+rpmRC pkgdbGet(dbiIndex dbi, dbiCursor dbc, unsigned int hdrNum,
+               unsigned char **hdrBlob, unsigned int *hdrLen);
+RPM_GNUC_INTERNAL
+unsigned int pkgdbKey(dbiIndex dbi, dbiCursor dbc);
 
 RPM_GNUC_INTERNAL
-int pkgdbPut(dbiIndex dbi, dbiCursor dbc,  unsigned int hdrNum,
-             unsigned char *hdrBlob, unsigned int hdrLen,
-             unsigned int *hdrOffset);
+rpmRC idxdbGet(dbiIndex dbi, dbiCursor dbc, const char *keyp, size_t keylen,
+               dbiIndexSet *set, int curFlags);
 RPM_GNUC_INTERNAL
-int pkgdbDel(dbiIndex dbi, dbiCursor dbc,  unsigned int hdrNum);
-RPM_GNUC_INTERNAL
-int pkgdbGet(dbiIndex dbi, dbiCursor dbc, unsigned int hdrNum,
-             unsigned char **hdrBlob, unsigned int *hdrLen,
-             unsigned int *hdrOffset);
+rpmRC idxdbPut(dbiIndex dbi, rpmTagVal rpmtag, unsigned int hdrNum, Header h);
 
 RPM_GNUC_INTERNAL
-rpmRC dbcCursorGet(dbiCursor dbc, const char *keyp, size_t keylen,
-                   dbiIndexSet *set);
+rpmRC idxdbDel(dbiIndex dbi, rpmTagVal rpmtag, unsigned int hdrNum, Header h);
+
 RPM_GNUC_INTERNAL
-rpmRC dbcCursorPut(dbiCursor dbc, const char *keyp, size_t keylen,
-		   dbiIndexItem rec);
+const void * idxdbKey(dbiIndex dbi, dbiCursor dbc, unsigned int *keylen);
+
+struct rpmdbOps_s {
+    const char *name; /* backend name */
+    const char *path; /* main database name */
+
+    int (*open)(rpmdb rdb, rpmDbiTagVal rpmtag, dbiIndex * dbip, int flags);
+    int (*close)(dbiIndex dbi, unsigned int flags);
+    int (*verify)(dbiIndex dbi, unsigned int flags);
+    void (*setFSync)(rpmdb rdb, int enable);
+    int (*ctrl)(rpmdb rdb, dbCtrlOp ctrl);
+
+    dbiCursor (*cursorInit)(dbiIndex dbi, unsigned int flags);
+    dbiCursor (*cursorFree)(dbiIndex dbi, dbiCursor dbc);
+
+    rpmRC (*pkgdbGet)(dbiIndex dbi, dbiCursor dbc, unsigned int hdrNum, unsigned char **hdrBlob, unsigned int *hdrLen);
+    rpmRC (*pkgdbPut)(dbiIndex dbi, dbiCursor dbc, unsigned int *hdrNum, unsigned char *hdrBlob, unsigned int hdrLen);
+    rpmRC (*pkgdbDel)(dbiIndex dbi, dbiCursor dbc,  unsigned int hdrNum);
+    unsigned int (*pkgdbKey)(dbiIndex dbi, dbiCursor dbc);
+
+    rpmRC (*idxdbGet)(dbiIndex dbi, dbiCursor dbc, const char *keyp, size_t keylen, dbiIndexSet *set, int curFlags);
+    rpmRC (*idxdbPut)(dbiIndex dbi, rpmTagVal rpmtag, unsigned int hdrNum, Header h);
+    rpmRC (*idxdbDel)(dbiIndex dbi, rpmTagVal rpmtag, unsigned int hdrNum, Header h);
+    const void * (*idxdbKey)(dbiIndex dbi, dbiCursor dbc, unsigned int *keylen);
+};
+
+#if defined(WITH_BDB)
 RPM_GNUC_INTERNAL
-rpmRC dbcCursorDel(dbiCursor dbc, const char *keyp, size_t keylen,
-		   dbiIndexItem rec);
+extern struct rpmdbOps_s db3_dbops;
+#endif
+
+#if defined(WITH_BDB_RO)
+RPM_GNUC_INTERNAL
+extern struct rpmdbOps_s bdbro_dbops;
+#endif
+
+#ifdef ENABLE_NDB
+RPM_GNUC_INTERNAL
+extern struct rpmdbOps_s ndb_dbops;
+#endif
+
+#if defined(WITH_SQLITE)
+RPM_GNUC_INTERNAL
+extern struct rpmdbOps_s sqlite_dbops;
+#endif
+
+RPM_GNUC_INTERNAL
+extern struct rpmdbOps_s dummydb_dbops;
 
 #ifdef __cplusplus
 }
