@@ -17,6 +17,7 @@
 #include <sys/types.h> /* already included from system.h */
 #endif
 #include <string.h>
+#include <fcntl.h>
 
 #include <rpm/rpmio.h>
 #include <rpm/rpmlog.h>
@@ -368,7 +369,7 @@ int rpmcpioHeaderRead(rpmcpio_t cpio, char ** path, int * fx)
     read = Fread(&magic, 6, 1, cpio->fd);
     cpio->offset += read;
     if (read != 6)
-        return RPMERR_READ_FAILED;
+	return RPMERR_BAD_MAGIC;
 
     /* read stripped header */
     if (!strncmp(CPIO_STRIPPED_MAGIC, magic,
@@ -377,7 +378,7 @@ int rpmcpioHeaderRead(rpmcpio_t cpio, char ** path, int * fx)
         read = Fread(&shdr, STRIPPED_PHYS_HDR_SIZE, 1, cpio->fd);
         cpio->offset += read;
         if (read != STRIPPED_PHYS_HDR_SIZE)
-            return RPMERR_READ_FAILED;
+	    return RPMERR_BAD_HEADER;
 
         GET_NUM_FIELD(shdr.fx, *fx);
         rc = rpmcpioReadPad(cpio);
@@ -395,10 +396,13 @@ int rpmcpioHeaderRead(rpmcpio_t cpio, char ** path, int * fx)
     read = Fread(&hdr, PHYS_HDR_SIZE, 1, cpio->fd);
     cpio->offset += read;
     if (read != PHYS_HDR_SIZE)
-	return RPMERR_READ_FAILED;
+        return RPMERR_BAD_HEADER;
 
     GET_NUM_FIELD(hdr.filesize, fsize);
     GET_NUM_FIELD(hdr.namesize, nameSize);
+    if (nameSize <= 0 || nameSize > 4096) {
+        return RPMERR_BAD_HEADER;
+    }
 
     char name[nameSize + 1];
     read = Fread(name, nameSize, 1, cpio->fd);
