@@ -17,6 +17,8 @@
 #include "lib/rpmfi_internal.h" /* rpmfiles stuff for now */
 #include "lib/misc.h"
 
+#include "lib/backend/dbiset.h"
+
 #include "debug.h"
 
 const char * const RPMVERSION = VERSION;
@@ -40,7 +42,7 @@ const int rpmFLAGS = RPMSENSE_EQUAL;
 #undef HTKEYTYPE
 #undef HTDATATYPE
 
-#define HASHTYPE removedHash
+#define HASHTYPE packageHash
 #define HTKEYTYPE unsigned int
 #define HTDATATYPE struct rpmte_s *
 #include "rpmhash.C"
@@ -48,26 +50,22 @@ const int rpmFLAGS = RPMSENSE_EQUAL;
 #undef HTKEYTYPE
 #undef HTDATATYPE
 
-#define HASHTYPE conflictsCache
-#define HTKEYTYPE const char *
+#define HASHTYPE filedepHash
+#define HTKEYTYPE rpmsid
+#define HTDATATYPE rpmsid
 #include "rpmhash.H"
 #include "rpmhash.C"
 #undef HASHTYPE
 #undef HTKEYTYPE
+#undef HTDATATYPE
 
-#define HASHTYPE requiresCache
-#define HTKEYTYPE const char *
-#include "rpmhash.H"
-#include "rpmhash.C"
+#define HASHTYPE depexistsHash
+#define HTKEYTYPE rpmsid
+#include "lib/rpmhash.H"
+#include "lib/rpmhash.C"
 #undef HASHTYPE
 #undef HTKEYTYPE
 
-enum addOp_e {
-    RPMTE_INSTALL	= 0,
-    RPMTE_UPGRADE	= 1,
-    RPMTE_REINSTALL	= 2,
-};
-
 /**
  * Check for supported payload format in header.
  * @param h		header to check
@@ -109,22 +107,24 @@ static rpmRC headerCheckPayloadFormat(Header h) {
 static int removePackage(rpmts ts, Header h, rpmte depends)
 {
     tsMembers tsmem = rpmtsMembers(ts);
-    rpmte p;
+    rpmte p, *pp;
     unsigned int dboffset = headerGetInstance(h);
 
     /* Can't remove what's not installed */
     if (dboffset == 0) return 1;
 
     /* Filter out duplicate erasures. */
-    if (removedHashHasEntry(tsmem->removedPackages, dboffset)) {
-        return 0;
+    if (packageHashGetEntry(tsmem->removedPackages, dboffset, &pp, NULL, NULL)) {
+	if (depends)
+	    rpmteSetDependsOn(pp[0], depends);
+	return 0;
     }
 
-    p = rpmteNew(ts, h, TR_REMOVED, NULL, NULL);
+    p = rpmteNew(ts, h, TR_REMOVED, NULL, NULL, 0);
     if (p == NULL)
 	return 1;
 
-    removedHashAddEntry(tsmem->removedPackages, dboffset, p);
+    packageHashAddEntry(tsmem->removedPackages, dboffset, p);
 
     if (tsmem->orderCount >= tsmem->orderAlloced) {
 	tsmem->orderAlloced += (tsmem->orderCount - tsmem->orderAlloced) + tsmem->delta;
@@ -172,23 +172,22 @@ static int addSelfErasures(rpmts ts, rpm_color_t tscolor, int op,
     Header oh;
     rpmdbMatchIterator mi = rpmtsInitIterator(ts, RPMDBI_NAME, rpmteN(p), 0);
     int rc = 0;
-    int cmp;
 
-    while((oh = rpmdbNextIterator(mi)) != NULL) {
+    while ((oh = rpmdbNextIterator(mi)) != NULL) {
 	/* Ignore colored packages not in our rainbow. */
 	if (skipColor(tscolor, hcolor, headerGetNumber(oh, RPMTAG_HEADERCOLOR)))
 	    continue;
 
-	cmp = rpmVersionCompare(h, oh);
+	/* On reinstall, skip packages with differing NEVRA. */
+	if (op != RPMTE_UPGRADE) {
+	    char * ohNEVRA = headerGetAsString(oh, RPMTAG_NEVRA);
+	    if (!rstreq(rpmteNEVRA(p), ohNEVRA)) {
+		free(ohNEVRA);
+		continue;
+	    }
+	    free(ohNEVRA);
+	}
 
-	/* On upgrade, skip packages that contain identical NEVR. */
-	if ((op == RPMTE_UPGRADE) && (cmp == 0))
-	    continue;
-
-	/* On reinstall, skip packages with differing NEVR. */
-	if ((op == RPMTE_REINSTALL) && (cmp != 0))
-	    continue;
-	
 	if (removePackage(ts, oh, p)) {
 	    rc = 1;
 	    break;
@@ -215,26 +214,16 @@ static int addObsoleteErasures(rpmts ts, rpm_color_t tscolor, rpmte p)
 
 	mi = rpmtsPrunedIterator(ts, RPMDBI_NAME, Name, 1);
 
-	while((oh = rpmdbNextIterator(mi)) != NULL) {
-	    const char *oarch = headerGetString(oh, RPMTAG_ARCH);
+	while ((oh = rpmdbNextIterator(mi)) != NULL) {
 	    int match;
 
-	    /* avoid self-obsoleting packages */
-	    if (rstreq(rpmteN(p), Name) && rstreq(rpmteA(p), oarch)) {
-		char * ohNEVRA = headerGetAsString(oh, RPMTAG_NEVRA);
-		rpmlog(RPMLOG_DEBUG, "  Not obsoleting: %s\n", ohNEVRA);
-		free(ohNEVRA);
-		continue;
-	    }
-
 	    /*
 	     * Rpm prior to 3.0.3 does not have versioned obsoletes.
 	     * If no obsoletes version info is available, match all names.
 	     */
 	    match = (rpmdsEVR(obsoletes) == NULL);
 	    if (!match)
-		match = rpmdsMatches(tspool, oh, -1, obsoletes, 1,
-					 _rpmds_nopromote);
+		match = rpmdsMatches(tspool, oh, -1, obsoletes, 1);
 
 	    if (match) {
 		char * ohNEVRA = headerGetAsString(oh, RPMTAG_NEVRA);
@@ -394,10 +383,8 @@ rpmal rpmtsCreateAl(rpmts ts, rpmElementTypes types)
     if (ts) {
 	rpmte p;
 	rpmtsi pi;
-	rpmstrPool tspool = rpmtsPool(ts);
 
-	al = rpmalCreate(tspool, (rpmtsNElements(ts) / 4) + 1, rpmtsFlags(ts),
-				rpmtsColor(ts), rpmtsPrefColor(ts));
+	al = rpmalCreate(ts, (rpmtsNElements(ts) / 4) + 1);
 	pi = rpmtsiInit(ts);
 	while ((p = rpmtsiNext(pi, types)))
 	    rpmalAdd(al, p);
@@ -432,7 +419,7 @@ static int addPackage(rpmts ts, Header h,
 	    goto exit;
     }
 
-    p = rpmteNew(ts, h, TR_ADDED, key, relocs);
+    p = rpmteNew(ts, h, TR_ADDED, key, relocs, op);
     if (p == NULL) {
 	ec = 1;
 	goto exit;
@@ -465,8 +452,7 @@ static int addPackage(rpmts ts, Header h,
     }
     
     if (tsmem->addedPackages == NULL) {
-	tsmem->addedPackages = rpmalCreate(rpmtsPool(ts), 5, rpmtsFlags(ts),
-					   tscolor, rpmtsPrefColor(ts));
+	tsmem->addedPackages = rpmalCreate(ts, 5);
     }
     rpmalAdd(tsmem->addedPackages, p);
 
@@ -507,7 +493,7 @@ int rpmtsAddEraseElement(rpmts ts, Header h, int dboffset)
 }
 
 /* Cached rpmdb provide lookup, returns 0 if satisfied, 1 otherwise */
-static int rpmdbProvides(rpmts ts, depCache dcache, rpmds dep)
+static int rpmdbProvides(rpmts ts, depCache dcache, rpmds dep, dbiIndexSet *matches)
 {
     const char * Name = rpmdsN(dep);
     const char * DNEVR = rpmdsDNEVR(dep);
@@ -521,7 +507,7 @@ static int rpmdbProvides(rpmts ts, depCache dcache, rpmds dep)
     unsigned int keyhash = 0;
 
     /* See if we already looked this up */
-    if (prune) {
+    if (prune && !matches) {
 	keyhash = depCacheKeyHash(dcache, DNEVR);
 	if (depCacheGetHEntry(dcache, DNEVR, keyhash, &cachedrc, NULL, NULL)) {
 	    rc = *cachedrc;
@@ -530,6 +516,8 @@ static int rpmdbProvides(rpmts ts, depCache dcache, rpmds dep)
 	}
     }
 
+    if (matches)
+	*matches = dbiIndexSetNew(0);
     /*
      * See if a filename dependency is a real file in some package,
      * taking file state into account: replaced, wrong colored and
@@ -544,6 +532,10 @@ static int rpmdbProvides(rpmts ts, depCache dcache, rpmds dep)
 		if (instance && instance == rpmdsInstance(dep))
 		    continue;
 	    }
+	    if (matches) {
+		dbiIndexSetAppendOne(*matches, headerGetInstance(h), 0, 0);
+		continue;
+	    }
 	    rpmdsNotify(dep, "(db files)", rc);
 	    break;
 	}
@@ -565,8 +557,7 @@ static int rpmdbProvides(rpmts ts, depCache dcache, rpmds dep)
 	while ((h = rpmdbNextIterator(mi)) != NULL) {
 	    /* Provide-indexes can't be used with nevr-only matching */
 	    int prix = (selfevr) ? -1 : rpmdbGetIteratorFileNum(mi);
-	    int match = rpmdsMatches(tspool, h, prix, dep, selfevr,
-					_rpmds_nopromote);
+	    int match = rpmdsMatches(tspool, h, prix, dep, selfevr);
 	    /* Ignore self-obsoletes and self-conflicts */
 	    if (match && (deptag == RPMTAG_OBSOLETENAME || deptag == RPMTAG_CONFLICTNAME)) {
 		unsigned int instance = headerGetInstance(h);
@@ -574,6 +565,10 @@ static int rpmdbProvides(rpmts ts, depCache dcache, rpmds dep)
 		    match = 0;
 	    }
 	    if (match) {
+		if (matches) {
+		    dbiIndexSetAppendOne(*matches, headerGetInstance(h), 0, 0);
+		    continue;
+		}
 		rpmdsNotify(dep, "(db provides)", rc);
 		break;
 	    }
@@ -582,16 +577,85 @@ static int rpmdbProvides(rpmts ts, depCache dcache, rpmds dep)
     }
     rc = (h != NULL) ? 0 : 1;
 
+    if (matches) {
+	dbiIndexSetUniq(*matches, 0);
+	rc = dbiIndexSetCount(*matches) ? 0 : 1;
+    }
+
     /* Cache the relatively expensive rpmdb lookup results */
     /* Caching the oddball non-pruned case would mess up other results */
-    if (prune)
+    if (prune && !matches)
 	depCacheAddHEntry(dcache, xstrdup(DNEVR), keyhash, rc);
     return rc;
 }
 
+static dbiIndexSet unsatisfiedDependSet(rpmts ts, rpmds dep)
+{
+    dbiIndexSet set1 = NULL, set2 = NULL;
+    tsMembers tsmem = rpmtsMembers(ts);
+    rpmsenseFlags dsflags = rpmdsFlags(dep);
+
+    if (dsflags & RPMSENSE_RPMLIB)
+	goto exit;
+
+    if (rpmdsIsRich(dep)) {
+	rpmds ds1, ds2; 
+	rpmrichOp op;
+	char *emsg = 0; 
+
+	if (rpmdsParseRichDep(dep, &ds1, &ds2, &op, &emsg) != RPMRC_OK) {
+	    rpmdsNotify(dep, emsg ? emsg : "(parse error)", 1);  
+	    _free(emsg);
+	    goto exit;
+	}
+	/* only a subset of ops is supported in set mode */
+	if (op != RPMRICHOP_WITH && op != RPMRICHOP_WITHOUT
+            && op != RPMRICHOP_OR && op != RPMRICHOP_SINGLE) {
+	    rpmdsNotify(dep, "(unsupported op in set mode)", 1);  
+	    goto exit_rich;
+	}
+
+	set1 = unsatisfiedDependSet(ts, ds1);
+	if (op == RPMRICHOP_SINGLE)
+	    goto exit_rich;
+	if (op != RPMRICHOP_OR && dbiIndexSetCount(set1) == 0)
+	    goto exit_rich;
+	set2 = unsatisfiedDependSet(ts, ds2);
+	if (op == RPMRICHOP_WITH) {
+	    dbiIndexSetFilterSet(set1, set2, 0);
+	} else if (op == RPMRICHOP_WITHOUT) {
+	    dbiIndexSetPruneSet(set1, set2, 0);
+	} else if (op == RPMRICHOP_OR) {
+	    dbiIndexSetAppendSet(set1, set2, 0);
+	}
+exit_rich:
+	ds1 = rpmdsFree(ds1);
+	ds2 = rpmdsFree(ds2);
+	goto exit;
+    }
+
+    /* match database entries */
+    rpmdbProvides(ts, NULL, dep, &set1);
+
+    /* Pretrans dependencies can't be satisfied by added packages. */
+    if (!(dsflags & RPMSENSE_PRETRANS)) {
+	rpmte *matches = rpmalAllSatisfiesDepend(tsmem->addedPackages, dep);
+	if (matches) {
+	    for (rpmte *p = matches; *p; p++)
+		dbiIndexSetAppendOne(set1, rpmalLookupTE(tsmem->addedPackages, *p), 1, 0);
+	}
+	_free(matches);
+    }
+
+exit:
+    set2 = dbiIndexSetFree(set2);
+    return set1 ? set1 : dbiIndexSetNew(0);
+}
+
 /**
  * Check dep for an unsatisfied dependency.
  * @param ts		transaction set
+ * @param dcache	dependency cache
  * @param dep		dependency
  * @return		0 if satisfied, 1 if not satisfied
  */
@@ -623,40 +687,81 @@ retry:
     }
 
     /* Dont look at pre-requisites of already installed packages */
-    if (!adding && isInstallPreReq(dsflags) && !isErasePreReq(dsflags))
+    if (!adding && isTransientReq(dsflags))
 	goto exit;
 
+    /* Handle rich dependencies */
+    if (rpmdsIsRich(dep)) {
+	rpmds ds1, ds2; 
+	rpmrichOp op;
+	char *emsg = 0; 
+	if (rpmdsParseRichDep(dep, &ds1, &ds2, &op, &emsg) != RPMRC_OK) {
+	    rc = rpmdsTagN(dep) == RPMTAG_CONFLICTNAME ? 0 : 1;
+	    if (rpmdsInstance(dep) != 0)
+		rc = !rc;	/* ignore errors for installed packages */
+	    rpmdsNotify(dep, emsg ? emsg : "(parse error)", rc);  
+	    _free(emsg);
+	    goto exit;
+	}
+	if (op == RPMRICHOP_WITH || op == RPMRICHOP_WITHOUT) {
+	    /* switch to set mode processing */
+	    dbiIndexSet set = unsatisfiedDependSet(ts, dep);
+	    rc = dbiIndexSetCount(set) ? 0 : 1;
+	    dbiIndexSetFree(set);
+	    ds1 = rpmdsFree(ds1);
+	    ds2 = rpmdsFree(ds2);
+	    rpmdsNotify(dep, "(rich)", rc);
+	    goto exit;
+	}
+	if (op == RPMRICHOP_IF || op == RPMRICHOP_UNLESS) {
+	    /* A IF B -> A OR NOT(B) */
+	    /* A UNLESS B -> A AND NOT(B) */
+	    if (rpmdsIsRich(ds2)) {
+		/* check if this has an ELSE clause */
+		rpmds ds21 = NULL, ds22 = NULL;
+		rpmrichOp op2;
+		if (rpmdsParseRichDep(ds2, &ds21, &ds22, &op2, NULL) == RPMRC_OK && op2 == RPMRICHOP_ELSE) {
+		    /* A IF B ELSE C -> (A OR NOT(B)) AND (C OR B) */
+		    /* A UNLESS B ELSE C -> (A AND NOT(B)) OR (C AND B) */
+		    rc = !unsatisfiedDepend(ts, dcache, ds21);	/* NOT(B) */
+		    if ((rc && op == RPMRICHOP_IF) || (!rc && op == RPMRICHOP_UNLESS)) {
+			rc = unsatisfiedDepend(ts, dcache, ds1);	/* A */
+		    } else {
+			rc = unsatisfiedDepend(ts, dcache, ds22);	/* C */
+		    }
+		    rpmdsFree(ds21);
+		    rpmdsFree(ds22);
+		    goto exitrich;
+		}
+		rpmdsFree(ds21);
+		rpmdsFree(ds22);
+	    }
+	    rc = !unsatisfiedDepend(ts, dcache, ds2);	/* NOT(B) */
+	    if ((rc && op == RPMRICHOP_IF) || (!rc && op == RPMRICHOP_UNLESS))
+		rc = unsatisfiedDepend(ts, dcache, ds1);
+	} else {
+	    rc = unsatisfiedDepend(ts, dcache, ds1);
+	    if ((rc && op == RPMRICHOP_OR) || (!rc && op == RPMRICHOP_AND))
+		rc = unsatisfiedDepend(ts, dcache, ds2);
+	}
+exitrich:
+	ds1 = rpmdsFree(ds1);
+	ds2 = rpmdsFree(ds2);
+	rpmdsNotify(dep, "(rich)", rc);
+	goto exit;
+    }
+
     /* Pretrans dependencies can't be satisfied by added packages. */
     if (!(dsflags & RPMSENSE_PRETRANS)) {
 	rpmte *matches = rpmalAllSatisfiesDepend(tsmem->addedPackages, dep);
-	int match = 0;
-
-	/*
-	 * Handle definitive matches within the added package set.
-	 * Self-obsoletes and -conflicts fall through here as we need to 
-	 * check for possible other matches in the rpmdb.
-	 */
-	for (rpmte *m = matches; m && *m; m++) {
-	    rpmTagVal dtag = rpmdsTagN(dep);
-	    /* Requires match, look no further */
-	    if (dtag == RPMTAG_REQUIRENAME) {
-		match = 1;
-		break;
-	    }
-
-	    /* Conflicts/obsoletes match on another package, look no further */
-	    if (rpmteDS(*m, dtag) != dep) {
-		match = 1;
-		break;
-	    }
-	}
-	free(matches);
+	int match = matches && *matches;
+	_free(matches);
 	if (match)
 	    goto exit;
     }
 
     /* See if the rpmdb provides it */
-    if (rpmdbProvides(ts, dcache, dep) == 0)
+    if (rpmdbProvides(ts, dcache, dep, NULL) == 0)
 	goto exit;
 
     /* Search for an unsatisfied dependency. */
@@ -707,16 +812,27 @@ static void checkDS(rpmts ts, depCache dcache, rpmte te,
 
 /* Check a given dependency against installed packages */
 static void checkInstDeps(rpmts ts, depCache dcache, rpmte te,
-			  rpmTag depTag, const char *dep)
+			  rpmTag depTag, const char *dep, rpmds depds, int neg)
 {
     Header h;
-    rpmdbMatchIterator mi = rpmtsPrunedIterator(ts, depTag, dep, 1);
+    rpmdbMatchIterator mi;
     rpmstrPool pool = rpmtsPool(ts);
+    char *ndep = NULL;
     /* require-problems are unsatisfied, others appear "satisfied" */
     int is_problem = (depTag == RPMTAG_REQUIRENAME);
 
+    if (depds)
+	dep = rpmdsN(depds);
+    if (neg) {
+	ndep = rmalloc(strlen(dep) + 2);
+	ndep[0] = '!';
+	strcpy(ndep + 1, dep);
+	dep = ndep;
+    }
+
+    mi = rpmtsPrunedIterator(ts, depTag, dep, 1);
     while ((h = rpmdbNextIterator(mi)) != NULL) {
-	char * pkgNEVRA;
+	int match = 1;
 	rpmds ds;
 
 	/* Ignore self-obsoletes and self-conflicts */
@@ -726,19 +842,126 @@ static void checkInstDeps(rpmts ts, depCache dcache, rpmte te,
 		continue;
 	}
 
-	pkgNEVRA = headerGetAsString(h, RPMTAG_NEVRA);
 	ds = rpmdsNewPool(pool, h, depTag, 0);
 	rpmdsSetIx(ds, rpmdbGetIteratorFileNum(mi));
 
-	if (unsatisfiedDepend(ts, dcache, ds) == is_problem)
+	/* Is it in our range at all? (but file deps have no range) */
+	if (depds)
+	    match = rpmdsCompare(ds, depds);
+
+	if (match && unsatisfiedDepend(ts, dcache, ds) == is_problem) {
+	    char *pkgNEVRA = headerGetAsString(h, RPMTAG_NEVRA);
 	    rpmteAddDepProblem(te, pkgNEVRA, ds, NULL);
+	    free(pkgNEVRA);
+	}
 
 	rpmdsFree(ds);
-	free(pkgNEVRA);
     }
     rpmdbFreeIterator(mi);
+    free(ndep);
 }
 
+static void checkInstFileDeps(rpmts ts, depCache dcache, rpmte te,
+			      rpmTag depTag, rpmfi fi, int is_not,
+			      filedepHash cache, fingerPrintCache *fpcp)
+{
+    rpmstrPool pool = rpmtsPool(ts);
+    fingerPrintCache fpc = *fpcp;
+    fingerPrint * fp = NULL;
+    rpmsid basename = rpmfiBNId(fi);
+    rpmsid dirname;
+    rpmsid *dirnames = 0;
+    int ndirnames = 0;
+    int i;
+
+    filedepHashGetEntry(cache, basename, &dirnames, &ndirnames, NULL);
+    if (!ndirnames)
+	return;
+    dirname = rpmfiDNId(fi);
+    for (i = 0; i < ndirnames; i++) {
+	char *fpdep = 0;
+	const char *dep;
+	if (dirnames[i] == dirname) {
+	    dep = rpmfiFN(fi);
+	} else {
+	    if (!fpc)
+		*fpcp = fpc = fpCacheCreate(1001, pool);
+	    if (!fp)
+		fpLookupId(fpc, dirname, basename, &fp);
+	    if (!fpLookupEqualsId(fpc, fp, dirnames[i], basename))
+		continue;
+	    rstrscat(&fpdep, rpmstrPoolStr(pool, dirnames[i]),
+                             rpmstrPoolStr(pool, basename), NULL);
+	    dep = fpdep;
+	}
+	checkInstDeps(ts, dcache, te, depTag, dep, NULL, is_not);
+	_free(fpdep);
+    }
+    _free(fp);
+}
+
+static void addFileDepToHash(rpmstrPool pool, filedepHash hash, char *key, size_t keylen)
+{
+    int i;
+    rpmsid basename, dirname;
+    if (!keylen || key[0] != '/')
+	return;
+    for (i = keylen - 1; key[i] != '/'; i--) 
+	;
+    i++;	/* include '/' in dirname */
+    dirname = rpmstrPoolIdn(pool, key, i, 1);
+    basename = rpmstrPoolIdn(pool, key + i, keylen - i, 1);
+    filedepHashAddEntry(hash, basename, dirname);
+}
+
+static void addDepToHash(rpmstrPool pool, depexistsHash hash, char *key, size_t keylen)
+{
+    if (keylen)
+	depexistsHashAddEntry(hash, rpmstrPoolIdn(pool, key, keylen, 1));
+}
+
+static void addIndexToDepHashes(rpmts ts, rpmDbiTag tag,
+				depexistsHash dephash, filedepHash filehash,
+				depexistsHash depnothash, filedepHash filenothash)
+{
+    rpmstrPool pool = rpmtsPool(ts);
+    char *key;
+    size_t keylen;
+    rpmdbIndexIterator ii = rpmdbIndexKeyIteratorInit(rpmtsGetRdb(ts), tag);
+
+    if (!ii)
+	return;
+    while ((rpmdbIndexIteratorNext(ii, (const void**)&key, &keylen)) == 0) {
+	if (!key || !keylen)
+	    continue;
+	if (*key == '!' && keylen > 1) {
+	    key++;
+	    keylen--;
+	    if (*key == '/' && filenothash)
+		addFileDepToHash(pool, filenothash, key, keylen);
+	    if (depnothash)
+		addDepToHash(pool, depnothash, key, keylen);
+	} else {
+	    if (*key == '/' && filehash)
+		addFileDepToHash(pool, filehash, key, keylen);
+	    if (dephash)
+		addDepToHash(pool, dephash, key, keylen);
+	}
+    }
+    rpmdbIndexIteratorFree(ii);
+}
+
+static unsigned int sidHash(rpmsid sid)
+{
+    return sid;
+}
+
+static int sidCmp(rpmsid a, rpmsid b)
+{
+    return (a != b);
+}
+
+
 int rpmtsCheck(rpmts ts)
 {
     rpm_color_t tscolor = rpmtsColor(ts);
@@ -746,77 +969,57 @@ int rpmtsCheck(rpmts ts)
     int closeatexit = 0;
     int rc = 0;
     depCache dcache = NULL;
-    conflictsCache confcache = NULL;
-    requiresCache reqcache = NULL;
+    filedepHash confilehash = NULL;	/* file conflicts of installed packages */
+    filedepHash connotfilehash = NULL;	/* file conflicts of installed packages */
+    depexistsHash connothash = NULL;
+    filedepHash reqfilehash = NULL;	/* file requires of installed packages */
+    filedepHash reqnotfilehash = NULL;	/* file requires of installed packages */
+    depexistsHash reqnothash = NULL;
+    fingerPrintCache fpc = NULL;
+    rpmdb rdb = NULL;
     
     (void) rpmswEnter(rpmtsOp(ts, RPMTS_OP_CHECK), 0);
 
     /* Do lazy, readonly, open of rpm database. */
-    if (rpmtsGetRdb(ts) == NULL && rpmtsGetDBMode(ts) != -1) {
+    rdb = rpmtsGetRdb(ts);
+    if (rdb == NULL && rpmtsGetDBMode(ts) != -1) {
 	if ((rc = rpmtsOpenDB(ts, rpmtsGetDBMode(ts))) != 0)
 	    goto exit;
+	rdb = rpmtsGetRdb(ts);
 	closeatexit = 1;
     }
 
+    if (rdb)
+	rpmdbCtrl(rdb, RPMDB_CTRL_LOCK_RO);
+
     /* XXX FIXME: figure some kind of heuristic for the cache size */
     dcache = depCacheCreate(5001, rstrhash, strcmp,
 				     (depCacheFreeKey)rfree, NULL);
 
-    /* build cache of all file conflicts dependencies */
-    confcache = conflictsCacheCreate(257, rstrhash, strcmp,
-				     (depCacheFreeKey)rfree);
-    if (confcache) {
-	rpmdbIndexIterator ii = rpmdbIndexIteratorInit(rpmtsGetRdb(ts), RPMTAG_CONFLICTNAME);
-	if (ii) {
-	    char *key;
-	    size_t keylen;
-	    while ((rpmdbIndexIteratorNext(ii, (const void**)&key, &keylen)) == 0) {
-		char *k;
-		if (!key || keylen == 0 || key[0] != '/')
-		    continue;
-		while (keylen > 1 && (k = memchr(key + 1, '/', keylen - 1))) {
-		    keylen -= k - key;
-		    key += k - key;
-		}
-		if (keylen <= 1)
-		    continue;
-		k = rmalloc(keylen);
-		memcpy(k, key + 1, keylen - 1);
-		k[keylen - 1] = 0;
-		conflictsCacheAddEntry(confcache, k);
-	    }
-	    rpmdbIndexIteratorFree(ii);
-	}
-    }
+    /* build hashes of all confilict sdependencies */
+    confilehash = filedepHashCreate(257, sidHash, sidCmp, NULL, NULL);
+    connothash = depexistsHashCreate(257, sidHash, sidCmp, NULL);
+    connotfilehash = filedepHashCreate(257, sidHash, sidCmp, NULL, NULL);
+    addIndexToDepHashes(ts, RPMTAG_CONFLICTNAME, NULL, confilehash, connothash, connotfilehash);
+    if (!filedepHashNumKeys(confilehash))
+	confilehash = filedepHashFree(confilehash);
+    if (!depexistsHashNumKeys(connothash))
+	connothash= depexistsHashFree(connothash);
+    if (!filedepHashNumKeys(connotfilehash))
+	connotfilehash = filedepHashFree(connotfilehash);
 
-    /* build cache of all file requires dependencies */
-    reqcache = requiresCacheCreate(8191, rstrhash, strcmp,
-				     (depCacheFreeKey)rfree);
-    if (reqcache) {
-	rpmdbIndexIterator ii = rpmdbIndexIteratorInit(rpmtsGetRdb(ts), RPMTAG_REQUIRENAME);
-	if (ii) {
-	    char *key;
-	    size_t keylen;
-	    while ((rpmdbIndexIteratorNext(ii, (const void**)&key, &keylen)) == 0) {
-		char *k;
-		if (!key || keylen == 0 || key[0] != '/')
-		    continue;
-		while (keylen > 1 && (k = memchr(key + 1, '/', keylen - 1))) {
-		    keylen -= k - key;
-		    key += k - key;
-		}
-		if (keylen <= 1)
-		    continue;
-		k = rmalloc(keylen);
-		memcpy(k, key + 1, keylen - 1);
-		k[keylen - 1] = 0;
-		requiresCacheAddEntry(reqcache, k);
-	    }
-	    rpmdbIndexIteratorFree(ii);
-	}
-    }
+    /* build hashes of all requires dependencies */
+    reqfilehash = filedepHashCreate(8191, sidHash, sidCmp, NULL, NULL);
+    reqnothash = depexistsHashCreate(257, sidHash, sidCmp, NULL);
+    reqnotfilehash = filedepHashCreate(257, sidHash, sidCmp, NULL, NULL);
+    addIndexToDepHashes(ts, RPMTAG_REQUIRENAME, NULL, reqfilehash, reqnothash, reqnotfilehash);
+    if (!filedepHashNumKeys(reqfilehash))
+	reqfilehash = filedepHashFree(reqfilehash);
+    if (!depexistsHashNumKeys(reqnothash))
+	reqnothash= depexistsHashFree(reqnothash);
+    if (!filedepHashNumKeys(reqnotfilehash))
+	reqnotfilehash = filedepHashFree(reqnotfilehash);
 
-    
     /*
      * Look at all of the added packages and make sure their dependencies
      * are satisfied.
@@ -835,26 +1038,29 @@ int rpmtsCheck(rpmts ts)
 	checkDS(ts, dcache, p, rpmteNEVRA(p), rpmteDS(p, RPMTAG_OBSOLETENAME),
 		tscolor);
 
-	/* Check provides against conflicts in installed packages. */
-	while (rpmdsNext(provides) >= 0) {
-	    checkInstDeps(ts, dcache, p, RPMTAG_CONFLICTNAME, rpmdsN(provides));
-	}
-
-	/* Skip obsoletion checks for source packages (ie build) */
+	/* Skip obsoletion and provides checks for source packages (ie build) */
 	if (rpmteIsSource(p))
 	    continue;
 
+	/* Check provides against conflicts in installed packages. */
+	while (rpmdsNext(provides) >= 0) {
+	    checkInstDeps(ts, dcache, p, RPMTAG_CONFLICTNAME, NULL, provides, 0);
+	    if (reqnothash && depexistsHashHasEntry(reqnothash, rpmdsNId(provides)))
+		checkInstDeps(ts, dcache, p, RPMTAG_REQUIRENAME, NULL, provides, 1);
+	}
+
 	/* Check package name (not provides!) against installed obsoletes */
-	checkInstDeps(ts, dcache, p, RPMTAG_OBSOLETENAME, rpmteN(p));
+	checkInstDeps(ts, dcache, p, RPMTAG_OBSOLETENAME, NULL, rpmteDS(p, RPMTAG_NAME), 0);
 
 	/* Check filenames against installed conflicts */
-        if (conflictsCacheNumKeys(confcache)) {
+        if (confilehash || reqnotfilehash) {
 	    rpmfiles files = rpmteFiles(p);
 	    rpmfi fi = rpmfilesIter(files, RPMFI_ITER_FWD);
 	    while (rpmfiNext(fi) >= 0) {
-		if (!conflictsCacheHasEntry(confcache, rpmfiBN(fi)))
-		    continue;
-		checkInstDeps(ts, dcache, p, RPMTAG_CONFLICTNAME, rpmfiFN(fi));
+		if (confilehash)
+		    checkInstFileDeps(ts, dcache, p, RPMTAG_CONFLICTNAME, fi, 0, confilehash, &fpc);
+		if (reqnotfilehash)
+		    checkInstFileDeps(ts, dcache, p, RPMTAG_REQUIRENAME, fi, 1, reqnotfilehash, &fpc);
 	    }
 	    rpmfiFree(fi);
 	    rpmfilesFree(files);
@@ -868,33 +1074,46 @@ int rpmtsCheck(rpmts ts)
     pi = rpmtsiInit(ts);
     while ((p = rpmtsiNext(pi, TR_REMOVED)) != NULL) {
 	rpmds provides = rpmdsInit(rpmteDS(p, RPMTAG_PROVIDENAME));
-	rpmfiles files = rpmteFiles(p);
-	rpmfi fi = rpmfilesIter(files, RPMFI_ITER_FWD);;
 
 	rpmlog(RPMLOG_DEBUG, "========== --- %s %s/%s 0x%x\n",
 		rpmteNEVR(p), rpmteA(p), rpmteO(p), rpmteColor(p));
 
 	/* Check provides and filenames against installed dependencies. */
 	while (rpmdsNext(provides) >= 0) {
-	    checkInstDeps(ts, dcache, p, RPMTAG_REQUIRENAME, rpmdsN(provides));
+	    checkInstDeps(ts, dcache, p, RPMTAG_REQUIRENAME, NULL, provides, 0);
+	    if (connothash && depexistsHashHasEntry(connothash, rpmdsNId(provides)))
+		checkInstDeps(ts, dcache, p, RPMTAG_CONFLICTNAME, NULL, provides, 1);
 	}
 
-	while (rpmfiNext(fi) >= 0) {
-	    if (RPMFILE_IS_INSTALLED(rpmfiFState(fi))) {
-		if (!requiresCacheHasEntry(reqcache, rpmfiBN(fi)))
-		    continue;
-		checkInstDeps(ts, dcache, p, RPMTAG_REQUIRENAME, rpmfiFN(fi));
+	if (reqfilehash || connotfilehash) {
+	    rpmfiles files = rpmteFiles(p);
+	    rpmfi fi = rpmfilesIter(files, RPMFI_ITER_FWD);
+	    while (rpmfiNext(fi) >= 0) {
+		if (RPMFILE_IS_INSTALLED(rpmfiFState(fi))) {
+		    if (reqfilehash)
+			checkInstFileDeps(ts, dcache, p, RPMTAG_REQUIRENAME, fi, 0, reqfilehash, &fpc);
+		    if (connotfilehash)
+			checkInstFileDeps(ts, dcache, p, RPMTAG_CONFLICTNAME, fi, 1, connotfilehash, &fpc);
+		}
 	    }
+	    rpmfiFree(fi);
+	    rpmfilesFree(files);
 	}
-	rpmfiFree(fi);
-	rpmfilesFree(files);
     }
     rpmtsiFree(pi);
 
+    if (rdb)
+	rpmdbCtrl(rdb, RPMDB_CTRL_UNLOCK_RO);
+
 exit:
     depCacheFree(dcache);
-    conflictsCacheFree(confcache);
-    requiresCacheFree(reqcache);
+    filedepHashFree(confilehash);
+    filedepHashFree(connotfilehash);
+    depexistsHashFree(connothash);
+    filedepHashFree(reqfilehash);
+    filedepHashFree(reqnotfilehash);
+    depexistsHashFree(reqnothash);
+    fpCacheFree(fpc);
 
     (void) rpmswExit(rpmtsOp(ts, RPMTS_OP_CHECK), 0);
 
