@@ -14,6 +14,7 @@
 #include <rpm/rpmte.h>
 #include <rpm/rpmts.h>
 #include <rpm/rpmlog.h>
+#include <rpm/rpmmacro.h>
 
 #include "rpmio/rpmio_internal.h"	/* fdInit/FiniDigest */
 #include "lib/fsm.h"
@@ -67,7 +68,7 @@ typedef struct dnli_s {
 
 /** \ingroup payload
  * Destroy directory name iterator.
- * @param a		directory name iterator
+ * @param dnli		directory name iterator
  * @retval		NULL always
  */
 static DNLI_t dnlFreeIterator(DNLI_t dnli)
@@ -200,41 +201,79 @@ static int fsmSetFCaps(const char *path, const char *captxt)
 	if (fcaps == NULL || cap_set_file(path, fcaps) != 0) {
 	    rc = RPMERR_SETCAP_FAILED;
 	}
+	if (_fsm_debug) {
+	    rpmlog(RPMLOG_DEBUG, " %8s (%s, %s) %s\n", __func__,
+		   path, captxt, (rc < 0 ? strerror(errno) : ""));
+	}
 	cap_free(fcaps);
     } 
 #endif
     return rc;
 }
 
+static void wfd_close(FD_t *wfdp)
+{
+    if (wfdp && *wfdp) {
+	int myerrno = errno;
+	static int oneshot = 0;
+	static int flush_io = 0;
+	if (!oneshot) {
+	    flush_io = (rpmExpandNumeric("%{?_flush_io}") > 0);
+	    oneshot = 1;
+	}
+	if (flush_io) {
+	    int fdno = Fileno(*wfdp);
+	    fsync(fdno);
+	}
+	Fclose(*wfdp);
+	*wfdp = NULL;
+	errno = myerrno;
+    }
+}
+
+static int wfd_open(FD_t *wfdp, const char *dest)
+{
+    int rc = 0;
+    /* Create the file with 0200 permissions (write by owner). */
+    {
+	mode_t old_umask = umask(0577);
+	*wfdp = Fopen(dest, "wx.ufdio");
+	umask(old_umask);
+    }
+    if (Ferror(*wfdp)) {
+	rc = RPMERR_OPEN_FAILED;
+	goto exit;
+    }
+
+    return 0;
+
+exit:
+    wfd_close(wfdp);
+    return rc;
+}
+
 /** \ingroup payload
  * Create file from payload stream.
  * @return		0 on success
  */
-static int expandRegular(rpmfi fi, const char *dest, rpmpsm psm, int nodigest, int nocontent)
+static int expandRegular(rpmfi fi, const char *dest, rpmpsm psm, int nodigest)
 {
     FD_t wfd = NULL;
-    int rc = 0;
+    int rc;
 
-    wfd = Fopen(dest, "w.ufdio");
-    if (Ferror(wfd)) {
-	rc = RPMERR_OPEN_FAILED;
-	goto exit;
-    }
+    rc = wfd_open(&wfd, dest);
+    if (rc != 0)
+        goto exit;
 
-    if (!nocontent)
-	rc = rpmfiArchiveReadToFilePsm(fi, wfd, nodigest, psm);
+    rc = rpmfiArchiveReadToFilePsm(fi, wfd, nodigest, psm);
+    wfd_close(&wfd);
 exit:
-    if (wfd) {
-	int myerrno = errno;
-	Fclose(wfd);
-	errno = myerrno;
-    }
     return rc;
 }
 
 static int fsmMkfile(rpmfi fi, const char *dest, rpmfiles files,
 		     rpmpsm psm, int nodigest, int *setmeta,
-		     int * firsthardlink)
+		     int * firsthardlink, FD_t *firstlinkfile)
 {
     int rc = 0;
     int numHardlinks = rpmfiFNlink(fi);
@@ -243,7 +282,7 @@ static int fsmMkfile(rpmfi fi, const char *dest, rpmfiles files,
 	/* Create first hardlinked file empty */
 	if (*firsthardlink < 0) {
 	    *firsthardlink = rpmfiFX(fi);
-	    rc = expandRegular(fi, dest, psm, nodigest, 1);
+	    rc = wfd_open(firstlinkfile, dest);
 	} else {
 	    /* Create hard links for others */
 	    char *fn = rpmfilesFN(files, *firsthardlink);
@@ -258,10 +297,11 @@ static int fsmMkfile(rpmfi fi, const char *dest, rpmfiles files,
        existing) file with content */
     if (numHardlinks<=1) {
 	if (!rc)
-	    rc = expandRegular(fi, dest, psm, nodigest, 0);
+	    rc = expandRegular(fi, dest, psm, nodigest);
     } else if (rpmfiArchiveHasContent(fi)) {
 	if (!rc)
-	    rc = expandRegular(fi, dest, psm, nodigest, 0);
+	    rc = rpmfiArchiveReadToFilePsm(fi, *firstlinkfile, nodigest, psm);
+	wfd_close(firstlinkfile);
 	*firsthardlink = -1;
     } else {
 	*setmeta = 0;
@@ -371,7 +411,8 @@ static int fsmMknod(const char *path, mode_t mode, dev_t dev)
 
 /**
  * Create (if necessary) directories not explicitly included in package.
- * @param dnli		file state machine data
+ * @param files		file data
+ * @param fs		file states
  * @param plugins	rpm plugins handle
  * @return		0 on success
  */
@@ -380,28 +421,19 @@ static int fsmMkdirs(rpmfiles files, rpmfs fs, rpmPlugins plugins)
     DNLI_t dnli = dnlInitIterator(files, fs, 0);
     struct stat sb;
     const char *dpath;
-    int dc = rpmfilesDC(files);
     int rc = 0;
     int i;
-    int ldnlen = 0;
-    int ldnalloc = 0;
-    char * ldn = NULL;
-    short * dnlx = NULL; 
+    size_t ldnlen = 0;
+    const char * ldn = NULL;
 
-    dnlx = (dc ? xcalloc(dc, sizeof(*dnlx)) : NULL);
-
-    if (dnlx != NULL)
     while ((dpath = dnlNextIterator(dnli)) != NULL) {
 	size_t dnlen = strlen(dpath);
 	char * te, dn[dnlen+1];
 
-	dc = dnli->isave;
-	if (dc < 0) continue;
-	dnlx[dc] = dnlen;
 	if (dnlen <= 1)
 	    continue;
 
-	if (dnlen <= ldnlen && rstreq(dpath, ldn))
+	if (dnlen == ldnlen && rstreq(dpath, ldn))
 	    continue;
 
 	/* Copy as we need to modify the string */
@@ -412,26 +444,19 @@ static int fsmMkdirs(rpmfiles files, rpmfs fs, rpmPlugins plugins)
 	    if (*te != '/')
 		continue;
 
-	    *te = '\0';
-
 	    /* Already validated? */
 	    if (i < ldnlen &&
 		(ldn[i] == '/' || ldn[i] == '\0') && rstreqn(dn, ldn, i))
-	    {
-		*te = '/';
-		/* Move pre-existing path marker forward. */
-		dnlx[dc] = (te - dn);
 		continue;
-	    }
 
 	    /* Validate next component of path. */
+	    *te = '\0';
 	    rc = fsmStat(dn, 1, &sb); /* lstat */
 	    *te = '/';
 
 	    /* Directory already exists? */
 	    if (rc == 0 && S_ISDIR(sb.st_mode)) {
-		/* Move pre-existing path marker forward. */
-		dnlx[dc] = (te - dn);
+		continue;
 	    } else if (rc == RPMERR_ENOENT) {
 		*te = '\0';
 		mode_t mode = S_IFDIR | (_dirPerms & 07777);
@@ -464,17 +489,9 @@ static int fsmMkdirs(rpmfiles files, rpmfs fs, rpmPlugins plugins)
 	if (rc) break;
 
 	/* Save last validated path. */
-	if (ldnalloc < (dnlen + 1)) {
-	    ldnalloc = dnlen + 100;
-	    ldn = xrealloc(ldn, ldnalloc);
-	}
-	if (ldn != NULL) { /* XXX can't happen */
-	    strcpy(ldn, dn);
-	    ldnlen = dnlen;
-	}
+	ldn = dpath;
+	ldnlen = dnlen;
     }
-    free(dnlx);
-    free(ldn);
     dnlFreeIterator(dnli);
 
     return rc;
@@ -549,7 +566,8 @@ static int fsmRename(const char *opath, const char *path)
     if (_fsm_debug)
 	rpmlog(RPMLOG_DEBUG, " %8s (%s, %s) %s\n", __func__,
 	       opath, path, (rc < 0 ? strerror(errno) : ""));
-    if (rc < 0)	rc = RPMERR_RENAME_FAILED;
+    if (rc < 0)
+	rc = (errno == EISDIR ? RPMERR_EXIST_AS_DIR : RPMERR_RENAME_FAILED);
     return rc;
 }
 
@@ -640,11 +658,14 @@ static int fsmVerify(const char *path, rpmfi fi)
     } else if (S_ISDIR(mode)) {
         if (S_ISDIR(dsb.st_mode)) return 0;
         if (S_ISLNK(dsb.st_mode)) {
+	    uid_t luid = dsb.st_uid;
             rc = fsmStat(path, 0, &dsb);
             if (rc == RPMERR_ENOENT) rc = 0;
             if (rc) return rc;
             errno = saveerrno;
-            if (S_ISDIR(dsb.st_mode)) return 0;
+	    /* Only permit directory symlinks by target owner and root */
+            if (S_ISDIR(dsb.st_mode) && (luid == 0 || luid == dsb.st_uid))
+		    return 0;
         }
     } else if (S_ISLNK(mode)) {
         if (S_ISLNK(dsb.st_mode)) {
@@ -710,7 +731,8 @@ static int fsmBackup(rpmfi fi, rpmFileAction action)
 }
 
 static int fsmSetmeta(const char *path, rpmfi fi, rpmPlugins plugins,
-		      rpmFileAction action, const struct stat * st)
+		      rpmFileAction action, const struct stat * st,
+		      int nofcaps)
 {
     int rc = 0;
     const char *dest = rpmfiFN(fi);
@@ -722,7 +744,7 @@ static int fsmSetmeta(const char *path, rpmfi fi, rpmPlugins plugins,
 	rc = fsmChmod(path, st->st_mode);
     }
     /* Set file capabilities (if enabled) */
-    if (!rc && S_ISREG(st->st_mode) && !getuid()) {
+    if (!rc && !nofcaps && S_ISREG(st->st_mode) && !getuid()) {
 	rc = fsmSetFCaps(path, rpmfiFCaps(fi));
     }
     if (!rc) {
@@ -783,6 +805,7 @@ static const char * fileActionString(rpmFileAction a)
     case FA_SKIPNSTATE: return "skipnstate";
     case FA_SKIPNETSHARED: return "skipnetshared";
     case FA_SKIPCOLOR:	return "skipcolor";
+    case FA_TOUCH:     return "touch";
     default:		return "???";
     }
 }
@@ -800,6 +823,9 @@ static void setFileState(rpmfs fs, int i)
     case FA_SKIPCOLOR:
 	rpmfsSetState(fs, i, RPMFILE_STATE_WRONGCOLOR);
 	break;
+    case FA_TOUCH:
+	rpmfsSetState(fs, i, RPMFILE_STATE_NORMAL);
+	break;
     default:
 	break;
     }
@@ -816,7 +842,9 @@ int rpmPackageFilesInstall(rpmts ts, rpmte te, rpmfiles files,
     int saveerrno = errno;
     int rc = 0;
     int nodigest = (rpmtsFlags(ts) & RPMTRANS_FLAG_NOFILEDIGEST) ? 1 : 0;
+    int nofcaps = (rpmtsFlags(ts) & RPMTRANS_FLAG_NOCAPS) ? 1 : 0;
     int firsthardlink = -1;
+    FD_t firstlinkfile = NULL;
     int skip;
     rpmFileAction action;
     char *tid = NULL;
@@ -846,7 +874,11 @@ int rpmPackageFilesInstall(rpmts ts, rpmte te, rpmfiles files,
 
 	action = rpmfsGetAction(fs, rpmfiFX(fi));
 	skip = XFA_SKIPPING(action);
-	suffix = S_ISDIR(rpmfiFMode(fi)) ? NULL : tid;
+	if (action != FA_TOUCH) {
+	    suffix = S_ISDIR(rpmfiFMode(fi)) ? NULL : tid;
+	} else {
+	    suffix = NULL;
+	}
 	fpath = fsmFsPath(fi, suffix);
 
 	/* Remap file perms, owner, and group. */
@@ -881,10 +913,21 @@ int rpmPackageFilesInstall(rpmts ts, rpmte te, rpmfiles files,
 		rc = RPMERR_ENOENT;
 	    }
 
+	    /* See if the file was removed while our attention was elsewhere */
+	    if (rc == RPMERR_ENOENT && action == FA_TOUCH) {
+		rpmlog(RPMLOG_DEBUG, "file %s vanished unexpectedly\n", fpath);
+		action = FA_CREATE;
+		fsmDebug(fpath, action, &sb);
+	    }
+
+	    /* When touching we don't need any of this... */
+	    if (action == FA_TOUCH)
+		goto touch;
+
             if (S_ISREG(sb.st_mode)) {
 		if (rc == RPMERR_ENOENT) {
 		    rc = fsmMkfile(fi, fpath, files, psm, nodigest,
-				   &setmeta, &firsthardlink);
+				   &setmeta, &firsthardlink, &firstlinkfile);
 		}
             } else if (S_ISDIR(sb.st_mode)) {
                 if (rc == RPMERR_ENOENT) {
@@ -914,15 +957,19 @@ int rpmPackageFilesInstall(rpmts ts, rpmte te, rpmfiles files,
                 if (!IS_DEV_LOG(fpath))
                     rc = RPMERR_UNKNOWN_FILETYPE;
             }
+
+touch:
 	    /* Set permissions, timestamps etc for non-hardlink entries */
 	    if (!rc && setmeta) {
-		rc = fsmSetmeta(fpath, fi, plugins, action, &sb);
+		rc = fsmSetmeta(fpath, fi, plugins, action, &sb, nofcaps);
 	    }
         } else if (firsthardlink >= 0 && rpmfiArchiveHasContent(fi)) {
+	    /* On FA_TOUCH no hardlinks are created thus this is skipped. */
 	    /* we skip the hard linked file containing the content */
 	    /* write the content to the first used instead */
 	    char *fn = rpmfilesFN(files, firsthardlink);
-	    rc = expandRegular(fi, fn, psm, nodigest, 0);
+	    rc = rpmfiArchiveReadToFilePsm(fi, firstlinkfile, nodigest, psm);
+	    wfd_close(&firstlinkfile);
 	    firsthardlink = -1;
 	    free(fn);
 	}
