@@ -55,23 +55,25 @@ static void rpmTSIFree(tsortInfo tsi)
     while (tsi->tsi_relations != NULL) {
 	rel = tsi->tsi_relations;
 	tsi->tsi_relations = tsi->tsi_relations->rel_next;
-	rel = _free(rel);
+	free(rel);
     }
     while (tsi->tsi_forward_relations != NULL) {
 	rel = tsi->tsi_forward_relations;
 	tsi->tsi_forward_relations = \
 	    tsi->tsi_forward_relations->rel_next;
-	rel = _free(rel);
+	free(rel);
     }
 }
 
 static inline int addSingleRelation(rpmte p,
 				    rpmte q,
-				    rpmsenseFlags dsflags)
+				    rpmds dep)
 {
     struct tsortInfo_s *tsi_p, *tsi_q;
     relation rel;
     rpmElementType teType = rpmteType(p);
+    rpmsenseFlags dsflags = rpmdsFlags(dep);
+    int reversed = rpmdsIsReverse(dep);
     rpmsenseFlags flags;
 
     /* Avoid deps outside this transaction and self dependencies */
@@ -80,9 +82,7 @@ static inline int addSingleRelation(rpmte p,
 
     /* Erasures are reversed installs. */
     if (teType == TR_REMOVED) {
-	rpmte r = p;
-	p = q;
-	q = r;
+	reversed = ! reversed;
 	flags = isErasePreReq(dsflags);
     } else {
 	flags = isInstallPreReq(dsflags);
@@ -94,21 +94,51 @@ static inline int addSingleRelation(rpmte p,
 	    RPMSENSE_SCRIPT_PRE : RPMSENSE_SCRIPT_PREUN;
     }
 
+    if (reversed) {
+	rpmte r = p;
+	p = q;
+	q = r;
+    }
+
     tsi_p = rpmteTSI(p);
     tsi_q = rpmteTSI(q);
 
     /* if relation got already added just update the flags */
-    if (tsi_q->tsi_relations && tsi_q->tsi_relations->rel_suc == tsi_p) {
+    if (!reversed &&
+	tsi_q->tsi_relations && tsi_q->tsi_relations->rel_suc == tsi_p) {
+	/* must be latest one added to q as we add all rels to p at once */
 	tsi_q->tsi_relations->rel_flags |= flags;
-	tsi_p->tsi_forward_relations->rel_flags |= flags;
-	return 0;
+	/* search entry in p */
+	for (struct relation_s * tsi = tsi_p->tsi_forward_relations;
+	     tsi; tsi = tsi->rel_next) {
+	    if (tsi->rel_suc == tsi_q) {
+		tsi->rel_flags |= flags;
+		return 0;
+	    }
+	}
+	assert(0);
+    }
+
+    /* if relation got already added just update the flags */
+    if (reversed && tsi_q->tsi_forward_relations &&
+	tsi_q->tsi_forward_relations->rel_suc == tsi_p) {
+	/* must be latest one added to q as we add all rels to p at once */
+	tsi_q->tsi_forward_relations->rel_flags |= flags;
+	/* search entry in p */
+	for (struct relation_s * tsi = tsi_p->tsi_relations;
+	     tsi; tsi = tsi->rel_next) {
+	    if (tsi->rel_suc == tsi_q) {
+		tsi->rel_flags |= flags;
+		return 0;
+	    }
+	}
+	assert(0);
     }
 
     /* Record next "q <- p" relation (i.e. "p" requires "q"). */
-    if (p != q) {
-	/* bump p predecessor count */
-	tsi_p->tsi_count++;
-    }
+
+    /* bump p predecessor count */
+    tsi_p->tsi_count++;
 
     rel = xcalloc(1, sizeof(*rel));
     rel->rel_suc = tsi_p;
@@ -116,10 +146,10 @@ static inline int addSingleRelation(rpmte p,
 
     rel->rel_next = tsi_q->tsi_relations;
     tsi_q->tsi_relations = rel;
-    if (p != q) {
-	/* bump q successor count */
-	tsi_q->tsi_qcnt++;
-    }
+
+    
+    /* bump q successor count */
+    tsi_q->tsi_qcnt++;
 
     rel = xcalloc(1, sizeof(*rel));
     rel->rel_suc = tsi_q;
@@ -134,65 +164,51 @@ static inline int addSingleRelation(rpmte p,
 /**
  * Record next "q <- p" relation (i.e. "p" requires "q").
  * @param ts		transaction set
+ * @param al		packages list
  * @param p		predecessor (i.e. package that "Requires: q")
- * @param requires	relation
+ * @param dep		dependency relation
  * @return		0 always
  */
 static inline int addRelation(rpmts ts,
 			      rpmal al,
 			      rpmte p,
-			      rpmds requires)
+			      rpmds dep)
 {
     rpmte q;
-    rpmsenseFlags dsflags;
-
-    dsflags = rpmdsFlags(requires);
 
     /* Avoid dependendencies which are not relevant for ordering */
-    if (dsflags & (RPMSENSE_RPMLIB|RPMSENSE_CONFIG|RPMSENSE_PRETRANS|RPMSENSE_POSTTRANS))
+    if (isUnorderedReq(rpmdsFlags(dep)))
 	return 0;
 
-    q = rpmalSatisfiesDepend(al, requires);
+    if (rpmdsIsRich(dep)) {
+	rpmds ds1, ds2;
+	rpmrichOp op;
+	if (rpmdsParseRichDep(dep, &ds1, &ds2, &op, NULL) == RPMRC_OK) {
+	    if (op != RPMRICHOP_ELSE)
+		addRelation(ts, al, p, ds1);
+	    if (op == RPMRICHOP_IF || op == RPMRICHOP_UNLESS) {
+	      rpmds ds21, ds22;
+	      rpmrichOp op2;
+	      if (rpmdsParseRichDep(dep, &ds21, &ds22, &op2, NULL) == RPMRC_OK && op2 == RPMRICHOP_ELSE) {
+		  addRelation(ts, al, p, ds22);
+	      }
+	      ds21 = rpmdsFree(ds21);
+	      ds22 = rpmdsFree(ds22);
+	    }
+	    if (op == RPMRICHOP_AND || op == RPMRICHOP_OR)
+		addRelation(ts, al, p, ds2);
+	    ds1 = rpmdsFree(ds1);
+	    ds2 = rpmdsFree(ds2);
+	}
+	return 0;
+    }
+    q = rpmalSatisfiesDepend(al, p, dep);
 
     /* Avoid deps outside this transaction and self dependencies */
     if (q == NULL || q == p)
 	return 0;
 
-    addSingleRelation(p, q, dsflags);
-
-    return 0;
-}
-
-/*
- * Collections might have special ordering requirements. Notably
- * sepolicy collection requires having all the bits in the collection
- * close to each other. We try to ensure this by creating a strongly
- * connected component of such "grouped" collections, by introducing 
- * an artificial relation loop across the all its members.
- */
-static int addCollRelations(rpmal al, rpmte p, ARGV_t *seenColls)
-{
-    ARGV_const_t qcolls;
-
-    for (qcolls = rpmteCollections(p); qcolls && *qcolls; qcolls++) {
-	char * flags;
-	if (argvSearch(*seenColls, *qcolls, NULL))
-	    continue;
-
-	flags = rstrscat(NULL, "%{__collection_", *qcolls, "_flags}", NULL);
-	if (rpmExpandNumeric(flags) & 0x1) {
-	    rpmte *tes = rpmalAllInCollection(al, *qcolls);
-	    for (rpmte *te = tes; te && *te; te++) {
-		rpmte next = (*(te + 1) != NULL) ? *(te + 1) : *tes;
-		addSingleRelation(*te, next, RPMSENSE_ANY);
-	    }
-	    _free(tes);
-	}
-	free(flags);
-
-	argvAdd(seenColls, *qcolls);
-	argvSort(*seenColls, NULL);
-    }
+    addSingleRelation(p, q, dep);
 
     return 0;
 }
@@ -258,7 +274,8 @@ typedef struct sccData_s {
     int sccCnt;			/* Number of SCC's found */
 } * sccData;
 
-static void tarjan(sccData sd, tsortInfo tsi) {
+static void tarjan(sccData sd, tsortInfo tsi)
+{
     tsortInfo tsi_q;
     relation rel;
 
@@ -429,38 +446,21 @@ static void collectTE(rpm_color_t prefcolor, tsortInfo q,
     q->tsi_SccIdx = 0;
 }
 
-static void collectSCC(rpm_color_t prefcolor, tsortInfo p_tsi,
-		       rpmte * newOrder, int * newOrderCount,
-		       scc SCCs, tsortInfo * queue_end)
+static void dijkstra(const struct scc_s *SCC, int sccNr)
 {
-    int sccNr = p_tsi->tsi_SccIdx;
-    struct scc_s SCC = SCCs[sccNr];
-    int i;
     int start, end;
     relation rel;
 
-    /* remove p from the outer queue */
-    tsortInfo outer_queue_start = p_tsi->tsi_suc;
-    p_tsi->tsi_suc = NULL;
-
-    /*
-     * Run a multi source Dijkstra's algorithm to find relations
-     * that can be zapped with least danger to pre reqs.
-     * As weight of the edges is always 1 it is not necessary to
-     * sort the vertices by distance as the queue gets them
-     * already in order
-    */
-
     /* can use a simple queue as edge weights are always 1 */
-    tsortInfo * queue = xmalloc((SCC.size+1) * sizeof(*queue));
+    tsortInfo * queue = xmalloc((SCC->size+1) * sizeof(*queue));
 
     /*
      * Find packages that are prerequired and use them as
      * starting points for the Dijkstra algorithm
      */
     start = end = 0;
-    for (i = 0; i < SCC.size; i++) {
-	tsortInfo tsi = SCC.members[i];
+    for (int i = 0; i < SCC->size; i++) {
+	tsortInfo tsi = SCC->members[i];
 	tsi->tsi_SccLowlink = INT_MAX;
 	for (rel=tsi->tsi_forward_relations; rel != NULL; rel=rel->rel_next) {
 	    if (rel->rel_flags && rel->rel_suc->tsi_SccIdx == sccNr) {
@@ -476,8 +476,8 @@ static void collectSCC(rpm_color_t prefcolor, tsortInfo p_tsi,
     }
 
     if (start == end) { /* no regular prereqs; add self prereqs to queue */
-	for (i = 0; i < SCC.size; i++) {
-	    tsortInfo tsi = SCC.members[i];
+	for (int i = 0; i < SCC->size; i++) {
+	    tsortInfo tsi = SCC->members[i];
 	    if (tsi->tsi_SccLowlink != INT_MAX) {
 		queue[end++] = tsi;
 	    }
@@ -496,8 +496,28 @@ static void collectSCC(rpm_color_t prefcolor, tsortInfo p_tsi,
 	    }
 	}
     }
-    queue = _free(queue);
+    free(queue);
+}
 
+static void collectSCC(rpm_color_t prefcolor, tsortInfo p_tsi,
+		       rpmte * newOrder, int * newOrderCount,
+		       scc SCCs, tsortInfo * queue_end)
+{
+    int sccNr = p_tsi->tsi_SccIdx;
+    const struct scc_s * SCC = SCCs+sccNr;
+
+    /* remove p from the outer queue */
+    tsortInfo outer_queue_start = p_tsi->tsi_suc;
+    p_tsi->tsi_suc = NULL;
+
+    /*
+     * Run a multi source Dijkstra's algorithm to find relations
+     * that can be zapped with least danger to pre reqs.
+     * As weight of the edges is always 1 it is not necessary to
+     * sort the vertices by distance as the queue gets them
+     * already in order
+    */
+    dijkstra(SCC, sccNr);
 
     while (1) {
 	tsortInfo best = NULL;
@@ -505,8 +525,8 @@ static void collectSCC(rpm_color_t prefcolor, tsortInfo p_tsi,
 	int best_score = 0;
 
 	/* select best candidate to start with */
-	for (int i = 0; i < SCC.size; i++) {
-	    tsortInfo tsi = SCC.members[i];
+	for (int i = 0; i < SCC->size; i++) {
+	    tsortInfo tsi = SCC->members[i];
 	    if (tsi->tsi_SccIdx == 0) /* package already collected */
 		continue;
 	    if (tsi->tsi_SccLowlink >= best_score) {
@@ -544,20 +564,15 @@ int rpmtsOrder(rpmts ts)
     rpmte * newOrder;
     int newOrderCount = 0;
     int rc;
-    rpmal erasedPackages = rpmalCreate(5, rpmtsColor(ts), prefcolor);
+    rpmal erasedPackages;
     scc SCCs;
     int nelem = rpmtsNElements(ts);
     tsortInfo sortInfo = xcalloc(nelem, sizeof(struct tsortInfo_s));
-    ARGV_t seenColls = NULL;
 
     (void) rpmswEnter(rpmtsOp(ts, RPMTS_OP_ORDER), 0);
 
     /* Create erased package index. */
-    pi = rpmtsiInit(ts);
-    while ((p = rpmtsiNext(pi, TR_REMOVED)) != NULL) {
-        rpmalAdd(erasedPackages, p);
-    }
-    pi = rpmtsiFree(pi);
+    erasedPackages = rpmtsCreateAl(ts, TR_REMOVED);
 
     for (int i = 0; i < nelem; i++) {
 	sortInfo[i].te = tsmem->order[i];
@@ -570,24 +585,24 @@ int rpmtsOrder(rpmts ts)
     while ((p = rpmtsiNext(pi, 0)) != NULL) {
 	rpmal al = (rpmteType(p) == TR_REMOVED) ? 
 		   erasedPackages : tsmem->addedPackages;
-	rpmds requires = rpmdsInit(rpmteDS(p, RPMTAG_REQUIRENAME));
-	rpmds order = rpmdsInit(rpmteDS(p, RPMTAG_ORDERNAME));
+	rpmTag ordertags[] = {
+		RPMTAG_REQUIRENAME,
+		RPMTAG_RECOMMENDNAME,
+		RPMTAG_SUGGESTNAME,
+		RPMTAG_SUPPLEMENTNAME,
+		RPMTAG_ENHANCENAME,
+		RPMTAG_ORDERNAME,
+		0,
+	};
 
-	while (rpmdsNext(requires) >= 0) {
-	    /* Record next "q <- p" relation (i.e. "p" requires "q"). */
-	    (void) addRelation(ts, al, p, requires);
+	for (int i = 0; ordertags[i]; i++) {
+	    rpmds dep = rpmdsInit(rpmteDS(p, ordertags[i]));
+	    while (rpmdsNext(dep) >= 0)
+		addRelation(ts, al, p, dep);
 	}
-
-	while (rpmdsNext(order) >= 0) {
-	    /* Record next "q <- p" ordering request */
-	    (void) addRelation(ts, al, p, order);
-	}
-
-	addCollRelations(al, p, &seenColls);
     }
 
-    seenColls = argvFree(seenColls);
-    pi = rpmtsiFree(pi);
+    rpmtsiFree(pi);
 
     newOrder = xcalloc(tsmem->orderCount, sizeof(*newOrder));
     SCCs = detectSCCs(sortInfo, nelem, (rpmtsFlags(ts) & RPMTRANS_FLAG_DEPLOOPS));
