@@ -14,6 +14,7 @@
 #include <rpm/rpmdb.h>
 #include <rpm/rpmfi.h>
 #include <rpm/rpmts.h>
+#include <rpm/rpmsq.h>
 #include <rpm/rpmlog.h>
 #include <rpm/rpmfileutil.h>	/* rpmCleanPath */
 
@@ -30,25 +31,16 @@ static void printFileInfo(const char * name,
 			  unsigned int mtime,
 			  unsigned short rdev, unsigned int nlink,
 			  const char * owner, const char * group,
-			  const char * linkto)
+			  const char * linkto, time_t now)
 {
     char sizefield[21];
     char ownerfield[8+1], groupfield[8+1];
     char timefield[100];
     time_t when = mtime;  /* important if sizeof(int32_t) ! sizeof(time_t) */
-    struct tm * tm;
-    static time_t now;
-    static struct tm nowtm;
+    struct tm * tm, _tm;
     char * perms = rpmPermsString(mode);
     char *link = NULL;
 
-    /* On first call, grab snapshot of now */
-    if (now == 0) {
-	now = time(NULL);
-	tm = localtime(&now);
-	if (tm) nowtm = *tm;	/* structure assignment */
-    }
-
     rstrlcpy(ownerfield, owner, sizeof(ownerfield));
     rstrlcpy(groupfield, group, sizeof(groupfield));
 
@@ -70,7 +62,7 @@ static void printFileInfo(const char * name,
     }
 
     /* Convert file mtime to display format */
-    tm = localtime(&when);
+    tm = localtime_r(&when, &_tm);
     timefield[0] = '\0';
     if (tm != NULL)
     {	const char *fmt;
@@ -91,7 +83,7 @@ static void printFileInfo(const char * name,
 	(void)strftime(timefield, sizeof(timefield) - 1, fmt, tm);
     }
 
-    rpmlog(RPMLOG_NOTICE, "%s %4d %-8s%-8s %10s %s %s\n", perms,
+    rpmlog(RPMLOG_NOTICE, "%s %4d %-8s %-8s %10s %s %s\n", perms,
 	(int)nlink, ownerfield, groupfield, sizefield, timefield, 
 	link ? link : name);
     free(perms);
@@ -103,6 +95,7 @@ int showQueryPackage(QVA_t qva, rpmts ts, Header h)
     rpmfi fi = NULL;
     rpmfiFlags fiflags =  (RPMFI_NOHEADER | RPMFI_FLAGS_QUERY);
     int rc = 0;		/* XXX FIXME: need real return code */
+    time_t now = 0;
 
     if (qva->qva_queryFormat != NULL) {
 	const char *errstr;
@@ -116,6 +109,10 @@ int showQueryPackage(QVA_t qva, rpmts ts, Header h)
 	}
     }
 
+    /* Inclusion flags traditionally imply list mode */
+    if (qva->qva_incattr)
+	qva->qva_flags |= QUERY_FOR_LIST;
+
     if (!(qva->qva_flags & QUERY_FOR_LIST))
 	goto exit;
 
@@ -142,16 +139,12 @@ int showQueryPackage(QVA_t qva, rpmts ts, Header h)
 	const char *flink = rpmfiFLink(fi);
 	char *buf = NULL;
 
-	/* If querying only docs, skip non-doc files. */
-	if ((qva->qva_flags & QUERY_FOR_DOCS) && !(fflags & RPMFILE_DOC))
+	/* If filtering by inclusion, skip non-matching (eg --configfiles) */
+	if (qva->qva_incattr && !(fflags & qva->qva_incattr))
 	    continue;
 
-	/* If querying only configs, skip non-config files. */
-	if ((qva->qva_flags & QUERY_FOR_CONFIG) && !(fflags & RPMFILE_CONFIG))
-	    continue;
-
-	/* If not querying %ghost, skip ghost files. */
-	if ((qva->qva_fflags & RPMFILE_GHOST) && (fflags & RPMFILE_GHOST))
+	/* Skip on attributes (eg from --noghost) */
+	if (fflags & qva->qva_excattr)
 	    continue;
 
 	if (qva->qva_flags & QUERY_FOR_STATE) {
@@ -219,11 +212,14 @@ int showQueryPackage(QVA_t qva, rpmts ts, Header h)
 	    }
 
 	    if (fuser && fgroup) {
+		/* On first call, grab snapshot of now */
+		if (now == 0)
+		    now = time(NULL);
 		if (buf) {
 		    rpmlog(RPMLOG_NOTICE, "%s", buf);
 		}
 		printFileInfo(fn, fsize, fmode, fmtime, frdev, fnlink,
-					fuser, fgroup, flink);
+					fuser, fgroup, flink, now);
 	    } else {
 		rpmlog(RPMLOG_ERR,
 			_("package has neither file owner or id lists\n"));
@@ -235,7 +231,7 @@ int showQueryPackage(QVA_t qva, rpmts ts, Header h)
     rc = 0;
 
 exit:
-    fi = rpmfiFree(fi);
+    rpmfiFree(fi);
     return rc;
 }
 
@@ -262,7 +258,6 @@ void rpmDisplayQueryTags(FILE * fp)
 	}
 	fprintf(fp, "\n");
     }
-    rpmtdFreeData(names);
     rpmtdFree(names);
 }
 
@@ -274,7 +269,7 @@ static int rpmgiShowMatches(QVA_t qva, rpmts ts, rpmgi gi)
     while ((h = rpmgiNext(gi)) != NULL) {
 	int rc;
 
-	rpmdbCheckSignals();
+	rpmsqPoll();
 	if ((rc = qva->qva_showPackage(qva, ts, h)) != 0)
 	    ec = rc;
 	headerFree(h);
@@ -292,21 +287,33 @@ static int rpmcliShowMatches(QVA_t qva, rpmts ts, rpmdbMatchIterator mi)
 
     while ((h = rpmdbNextIterator(mi)) != NULL) {
 	int rc;
-	rpmdbCheckSignals();
+	rpmsqPoll();
 	if ((rc = qva->qva_showPackage(qva, ts, h)) != 0)
 	    ec = rc;
     }
     return ec;
 }
 
+static rpmdbMatchIterator matchesIterator(rpmts ts, rpmDbiTag dbi,
+					const void *arg, size_t arglen)
+{
+    unsigned int matches = 0;
+    rpmdbMatchIterator mi = rpmtsInitIterator(ts, dbi, arg, arglen);
+    while (rpmdbNextIterator(mi) != NULL)
+	matches++;
+    mi = rpmdbFreeIterator(mi);
+    if (matches)
+	mi = rpmtsInitIterator(ts, dbi, arg, arglen);
+    return mi;
+}
+
 static rpmdbMatchIterator initQueryIterator(QVA_t qva, rpmts ts, const char * arg)
 {
     const char * s;
     int i;
-    int provides_checked = 0;
     rpmdbMatchIterator mi = NULL;
 
-    (void) rpmdbCheckSignals();
+    (void) rpmsqPoll();
 
     if (qva->qva_showPackage == NULL)
 	goto exit;
@@ -379,6 +386,20 @@ static rpmdbMatchIterator initQueryIterator(QVA_t qva, rpmts ts, const char * ar
 	}
     }	break;
 
+    case RPMQV_WHATCONFLICTS:
+	mi = rpmtsInitIterator(ts, RPMDBI_CONFLICTNAME, arg, 0);
+	if (mi == NULL) {
+	    rpmlog(RPMLOG_NOTICE, _("no package conflicts %s\n"), arg);
+	}
+	break;
+
+    case RPMQV_WHATOBSOLETES:
+	mi = rpmtsInitIterator(ts, RPMDBI_OBSOLETENAME, arg, 0);
+	if (mi == NULL) {
+	    rpmlog(RPMLOG_NOTICE, _("no package obsoletes %s\n"), arg);
+	}
+	break;
+
     case RPMQV_WHATREQUIRES:
 	mi = rpmtsInitIterator(ts, RPMDBI_REQUIRENAME, arg, 0);
 	if (mi == NULL) {
@@ -386,15 +407,43 @@ static rpmdbMatchIterator initQueryIterator(QVA_t qva, rpmts ts, const char * ar
 	}
 	break;
 
+    case RPMQV_WHATRECOMMENDS:
+	mi = rpmtsInitIterator(ts, RPMDBI_RECOMMENDNAME, arg, 0);
+	if (mi == NULL) {
+	    rpmlog(RPMLOG_NOTICE, _("no package recommends %s\n"), arg);
+	}
+	break;
+
+    case RPMQV_WHATSUGGESTS:
+	mi = rpmtsInitIterator(ts, RPMDBI_SUGGESTNAME, arg, 0);
+	if (mi == NULL) {
+	    rpmlog(RPMLOG_NOTICE, _("no package suggests %s\n"), arg);
+	}
+	break;
+
+    case RPMQV_WHATSUPPLEMENTS:
+	mi = rpmtsInitIterator(ts, RPMDBI_SUPPLEMENTNAME, arg, 0);
+	if (mi == NULL) {
+	    rpmlog(RPMLOG_NOTICE, _("no package supplements %s\n"), arg);
+	}
+	break;
+
+    case RPMQV_WHATENHANCES:
+	mi = rpmtsInitIterator(ts, RPMDBI_ENHANCENAME, arg, 0);
+	if (mi == NULL) {
+	    rpmlog(RPMLOG_NOTICE, _("no package enhances %s\n"), arg);
+	}
+	break;
+
     case RPMQV_WHATPROVIDES:
-	if (arg[0] != '/') {
-	    provides_checked = 1;
+	if (arg[0] != '/' && arg[0] != '.') {
 	    mi = rpmtsInitIterator(ts, RPMDBI_PROVIDENAME, arg, 0);
 	    if (mi == NULL) {
 		rpmlog(RPMLOG_NOTICE, _("no package provides %s\n"), arg);
 	    }
 	    break;
 	}
+	/* fallthrough on absolute and relative paths */
     case RPMQV_PATH:
     {   char * fn;
 
@@ -409,13 +458,14 @@ static rpmdbMatchIterator initQueryIterator(QVA_t qva, rpmts ts, const char * ar
 	} else if (*arg != '/') {
 	    char *curDir = rpmGetCwd();
 	    fn = (char *) rpmGetPath(curDir, "/", arg, NULL);
-	    curDir = _free(curDir);
+	    free(curDir);
 	} else
 	    fn = xstrdup(arg);
 	(void) rpmCleanPath(fn);
 
-	mi = rpmtsInitIterator(ts, RPMDBI_BASENAMES, fn, 0);
-	if (mi == NULL && !provides_checked)
+	/* XXX Add a switch to enable former BASENAMES behavior? */
+	mi = rpmtsInitIterator(ts, RPMDBI_INSTFILENAMES, fn, 0);
+	if (mi == NULL)
 	    mi = rpmtsInitIterator(ts, RPMDBI_PROVIDENAME, fn, 0);
 
 	if (mi == NULL) {
@@ -427,7 +477,7 @@ static rpmdbMatchIterator initQueryIterator(QVA_t qva, rpmts ts, const char * ar
 			_("file %s is not owned by any package\n"), fn);
 	}
 
-	fn = _free(fn);
+	free(fn);
     }	break;
 
     case RPMQV_DBOFFSET:
@@ -440,7 +490,7 @@ static rpmdbMatchIterator initQueryIterator(QVA_t qva, rpmts ts, const char * ar
 	}
 	rpmlog(RPMLOG_DEBUG, "package record number: %u\n", recOffset);
 	/* RPMDBI_PACKAGES */
-	mi = rpmtsInitIterator(ts, RPMDBI_PACKAGES, &recOffset, sizeof(recOffset));
+	mi = matchesIterator(ts, RPMDBI_PACKAGES, &recOffset, sizeof(recOffset));
 	if (mi == NULL) {
 	    rpmlog(RPMLOG_ERR, _("record %u could not be read\n"), recOffset);
 	}
@@ -448,17 +498,10 @@ static rpmdbMatchIterator initQueryIterator(QVA_t qva, rpmts ts, const char * ar
 
     case RPMQV_PACKAGE:
     {
-	int matches = 0;
-	mi = rpmtsInitIterator(ts, RPMDBI_LABEL, arg, 0);
-	while (rpmdbNextIterator(mi) != NULL) {
-	    matches++;
-	}
-	mi = rpmdbFreeIterator(mi);
-	if (! matches) {
+	mi = matchesIterator(ts, RPMDBI_LABEL, arg, 0);
+
+	if (mi == NULL && !rpmFileHasSuffix(arg, ".rpm"))
 	    rpmlog(RPMLOG_NOTICE, _("package %s is not installed\n"), arg);
-	} else {
-	    mi = rpmtsInitIterator(ts, RPMDBI_LABEL, arg, 0);
-	}
 	break;
     }
     default:
@@ -510,26 +553,42 @@ int rpmcliArgIter(rpmts ts, QVA_t qva, ARGV_const_t argv)
     case RPMQV_ALL: {
 	rpmdbMatchIterator mi = initFilterIterator(ts, argv);
 	ec = rpmcliShowMatches(qva, ts, mi);
-	mi = rpmdbFreeIterator(mi);
+	rpmdbFreeIterator(mi);
 	break;
     }
     case RPMQV_RPM: {
 	rpmgi gi = rpmgiNew(ts, giFlags, argv);
 	ec = rpmgiShowMatches(qva, ts, gi);
-	gi = rpmgiFree(gi);
+	rpmgiFree(gi);
 	break;
     }
     case RPMQV_SPECRPMS:
-    case RPMQV_SPECSRPM:
+    case RPMQV_SPECBUILTRPMS:
+    case RPMQV_SPECSRPM: {
+	char *target = rpmExpand("%{_target}", NULL);
 	for (ARGV_const_t arg = argv; arg && *arg; arg++) {
 	    ec += ((qva->qva_specQuery != NULL)
 		    ? qva->qva_specQuery(ts, qva, *arg) : 1);
+	    rpmFreeMacros(NULL);
+	    rpmReadConfigFiles(rpmcliRcfile, target);
 	}
+	free(target);
 	break;
+    }
     default:
 	for (ARGV_const_t arg = argv; arg && *arg; arg++) {
+	    int ecLocal;
 	    rpmdbMatchIterator mi = initQueryIterator(qva, ts, *arg);
-	    ec += rpmcliShowMatches(qva, ts, mi);
+	    ecLocal = rpmcliShowMatches(qva, ts, mi);
+	    if (mi == NULL && qva->qva_source == RPMQV_PACKAGE) {
+		if (rpmFileHasSuffix(*arg, ".rpm")) {
+		    char * const argFirst[2] = { arg[0], NULL };
+		    rpmgi gi = rpmgiNew(ts, giFlags, argFirst);
+		    ecLocal = rpmgiShowMatches(qva, ts, gi);
+		    rpmgiFree(gi);
+		}
+	    }
+	    ec += ecLocal;
 	    rpmdbFreeIterator(mi);
 	}
 	break;
@@ -547,26 +606,22 @@ int rpmcliQuery(rpmts ts, QVA_t qva, char * const * argv)
 	qva->qva_showPackage = showQueryPackage;
 
     /* If --queryformat unspecified, then set default now. */
-    if (!(qva->qva_flags & _QUERY_FOR_BITS) && qva->qva_queryFormat == NULL) {
+    if (!(qva->qva_flags & _QUERY_FOR_BITS) && !(qva->qva_incattr) &&
+					    qva->qva_queryFormat == NULL) {
 	char * fmt = rpmExpand("%{?_query_all_fmt}\n", NULL);
 	if (fmt == NULL || strlen(fmt) <= 1) {
-	    fmt = _free(fmt);
+	    free(fmt);
 	    fmt = xstrdup("%{nvra}\n");
 	}
 	qva->qva_queryFormat = fmt;
     }
 
     vsflags = rpmExpandNumeric("%{?_vsflags_query}");
-    if (rpmcliQueryFlags & VERIFY_DIGEST)
-	vsflags |= _RPMVSF_NODIGESTS;
-    if (rpmcliQueryFlags & VERIFY_SIGNATURE)
-	vsflags |= _RPMVSF_NOSIGNATURES;
-    if (rpmcliQueryFlags & VERIFY_HDRCHK)
-	vsflags |= RPMVSF_NOHDRCHK;
+    vsflags |= rpmcliVSFlags;
 
     ovsflags = rpmtsSetVSFlags(ts, vsflags);
     ec = rpmcliArgIter(ts, qva, argv);
-    vsflags = rpmtsSetVSFlags(ts, ovsflags);
+    rpmtsSetVSFlags(ts, ovsflags);
 
     if (qva->qva_showPackage == showQueryPackage)
 	qva->qva_showPackage = NULL;
