@@ -10,6 +10,7 @@
 #include <rpm/rpmdb.h>
 #include <rpm/rpmds.h>
 #include <rpm/rpmts.h>
+#include <rpm/rpmsq.h>
 #include <rpm/rpmlog.h>
 #include <rpm/rpmfileutil.h>
 
@@ -17,11 +18,12 @@
 #include "lib/manifest.h"
 #include "debug.h"
 
-int rpmcliPackagesTotal = 0;
-int rpmcliHashesCurrent = 0;
-int rpmcliHashesTotal = 0;
-int rpmcliProgressCurrent = 0;
-int rpmcliProgressTotal = 0;
+static int rpmcliPackagesTotal = 0;
+static int rpmcliHashesCurrent = 0;
+static int rpmcliHashesTotal = 0;
+static int rpmcliProgressCurrent = 0;
+static int rpmcliProgressTotal = 0;
+static int rpmcliProgressState = 0;
 
 /**
  * Print a CLI progress bar.
@@ -33,7 +35,7 @@ static void printHash(const rpm_loff_t amount, const rpm_loff_t total)
 {
     int hashesNeeded;
 
-    rpmcliHashesTotal = (isatty (STDOUT_FILENO) ? 44 : 50);
+    rpmcliHashesTotal = (isatty (STDOUT_FILENO) ? 34 : 40);
 
     if (rpmcliHashesCurrent != rpmcliHashesTotal) {
 	float pct = (total ? (((float) amount) / total) : 1.0);
@@ -81,12 +83,7 @@ static rpmVSFlags setvsFlags(struct rpmInstallArguments_s * ia)
     else
 	vsflags = rpmExpandNumeric("%{?_vsflags_install}");
 
-    if (rpmcliQueryFlags & VERIFY_DIGEST)
-	vsflags |= _RPMVSF_NODIGESTS;
-    if (rpmcliQueryFlags & VERIFY_SIGNATURE)
-	vsflags |= _RPMVSF_NOSIGNATURES;
-    if (rpmcliQueryFlags & VERIFY_HDRCHK)
-	vsflags |= RPMVSF_NOHDRCHK;
+    vsflags |= rpmcliVSFlags;
 
     return vsflags;
 }
@@ -99,7 +96,6 @@ void * rpmShowProgress(const void * arg,
 			void * data)
 {
     Header h = (Header) arg;
-    char * s;
     int flags = (int) ((long)data);
     void * rc = NULL;
     const char * filename = (const char *)key;
@@ -133,28 +129,45 @@ void * rpmShowProgress(const void * arg,
 	break;
 
     case RPMCALLBACK_INST_START:
+    case RPMCALLBACK_UNINST_START:
+	if (rpmcliProgressState != what) {
+	    rpmcliProgressState = what;
+	    if (flags & INSTALL_HASH) {
+		if (what == RPMCALLBACK_INST_START) {
+		    fprintf(stdout, _("Updating / installing...\n"));
+		} else {
+		    fprintf(stdout, _("Cleaning up / removing...\n"));
+		}
+		fflush(stdout);
+	    }
+	}
+		
 	rpmcliHashesCurrent = 0;
 	if (h == NULL || !(flags & INSTALL_LABEL))
 	    break;
-	/* @todo Remove headerFormat() on a progress callback. */
 	if (flags & INSTALL_HASH) {
-	    s = headerFormat(h, "%{NAME}", NULL);
+	    char *s = headerGetAsString(h, RPMTAG_NEVR);
 	    if (isatty (STDOUT_FILENO))
-		fprintf(stdout, "%4d:%-23.23s", rpmcliProgressCurrent + 1, s);
+		fprintf(stdout, "%4d:%-33.33s", rpmcliProgressCurrent + 1, s);
 	    else
-		fprintf(stdout, "%-28.28s", s);
+		fprintf(stdout, "%-38.38s", s);
 	    (void) fflush(stdout);
-	    s = _free(s);
+	    free(s);
 	} else {
-	    s = headerFormat(h, "%{NAME}-%{VERSION}-%{RELEASE}", NULL);
+	    char *s = headerGetAsString(h, RPMTAG_NEVRA);
 	    fprintf(stdout, "%s\n", s);
 	    (void) fflush(stdout);
-	    s = _free(s);
+	    free(s);
 	}
 	break;
 
+    case RPMCALLBACK_INST_STOP:
+	break;
+
     case RPMCALLBACK_TRANS_PROGRESS:
     case RPMCALLBACK_INST_PROGRESS:
+    case RPMCALLBACK_UNINST_PROGRESS:
+    case RPMCALLBACK_VERIFY_PROGRESS:
 	if (flags & INSTALL_PERCENT)
 	    fprintf(stdout, "%%%% %f\n", (double) (total
 				? ((((float) amount) / total) * 100)
@@ -165,29 +178,32 @@ void * rpmShowProgress(const void * arg,
 	break;
 
     case RPMCALLBACK_TRANS_START:
+    case RPMCALLBACK_VERIFY_START:
 	rpmcliHashesCurrent = 0;
 	rpmcliProgressTotal = 1;
 	rpmcliProgressCurrent = 0;
+	rpmcliPackagesTotal = total;
+	rpmcliProgressState = what;
 	if (!(flags & INSTALL_LABEL))
 	    break;
-	if (flags & INSTALL_HASH)
-	    fprintf(stdout, "%-28s", _("Preparing..."));
-	else
-	    fprintf(stdout, "%s\n", _("Preparing packages for installation..."));
+	if (flags & INSTALL_HASH) {
+	    fprintf(stdout, "%-38s", (what == RPMCALLBACK_TRANS_START) ?
+		    _("Preparing...") : _("Verifying..."));
+	} else {
+	    fprintf(stdout, "%s\n", (what == RPMCALLBACK_TRANS_START) ?
+		    _("Preparing packages...") : _("Verifying packages..."));
+	}
 	(void) fflush(stdout);
 	break;
 
     case RPMCALLBACK_TRANS_STOP:
+    case RPMCALLBACK_VERIFY_STOP:
 	if (flags & INSTALL_HASH)
 	    printHash(1, 1);	/* Fixes "preparing..." progress bar */
 	rpmcliProgressTotal = rpmcliPackagesTotal;
 	rpmcliProgressCurrent = 0;
 	break;
 
-    case RPMCALLBACK_UNINST_PROGRESS:
-	break;
-    case RPMCALLBACK_UNINST_START:
-	break;
     case RPMCALLBACK_UNINST_STOP:
 	break;
     case RPMCALLBACK_UNPACK_ERROR:
@@ -196,6 +212,10 @@ void * rpmShowProgress(const void * arg,
 	break;
     case RPMCALLBACK_SCRIPT_ERROR:
 	break;
+    case RPMCALLBACK_SCRIPT_START:
+	break;
+    case RPMCALLBACK_SCRIPT_STOP:
+	break;
     case RPMCALLBACK_UNKNOWN:
     default:
 	break;
@@ -214,7 +234,6 @@ static void setNotifyFlag(struct rpmInstallArguments_s * ia,
 }
 
 struct rpmEIU {
-    Header h;
     int numFailed;
     int numPkgs;
     char ** pkgURL;
@@ -225,7 +244,6 @@ struct rpmEIU {
     int numRPMS;
     int numSRPMS;
     char ** sourceURL;
-    int isSource;
     int argc;
     char ** argv;
     rpmRelocation * relocations;
@@ -274,8 +292,8 @@ static int rpmcliTransaction(rpmts ts, struct rpmInstallArguments_s * ia,
 
 	ps = rpmtsProblems(ts);
 
-	if ((rpmpsNumProblems(ps) > 0) && (eflags? 1 : (rc > 0)))
-	    rpmpsPrint((eflags? NULL : stderr), ps);
+	if (rpmpsNumProblems(ps) > 0 && (eflags || rc > 0))
+	    rpmpsPrint(NULL, ps);
 	ps = rpmpsFree(ps);
     }
 
@@ -287,7 +305,7 @@ static int tryReadManifest(struct rpmEIU * eiu)
     int rc;
 
     /* Try to read a package manifest. */
-    FD_t fd = Fopen(*eiu->fnp, "r.fpio");
+    FD_t fd = Fopen(*eiu->fnp, "r.ufdio");
     if (fd == NULL || Ferror(fd)) {
         rpmlog(RPMLOG_ERR, _("open of %s failed: %s\n"), *eiu->fnp,
 	       Fstrerror(fd));
@@ -295,7 +313,8 @@ static int tryReadManifest(struct rpmEIU * eiu)
 	    Fclose(fd);
 	    fd = NULL;
 	}
-	eiu->numFailed++; *eiu->fnp = NULL;
+	eiu->numFailed++;
+	*eiu->fnp = NULL;
 	return RPMRC_FAIL;
     }
 
@@ -307,16 +326,16 @@ static int tryReadManifest(struct rpmEIU * eiu)
     Fclose(fd);
     fd = NULL;
 
-    if (rc != RPMRC_OK)
-        eiu->numFailed++; *eiu->fnp = NULL;
+    if (rc != RPMRC_OK) {
+        eiu->numFailed++;
+        *eiu->fnp = NULL;
+    }
 
     return rc;
 }
 
-static int tryReadHeader(rpmts ts, struct rpmEIU * eiu, rpmVSFlags vsflags)
+static int tryReadHeader(rpmts ts, struct rpmEIU * eiu, Header * hdrp)
 {
-   rpmVSFlags tvsflags;
-
    /* Try to read the header from a package file. */
    FD_t fd = Fopen(*eiu->fnp, "r.ufdio");
    if (fd == NULL || Ferror(fd)) {
@@ -326,14 +345,13 @@ static int tryReadHeader(rpmts ts, struct rpmEIU * eiu, rpmVSFlags vsflags)
            Fclose(fd);
 	   fd = NULL;
        }
-       eiu->numFailed++; *eiu->fnp = NULL;
+       eiu->numFailed++;
+       *eiu->fnp = NULL;
        return RPMRC_FAIL;
    }
 
    /* Read the header, verifying signatures (if present). */
-   tvsflags = rpmtsSetVSFlags(ts, vsflags);
-   eiu->rpmrc = rpmReadPackageFile(ts, fd, *eiu->fnp, &eiu->h);
-   tvsflags = rpmtsSetVSFlags(ts, tvsflags);
+   eiu->rpmrc = rpmReadPackageFile(ts, fd, *eiu->fnp, hdrp);
    Fclose(fd);
    fd = NULL;
    
@@ -341,9 +359,10 @@ static int tryReadHeader(rpmts ts, struct rpmEIU * eiu, rpmVSFlags vsflags)
    if (eiu->rpmrc == RPMRC_NOTFOUND && (giFlags & RPMGI_NOMANIFEST))
        eiu->rpmrc = RPMRC_FAIL;
 
-   if(eiu->rpmrc == RPMRC_FAIL) {
+   if (eiu->rpmrc == RPMRC_FAIL) {
        rpmlog(RPMLOG_ERR, _("%s cannot be installed\n"), *eiu->fnp);
-       eiu->numFailed++; *eiu->fnp = NULL;
+       eiu->numFailed++;
+       *eiu->fnp = NULL;
    }
 
    return RPMRC_OK;
@@ -351,11 +370,11 @@ static int tryReadHeader(rpmts ts, struct rpmEIU * eiu, rpmVSFlags vsflags)
 
 
 /* On --freshen, verify package is installed and newer */
-static int checkFreshenStatus(rpmts ts, struct rpmEIU * eiu)
+static int checkFreshenStatus(rpmts ts, Header h)
 {
     rpmdbMatchIterator mi = NULL;
-    const char * name = headerGetString(eiu->h, RPMTAG_NAME);
-    const char *arch = headerGetString(eiu->h, RPMTAG_ARCH);
+    const char * name = headerGetString(h, RPMTAG_NAME);
+    const char *arch = headerGetString(h, RPMTAG_ARCH);
     Header oldH = NULL;
 
     if (name != NULL)
@@ -365,15 +384,25 @@ static int checkFreshenStatus(rpmts ts, struct rpmEIU * eiu)
 
     while ((oldH = rpmdbNextIterator(mi)) != NULL) {
 	/* Package is newer than those currently installed. */
-        if (rpmVersionCompare(oldH, eiu->h) < 0)
+        if (rpmVersionCompare(oldH, h) < 0)
 	    break;
     }
 
-    mi = rpmdbFreeIterator(mi);
-    if (oldH == NULL) {
-        eiu->h = headerFree(eiu->h);
+    rpmdbFreeIterator(mi);
+    return (oldH != NULL);
+}
+
+static int rpmNoGlob(const char *fn, int *argcPtr, ARGV_t * argvPtr)
+{
+    struct stat sb;
+    int rc = stat(fn, &sb);
+    if (rc == 0) {
+	argvAdd(argvPtr, fn);
+	*argcPtr = 1;
+    } else {
+	*argcPtr = 0;
     }
-   return (oldH != NULL);
+    return rc;
 }
 
 /** @todo Generalize --freshen policies. */
@@ -383,20 +412,27 @@ int rpmInstall(rpmts ts, struct rpmInstallArguments_s * ia, ARGV_t fileArgv)
     rpmRelocation * relocations;
     char * fileURL = NULL;
     rpmVSFlags vsflags, ovsflags;
+    rpmVSFlags ovfyflags;
     int rc;
     int i;
 
+    vsflags = setvsFlags(ia);
+    ovsflags = rpmtsSetVSFlags(ts, (vsflags | RPMVSF_NEEDPAYLOAD));
+    /* for rpm cli, --nosignature/--nodigest applies to both vs and vfyflags */
+    ovfyflags = rpmtsSetVfyFlags(ts, (rpmtsVfyFlags(ts) | rpmcliVSFlags));
+
     if (fileArgv == NULL) goto exit;
 
-    rpmcliPackagesTotal = 0;
+    if (rpmcliVfyLevelMask) {
+	int vfylevel = rpmtsVfyLevel(ts);
+	vfylevel &= ~rpmcliVfyLevelMask;
+	rpmtsSetVfyLevel(ts, vfylevel);
+    }
 
     (void) rpmtsSetFlags(ts, ia->transFlags);
 
     relocations = ia->relocations;
 
-    vsflags = setvsFlags(ia);
-    ovsflags = rpmtsSetVSFlags(ts, (vsflags | RPMVSF_NEEDPAYLOAD));
-
     setNotifyFlag(ia, ts); 
 
     if ((eiu->relocations = relocations) != NULL) {
@@ -410,13 +446,20 @@ int rpmInstall(rpmts ts, struct rpmInstallArguments_s * ia, ARGV_t fileArgv)
     for (eiu->fnp = fileArgv; *eiu->fnp != NULL; eiu->fnp++) {
     	ARGV_t av = NULL;
     	int ac = 0;
-	char * fn;
 
-	fn = rpmEscapeSpaces(*eiu->fnp);
-	rc = rpmGlob(fn, &ac, &av);
-	fn = _free(fn);
+	if (giFlags & RPMGI_NOGLOB) {
+	    rc = rpmNoGlob(*eiu->fnp, &ac, &av);
+	} else {
+	    char * fn = rpmEscapeSpaces(*eiu->fnp);
+	    rc = rpmGlob(fn, &ac, &av);
+	    fn = _free(fn);
+	}
 	if (rc || ac == 0) {
-	    rpmlog(RPMLOG_ERR, _("File not found by glob: %s\n"), *eiu->fnp);
+	    if (giFlags & RPMGI_NOGLOB) {
+		rpmlog(RPMLOG_ERR, _("File not found: %s\n"), *eiu->fnp);
+	    } else {
+		rpmlog(RPMLOG_ERR, _("File not found by glob: %s\n"), *eiu->fnp);
+	    }
 	    eiu->numFailed++;
 	    continue;
 	}
@@ -495,12 +538,13 @@ restart:
 	 *eiu->fnp != NULL;
 	 eiu->fnp++, eiu->prevx++)
     {
+	Header h = NULL;
 	const char * fileName;
 
 	rpmlog(RPMLOG_DEBUG, "============== %s\n", *eiu->fnp);
 	(void) urlPath(*eiu->fnp, &fileName);
 
-	if (tryReadHeader(ts, eiu, vsflags) == RPMRC_FAIL)
+	if (tryReadHeader(ts, eiu, &h) == RPMRC_FAIL)
 	    continue;
 
 	if (eiu->rpmrc == RPMRC_NOTFOUND) {
@@ -511,9 +555,11 @@ restart:
 	    }
 	}
 
-	eiu->isSource = headerIsSource(eiu->h);
-
-	if (eiu->isSource) {
+	if (headerIsSource(h)) {
+	    if (ia->installInterfaceFlags & INSTALL_FRESHEN) {
+		headerFree(h);
+	        continue;
+	    }
 	    rpmlog(RPMLOG_DEBUG, "\tadded source package [%d]\n",
 		eiu->numSRPMS);
 	    eiu->sourceURL = xrealloc(eiu->sourceURL,
@@ -528,32 +574,36 @@ restart:
 	if (eiu->relocations) {
 	    struct rpmtd_s prefixes;
 
-	    headerGet(eiu->h, RPMTAG_PREFIXES, &prefixes, HEADERGET_DEFAULT);
+	    headerGet(h, RPMTAG_PREFIXES, &prefixes, HEADERGET_DEFAULT);
 	    if (rpmtdCount(&prefixes) == 1) {
 		eiu->relocations->oldPath = xstrdup(rpmtdGetString(&prefixes));
 		rpmtdFreeData(&prefixes);
 	    } else {
 		rpmlog(RPMLOG_ERR, _("package %s is not relocatable\n"),
-		       headerGetString(eiu->h, RPMTAG_NAME));
+		       headerGetString(h, RPMTAG_NAME));
 		eiu->numFailed++;
 		goto exit;
 	    }
 	}
 
 	if (ia->installInterfaceFlags & INSTALL_FRESHEN)
-	    if (checkFreshenStatus(ts, eiu) != 1)
+	    if (checkFreshenStatus(ts, h) != 1) {
+		headerFree(h);
 	        continue;
+	    }
 
-	rc = rpmtsAddInstallElement(ts, eiu->h, (fnpyKey)fileName,
+	if (ia->installInterfaceFlags & INSTALL_REINSTALL)
+	    rc = rpmtsAddReinstallElement(ts, h, (fnpyKey)fileName);
+	else
+	    rc = rpmtsAddInstallElement(ts, h, (fnpyKey)fileName,
 			(ia->installInterfaceFlags & INSTALL_UPGRADE) != 0,
 			relocations);
 
-	/* XXX reference held by transaction set */
-	eiu->h = headerFree(eiu->h);
+	headerFree(h);
 	if (eiu->relocations)
 	    eiu->relocations->oldPath = _free(eiu->relocations->oldPath);
 
-	switch(rc) {
+	switch (rc) {
 	case 0:
 	    rpmlog(RPMLOG_DEBUG, "\tadded binary package [%d]\n",
 			eiu->numRPMS);
@@ -564,13 +614,6 @@ restart:
 	    eiu->numFailed++;
 	    goto exit;
 	    break;
-	case 2:
-	    rpmlog(RPMLOG_ERR,
-			    _("file %s requires a newer version of RPM\n"),
-			    *eiu->fnp);
-	    eiu->numFailed++;
-	    goto exit;
-	    break;
 	default:
 	    eiu->numFailed++;
 	    goto exit;
@@ -594,8 +637,11 @@ restart:
     }
 
     if (eiu->numSRPMS && (eiu->sourceURL != NULL)) {
+	rpmcliProgressState = 0;
+	rpmcliProgressTotal = 0;
+	rpmcliProgressCurrent = 0;
 	for (i = 0; i < eiu->numSRPMS; i++) {
-	    rpmdbCheckSignals();
+	    rpmsqPoll();
 	    if (eiu->sourceURL[i] != NULL) {
 	        rc = RPMRC_OK;
 		if (!(rpmtsFlags(ts) & RPMTRANS_FLAG_TEST))
@@ -622,6 +668,8 @@ exit:
     free(eiu);
 
     rpmtsEmpty(ts);
+    rpmtsSetVSFlags(ts, ovsflags);
+    rpmtsSetVfyFlags(ts, ovfyflags);
 
     return rc;
 }
@@ -641,23 +689,14 @@ int rpmErase(rpmts ts, struct rpmInstallArguments_s * ia, ARGV_const_t argv)
 
     (void) rpmtsSetFlags(ts, ia->transFlags);
 
-#ifdef	NOTYET	/* XXX no callbacks on erase yet */
     setNotifyFlag(ia, ts);
-#endif
 
     qfmt = rpmExpand("%{?_query_all_fmt}\n", NULL);
     for (arg = argv; *arg; arg++) {
-	rpmdbMatchIterator mi;
-	int matches = 0;
+	rpmdbMatchIterator mi = rpmtsInitIterator(ts, RPMDBI_LABEL, *arg, 0);
+	int matches = rpmdbGetIteratorCount(mi);
 	int erasing = 1;
 
-	/* Iterator count isn't reliable with labels, count manually... */
-	mi = rpmtsInitIterator(ts, RPMDBI_LABEL, *arg, 0);
-	while (rpmdbNextIterator(mi) != NULL) {
-	    matches++;
-	}
-	rpmdbFreeIterator(mi);
-
 	if (! matches) {
 	    rpmlog(RPMLOG_ERR, _("package %s is not installed\n"), *arg);
 	    numFailed++;
@@ -672,7 +711,6 @@ int rpmErase(rpmts ts, struct rpmInstallArguments_s * ia, ARGV_const_t argv)
 		erasing = 0;
 	    }
 
-	    mi = rpmtsInitIterator(ts, RPMDBI_LABEL, *arg, 0);
 	    while ((h = rpmdbNextIterator(mi)) != NULL) {
 		if (erasing) {
 		    (void) rpmtsAddEraseElement(ts, h, -1);
@@ -683,8 +721,8 @@ int rpmErase(rpmts ts, struct rpmInstallArguments_s * ia, ARGV_const_t argv)
 		    free(nevra);
 		}
 	    }
-	    mi = rpmdbFreeIterator(mi);
 	}
+	rpmdbFreeIterator(mi);
     }
     free(qfmt);
 
@@ -692,8 +730,9 @@ int rpmErase(rpmts ts, struct rpmInstallArguments_s * ia, ARGV_const_t argv)
     numFailed = rpmcliTransaction(ts, ia, numPackages);
 exit:
     rpmtsEmpty(ts);
+    rpmtsSetVSFlags(ts, ovsflags);
 
-    return numFailed;
+    return (numFailed < 0) ? numPackages : numFailed;
 }
 
 int rpmInstallSource(rpmts ts, const char * arg,
@@ -718,7 +757,7 @@ int rpmInstallSource(rpmts ts, const char * arg,
 		rpmtsSetVSFlags(ts, (specFilePtr) ? (rpmtsVSFlags(ts) | RPMVSF_NEEDPAYLOAD) : rpmtsVSFlags(ts));
 	rpmRC rpmrc = rpmInstallSourcePackage(ts, fd, specFilePtr, cookie);
 	rc = (rpmrc == RPMRC_OK ? 0 : 1);
-	ovsflags = rpmtsSetVSFlags(ts, ovsflags);
+	rpmtsSetVSFlags(ts, ovsflags);
     }
     if (rc != 0) {
 	rpmlog(RPMLOG_ERR, _("%s cannot be installed\n"), arg);
