@@ -1,6 +1,12 @@
 #ifndef _RPMTAG_H
 #define _RPMTAG_H
 
+/** \ingroup rpmtag
+ *  \file lib/rpmtag.h
+ *
+ * Accessing RPM tags: values, types, ...
+ */
+
 #include <rpm/rpmtypes.h>
 
 #ifdef __cplusplus
@@ -57,6 +63,10 @@ typedef enum rpmTag_e {
 #define	RPMTAG_HDRID	RPMTAG_SHA1HEADER	/* s */
     RPMTAG_LONGSIGSIZE		= RPMTAG_SIG_BASE+14,	/* l */
     RPMTAG_LONGARCHIVESIZE	= RPMTAG_SIG_BASE+15,	/* l */
+    /* RPMTAG_SIG_BASE+16 reserved */
+    RPMTAG_SHA256HEADER		= RPMTAG_SIG_BASE+17,	/* s */
+    /* RPMTAG_SIG_BASE+18 reserved for RPMSIGTAG_FILESIGNATURES */
+    /* RPMTAG_SIG_BASE+19 reserved for RPMSIGTAG_FILESIGNATURELENGTH */
 
     RPMTAG_NAME  		= 1000,	/* s */
 #define	RPMTAG_N	RPMTAG_NAME	/* s */
@@ -314,8 +324,8 @@ typedef enum rpmTag_e {
     RPMTAG_RECOMMENDFLAGS	= 5048,	/* i[] */
     RPMTAG_SUGGESTNAME		= 5049,	/* s[] */
 #define	RPMTAG_SUGGESTS RPMTAG_SUGGESTNAME	/* s[] */
-    RPMTAG_SUGGESTVERSION	= 5050,	/* s[] extension */
-    RPMTAG_SUGGESTFLAGS		= 5051,	/* i[] extension */
+    RPMTAG_SUGGESTVERSION	= 5050,	/* s[] */
+    RPMTAG_SUGGESTFLAGS		= 5051,	/* i[] */
     RPMTAG_SUPPLEMENTNAME	= 5052,	/* s[] */
 #define	RPMTAG_SUPPLEMENTS RPMTAG_SUPPLEMENTNAME /* s[] */
     RPMTAG_SUPPLEMENTVERSION	= 5053,	/* s[] */
@@ -328,6 +338,42 @@ typedef enum rpmTag_e {
     RPMTAG_SUGGESTNEVRS		= 5059, /* s[] extension */
     RPMTAG_SUPPLEMENTNEVRS	= 5060, /* s[] extension */
     RPMTAG_ENHANCENEVRS		= 5061, /* s[] extension */
+    RPMTAG_ENCODING		= 5062, /* s */
+    RPMTAG_FILETRIGGERIN		= 5063, /* internal */
+    RPMTAG_FILETRIGGERUN		= 5064, /* internal */
+    RPMTAG_FILETRIGGERPOSTUN		= 5065, /* internal */
+    RPMTAG_FILETRIGGERSCRIPTS		= 5066, /* s[] */
+    RPMTAG_FILETRIGGERSCRIPTPROG	= 5067, /* s[] */
+    RPMTAG_FILETRIGGERSCRIPTFLAGS	= 5068, /* i[] */
+    RPMTAG_FILETRIGGERNAME		= 5069, /* s[] */
+    RPMTAG_FILETRIGGERINDEX		= 5070, /* i[] */
+    RPMTAG_FILETRIGGERVERSION		= 5071, /* s[] */
+    RPMTAG_FILETRIGGERFLAGS		= 5072, /* i[] */
+    RPMTAG_TRANSFILETRIGGERIN		= 5073, /* internal */
+    RPMTAG_TRANSFILETRIGGERUN		= 5074, /* internal */
+    RPMTAG_TRANSFILETRIGGERPOSTUN	= 5075, /* internal */
+    RPMTAG_TRANSFILETRIGGERSCRIPTS	= 5076, /* s[] */
+    RPMTAG_TRANSFILETRIGGERSCRIPTPROG	= 5077, /* s[] */
+    RPMTAG_TRANSFILETRIGGERSCRIPTFLAGS	= 5078, /* i[] */
+    RPMTAG_TRANSFILETRIGGERNAME		= 5079, /* s[] */
+    RPMTAG_TRANSFILETRIGGERINDEX	= 5080, /* i[] */
+    RPMTAG_TRANSFILETRIGGERVERSION	= 5081, /* s[] */
+    RPMTAG_TRANSFILETRIGGERFLAGS	= 5082, /* i[] */
+    RPMTAG_REMOVEPATHPOSTFIXES  = 5083, /* s internal */
+    RPMTAG_FILETRIGGERPRIORITIES	= 5084, /* i[] */
+    RPMTAG_TRANSFILETRIGGERPRIORITIES	= 5085, /* i[] */
+    RPMTAG_FILETRIGGERCONDS		= 5086, /* s[] extension */
+    RPMTAG_FILETRIGGERTYPE		= 5087, /* s[] extension */
+    RPMTAG_TRANSFILETRIGGERCONDS	= 5088, /* s[] extension */
+    RPMTAG_TRANSFILETRIGGERTYPE		= 5089, /* s[] extension */
+    RPMTAG_FILESIGNATURES	= 5090, /* s[] */
+    RPMTAG_FILESIGNATURELENGTH  = 5091, /* i */
+    RPMTAG_PAYLOADDIGEST	= 5092, /* s[] */
+    RPMTAG_PAYLOADDIGESTALGO	= 5093, /* i */
+    RPMTAG_AUTOINSTALLED	= 5094, /* i reservation (unimplemented) */
+    RPMTAG_IDENTITY		= 5095, /* s reservation (unimplemented) */
+    RPMTAG_MODULARITYLABEL	= 5096, /* s */
+    RPMTAG_PAYLOADDIGESTALT	= 5097, /* s[] */
 
     RPMTAG_FIRSTFREE_TAG	/*!< internal */
 } rpmTag;
@@ -353,6 +399,12 @@ typedef enum rpmDbiTag_e {
     RPMDBI_SIGMD5		= RPMTAG_SIGMD5,
     RPMDBI_SHA1HEADER		= RPMTAG_SHA1HEADER,
     RPMDBI_INSTFILENAMES	= RPMTAG_INSTFILENAMES,
+    RPMDBI_FILETRIGGERNAME	= RPMTAG_FILETRIGGERNAME,
+    RPMDBI_TRANSFILETRIGGERNAME = RPMTAG_TRANSFILETRIGGERNAME,
+    RPMDBI_RECOMMENDNAME	= RPMTAG_RECOMMENDNAME,
+    RPMDBI_SUGGESTNAME		= RPMTAG_SUGGESTNAME,
+    RPMDBI_SUPPLEMENTNAME	= RPMTAG_SUPPLEMENTNAME,
+    RPMDBI_ENHANCENAME		= RPMTAG_ENHANCENAME,
 } rpmDbiTag;
 
 /** \ingroup signature
@@ -370,11 +422,14 @@ typedef enum rpmSigTag_e {
     RPMSIGTAG_RESERVEDSPACE = 1008,/*!< internal space reserved for signatures */
     RPMSIGTAG_BADSHA1_1	= RPMTAG_BADSHA1_1,	/*!< internal Broken SHA1, take 1. */
     RPMSIGTAG_BADSHA1_2	= RPMTAG_BADSHA1_2,	/*!< internal Broken SHA1, take 2. */
-    RPMSIGTAG_SHA1	= RPMTAG_SHA1HEADER,	/*!< internal sha1 header digest. */
     RPMSIGTAG_DSA	= RPMTAG_DSAHEADER,	/*!< internal DSA header signature. */
     RPMSIGTAG_RSA	= RPMTAG_RSAHEADER,	/*!< internal RSA header signature. */
+    RPMSIGTAG_SHA1	= RPMTAG_SHA1HEADER,	/*!< internal sha1 header digest. */
     RPMSIGTAG_LONGSIZE	= RPMTAG_LONGSIGSIZE,	/*!< internal Header+Payload size (64bit) in bytes. */
     RPMSIGTAG_LONGARCHIVESIZE = RPMTAG_LONGARCHIVESIZE, /*!< internal uncompressed payload size (64bit) in bytes. */
+    RPMSIGTAG_SHA256	= RPMTAG_SHA256HEADER,
+    RPMSIGTAG_FILESIGNATURES		= RPMTAG_SIG_BASE + 18,
+    RPMSIGTAG_FILESIGNATURELENGTH	= RPMTAG_SIG_BASE + 19,
 } rpmSigTag;
 
 
@@ -408,24 +463,6 @@ typedef enum rpmTagClass_e {
     RPM_BINARY_CLASS	= 3,
 } rpmTagClass;
 
-/** \ingroup header
- * New rpm data types under consideration/development.
- * These data types may (or may not) be added to rpm at some point. In order
- * to avoid incompatibility with legacy versions of rpm, these data (sub-)types
- * are introduced into the header by overloading RPM_BIN_TYPE, with the binary
- * value of the tag a 16 byte image of what should/will be in the header index,
- * followed by per-tag private data.
- */
-typedef enum rpmSubTagType_e {
-    RPM_REGION_TYPE		= -10,
-    RPM_BIN_ARRAY_TYPE		= -11,
-  /*!<@todo Implement, kinda like RPM_STRING_ARRAY_TYPE for known (but variable)
-	length binary data. */
-    RPM_XREF_TYPE		= -12
-  /*!<@todo Implement, intent is to to carry a (???,tagNum,valNum) cross
-	reference to retrieve data from other tags. */
-} rpmSubTagType;
-
 /** \ingroup header
  *  * Identify how to return the header data type.
  *   */
