@@ -6,6 +6,7 @@
 
 #include <inttypes.h>
 #include <libgen.h>
+#include <fcntl.h>
 
 #include <rpm/rpmtypes.h>
 #include <rpm/rpmlib.h>			/* rpmReadPackage etc */
@@ -19,6 +20,7 @@
 #include <rpm/rpmds.h>
 #include <rpm/rpmfi.h>
 #include <rpm/rpmlog.h>
+#include <rpm/rpmsq.h>
 #include <rpm/rpmte.h>
 
 #include "rpmio/digest.h"
@@ -28,6 +30,7 @@
 #include "lib/rpmts_internal.h"
 #include "lib/rpmte_internal.h"
 #include "lib/misc.h"
+#include "lib/rpmtriggers.h"
 
 #include "debug.h"
 
@@ -133,17 +136,21 @@ int rpmtsRebuildDB(rpmts ts)
 {
     int rc = -1;
     rpmtxn txn = NULL;
+    int rebuildflags = 0;
 
     /* Cannot do this on a populated transaction set */
     if (rpmtsNElements(ts) > 0)
 	return -1;
 
+    if (rpmExpandNumeric("%{?_rebuilddb_salvage}"))
+	rebuildflags |= RPMDB_REBUILD_FLAG_SALVAGE;
+
     txn = rpmtxnBegin(ts, RPMTXN_WRITE);
     if (txn) {
 	if (!(ts->vsflags & RPMVSF_NOHDRCHK))
-	    rc = rpmdbRebuild(ts->rootDir, ts, headerCheck);
+	    rc = rpmdbRebuild(ts->rootDir, ts, headerCheck, rebuildflags);
 	else
-	    rc = rpmdbRebuild(ts->rootDir, NULL, NULL);
+	    rc = rpmdbRebuild(ts->rootDir, NULL, NULL, rebuildflags);
 	rpmtxnEnd(txn);
     }
     return rc;
@@ -170,12 +177,12 @@ rpmdbMatchIterator rpmtsInitIterator(const rpmts ts, rpmDbiTagVal rpmtag,
     if (ts == NULL)
 	return NULL;
 
-    if (ts && ts->keyring == NULL)
-	loadKeyring(ts);
-
     if (ts->rdb == NULL && rpmtsOpenDB(ts, ts->dbmode))
 	return NULL;
 
+    if (ts->keyring == NULL)
+	loadKeyring(ts);
+
     /* Parse out "N(EVR)" tokens from a label key if present */
     if (rpmtag == RPMDBI_LABEL && keyp != NULL && strchr(keyp, '(')) {
 	const char *se, *s = keyp;
@@ -274,7 +281,10 @@ static int loadKeyringFromFiles(rpmts ts)
     }
 
     for (char **f = files; *f; f++) {
+	int subkeysCount, i;
+	rpmPubkey *subkeys;
 	rpmPubkey key = rpmPubkeyRead(*f);
+
 	if (!key) {
 	    rpmlog(RPMLOG_ERR, _("%s: reading of public key failed.\n"), *f);
 	    continue;
@@ -283,7 +293,22 @@ static int loadKeyringFromFiles(rpmts ts)
 	    nkeys++;
 	    rpmlog(RPMLOG_DEBUG, "added key %s to keyring\n", *f);
 	}
+	subkeys = rpmGetSubkeys(key, &subkeysCount);
 	rpmPubkeyFree(key);
+
+	for (i = 0; i < subkeysCount; i++) {
+	    rpmPubkey subkey = subkeys[i];
+
+	    if (rpmKeyringAddKey(ts->keyring, subkey) == 0) {
+		rpmlog(RPMLOG_DEBUG,
+		    "added subkey %d of main key %s to keyring\n",
+		    i, *f);
+
+		nkeys++;
+	    }
+	    rpmPubkeyFree(subkey);
+	}
+	free(subkeys);
     }
 exit:
     free(pkpath);
@@ -312,6 +337,9 @@ static int loadKeyringFromDB(rpmts ts)
 
 	    if (rpmBase64Decode(key, (void **) &pkt, &pktlen) == 0) {
 		rpmPubkey key = rpmPubkeyNew(pkt, pktlen);
+		int subkeysCount, i;
+		rpmPubkey *subkeys = rpmGetSubkeys(key, &subkeysCount);
+
 		if (rpmKeyringAddKey(ts->keyring, key) == 0) {
 		    char *nvr = headerGetAsString(h, RPMTAG_NVR);
 		    rpmlog(RPMLOG_DEBUG, "added key %s to keyring\n", nvr);
@@ -319,6 +347,22 @@ static int loadKeyringFromDB(rpmts ts)
 		    nkeys++;
 		}
 		rpmPubkeyFree(key);
+
+		for (i = 0; i < subkeysCount; i++) {
+		    rpmPubkey subkey = subkeys[i];
+
+		    if (rpmKeyringAddKey(ts->keyring, subkey) == 0) {
+			char *nvr = headerGetAsString(h, RPMTAG_NVR);
+			rpmlog(RPMLOG_DEBUG,
+			    "added subkey %d of main key %s to keyring\n",
+			    i, nvr);
+
+			free(nvr);
+			nkeys++;
+		    }
+		    rpmPubkeyFree(subkey);
+		}
+		free(subkeys);
 		free(pkt);
 	    }
 	}
@@ -332,7 +376,8 @@ static int loadKeyringFromDB(rpmts ts)
 static void loadKeyring(rpmts ts)
 {
     /* Never load the keyring if signature checking is disabled */
-    if ((rpmtsVSFlags(ts) & _RPMVSF_NOSIGNATURES) != _RPMVSF_NOSIGNATURES) {
+    if ((rpmtsVSFlags(ts) & RPMVSF_MASK_NOSIGNATURES) !=
+	RPMVSF_MASK_NOSIGNATURES) {
 	ts->keyring = rpmKeyringNew();
 	if (loadKeyringFromFiles(ts) == 0) {
 	    if (loadKeyringFromDB(ts) > 0) {
@@ -343,28 +388,99 @@ static void loadKeyring(rpmts ts)
     }
 }
 
+static void addGpgProvide(Header h, const char *n, const char *v)
+{
+    rpmsenseFlags pflags = (RPMSENSE_KEYRING|RPMSENSE_EQUAL);
+    char * nsn = rstrscat(NULL, "gpg(", n, ")", NULL);
+
+    headerPutString(h, RPMTAG_PROVIDENAME, nsn);
+    headerPutString(h, RPMTAG_PROVIDEVERSION, v);
+    headerPutUint32(h, RPMTAG_PROVIDEFLAGS, &pflags, 1);
+
+    free(nsn);
+}
+
+struct pgpdata_s {
+    char *signid;
+    char *timestr;
+    char *verid;
+    const char *userid;
+    const char *shortid;
+    uint32_t time;
+};
+
+static void initPgpData(pgpDigParams pubp, struct pgpdata_s *pd)
+{
+    memset(pd, 0, sizeof(*pd));
+    pd->signid = pgpHexStr(pubp->signid, sizeof(pubp->signid));
+    pd->shortid = pd->signid + 8;
+    pd->userid = pubp->userid ? pubp->userid : "none";
+    pd->time = pubp->time;
+
+    rasprintf(&pd->timestr, "%x", pd->time);
+    rasprintf(&pd->verid, "%d:%s-%s", pubp->version, pd->signid, pd->timestr);
+}
+
+static void finiPgpData(struct pgpdata_s *pd)
+{
+    free(pd->timestr);
+    free(pd->verid);
+    free(pd->signid);
+    memset(pd, 0, sizeof(*pd));
+}
+
+static Header makeImmutable(Header h)
+{
+    h = headerReload(h, RPMTAG_HEADERIMMUTABLE);
+    if (h != NULL) {
+	char *sha1 = NULL;
+	char *sha256 = NULL;
+	unsigned int blen = 0;
+	void *blob = headerExport(h, &blen);
+
+	/* XXX FIXME: bah, this code is repeated in way too many places */
+	rpmDigestBundle bundle = rpmDigestBundleNew();
+	rpmDigestBundleAdd(bundle, PGPHASHALGO_SHA1, RPMDIGEST_NONE);
+	rpmDigestBundleAdd(bundle, PGPHASHALGO_SHA256, RPMDIGEST_NONE);
+
+	rpmDigestBundleUpdate(bundle, rpm_header_magic, sizeof(rpm_header_magic));
+	rpmDigestBundleUpdate(bundle, blob, blen);
+
+	rpmDigestBundleFinal(bundle, PGPHASHALGO_SHA1, (void **)&sha1, NULL, 1);
+	rpmDigestBundleFinal(bundle, PGPHASHALGO_SHA256, (void **)&sha256, NULL, 1);
+
+	if (sha1 && sha256) {
+	    headerPutString(h, RPMTAG_SHA1HEADER, sha1);
+	    headerPutString(h, RPMTAG_SHA256HEADER, sha256);
+	} else {
+	    h = headerFree(h);
+	}
+	free(sha1);
+	free(sha256);
+	free(blob);
+	rpmDigestBundleFree(bundle);
+    }
+    return h;
+}
+
 /* Build pubkey header. */
-static int makePubkeyHeader(rpmts ts, rpmPubkey key, Header * hdrp)
+static int makePubkeyHeader(rpmts ts, rpmPubkey key, rpmPubkey *subkeys,
+			    int subkeysCount, Header * hdrp)
 {
     Header h = headerNew();
     const char * afmt = "%{pubkeys:armor}";
     const char * group = "Public Keys";
     const char * license = "pubkey";
     const char * buildhost = "localhost";
-    const char * userid;
-    rpmsenseFlags pflags = (RPMSENSE_KEYRING|RPMSENSE_EQUAL);
     uint32_t zero = 0;
-    uint32_t keytime = 0;
     pgpDig dig = NULL;
     pgpDigParams pubp = NULL;
+    struct pgpdata_s kd;
     char * d = NULL;
     char * enc = NULL;
-    char * n = NULL;
-    char * u = NULL;
-    char * v = NULL;
-    char * r = NULL;
-    char * evr = NULL;
+    char * s = NULL;
     int rc = -1;
+    int i;
 
     if ((enc = rpmPubkeyBase64(key)) == NULL)
 	goto exit;
@@ -374,75 +490,54 @@ static int makePubkeyHeader(rpmts ts, rpmPubkey key, Header * hdrp)
 	goto exit;
 
     /* Build header elements. */
-    v = pgpHexStr(pubp->signid, sizeof(pubp->signid)); 
-    r = pgpHexStr(pubp->time, sizeof(pubp->time));
-    userid = pubp->userid ? pubp->userid : "none";
-    keytime = pgpGrab(pubp->time, sizeof(pubp->time));
-
-    rasprintf(&n, "gpg(%s)", v+8);
-    rasprintf(&u, "gpg(%s)", userid);
-    rasprintf(&evr, "%d:%s-%s", pubp->version, v, r);
+    initPgpData(pubp, &kd);
 
+    rasprintf(&s, "%s public key", kd.userid);
     headerPutString(h, RPMTAG_PUBKEYS, enc);
 
     if ((d = headerFormat(h, afmt, NULL)) == NULL)
 	goto exit;
 
     headerPutString(h, RPMTAG_NAME, "gpg-pubkey");
-    headerPutString(h, RPMTAG_VERSION, v+8);
-    headerPutString(h, RPMTAG_RELEASE, r);
+    headerPutString(h, RPMTAG_VERSION, kd.shortid);
+    headerPutString(h, RPMTAG_RELEASE, kd.timestr);
     headerPutString(h, RPMTAG_DESCRIPTION, d);
     headerPutString(h, RPMTAG_GROUP, group);
     headerPutString(h, RPMTAG_LICENSE, license);
-    headerPutString(h, RPMTAG_SUMMARY, u);
-    headerPutString(h, RPMTAG_PACKAGER, userid);
-
+    headerPutString(h, RPMTAG_SUMMARY, s);
+    headerPutString(h, RPMTAG_PACKAGER, kd.userid);
     headerPutUint32(h, RPMTAG_SIZE, &zero, 1);
-
-    headerPutString(h, RPMTAG_PROVIDENAME, u);
-    headerPutString(h, RPMTAG_PROVIDEVERSION, evr);
-    headerPutUint32(h, RPMTAG_PROVIDEFLAGS, &pflags, 1);
-	
-    headerPutString(h, RPMTAG_PROVIDENAME, n);
-    headerPutString(h, RPMTAG_PROVIDEVERSION, evr);
-    headerPutUint32(h, RPMTAG_PROVIDEFLAGS, &pflags, 1);
-
     headerPutString(h, RPMTAG_RPMVERSION, RPMVERSION);
     headerPutString(h, RPMTAG_BUILDHOST, buildhost);
-    headerPutUint32(h, RPMTAG_BUILDTIME, &keytime, 1);
+    headerPutUint32(h, RPMTAG_BUILDTIME, &kd.time, 1);
     headerPutString(h, RPMTAG_SOURCERPM, "(none)");
 
-    /* Reload the lot to immutable region and stomp sha1 digest on it */
-    h = headerReload(h, RPMTAG_HEADERIMMUTABLE);
+    addGpgProvide(h, kd.userid, kd.verid);
+    addGpgProvide(h, kd.shortid, kd.verid);
+    addGpgProvide(h, kd.signid, kd.verid);
+
+    for (i = 0; i < subkeysCount; i++) {
+	struct pgpdata_s skd;
+	initPgpData(rpmPubkeyPgpDigParams(subkeys[i]), &skd);
+	addGpgProvide(h, skd.shortid, skd.verid);
+	addGpgProvide(h, skd.signid, skd.verid);
+	finiPgpData(&skd);
+    }
+
+    /* Reload it into immutable region and stomp standard digests on it */
+    h = makeImmutable(h);
     if (h != NULL) {
-	char *sha1 = NULL;
-	unsigned int blen = 0;
-	const void *blob = headerExport(h, &blen);
-
-	/* XXX FIXME: bah, this code is repeated in way too many places */
-	DIGEST_CTX ctx = rpmDigestInit(PGPHASHALGO_SHA1, RPMDIGEST_NONE);
-	rpmDigestUpdate(ctx, rpm_header_magic, sizeof(rpm_header_magic));
-	rpmDigestUpdate(ctx, blob, blen);
-	rpmDigestFinal(ctx, (void **)&sha1, NULL, 1);
-
-	if (sha1) {
-	    headerPutString(h, RPMTAG_SHA1HEADER, sha1);
-	    *hdrp = headerLink(h);
-	    rc = 0;
-	}
-	free(sha1);
+	*hdrp = headerLink(h);
+	rc = 0;
     }
 
 exit:
     headerFree(h);
     pgpFreeDig(dig);
-    free(n);
-    free(u);
-    free(v);
-    free(r);
-    free(evr);
+    finiPgpData(&kd);
     free(enc);
     free(d);
+    free(s);
 
     return rc;
 }
@@ -464,21 +559,27 @@ rpmRC rpmtsImportPubkey(const rpmts ts, const unsigned char * pkt, size_t pktlen
     Header h = NULL;
     rpmRC rc = RPMRC_FAIL;		/* assume failure */
     rpmPubkey pubkey = NULL;
+    rpmPubkey *subkeys = NULL;
+    int subkeysCount = 0;
     rpmVSFlags oflags = rpmtsVSFlags(ts);
     rpmKeyring keyring;
     rpmtxn txn = rpmtxnBegin(ts, RPMTXN_WRITE);
-    int krc;
+    int krc, i;
 
     if (txn == NULL)
 	return rc;
 
     /* XXX keyring wont load if sigcheck disabled, force it temporarily */
-    rpmtsSetVSFlags(ts, (oflags & ~_RPMVSF_NOSIGNATURES));
+    rpmtsSetVSFlags(ts, (oflags & ~RPMVSF_MASK_NOSIGNATURES));
     keyring = rpmtsGetKeyring(ts, 1);
     rpmtsSetVSFlags(ts, oflags);
 
     if ((pubkey = rpmPubkeyNew(pkt, pktlen)) == NULL)
 	goto exit;
+
+    if ((subkeys = rpmGetSubkeys(pubkey, &subkeysCount)) == NULL)
+	goto exit;
+
     krc = rpmKeyringAddKey(keyring, pubkey);
     if (krc < 0)
 	goto exit;
@@ -487,7 +588,7 @@ rpmRC rpmtsImportPubkey(const rpmts ts, const unsigned char * pkt, size_t pktlen
     if (krc == 0) {
 	rpm_tid_t tid = rpmtsGetTid(ts);
 
-	if (makePubkeyHeader(ts, pubkey, &h) != 0) 
+	if (makePubkeyHeader(ts, pubkey, subkeys, subkeysCount, &h) != 0)
 	    goto exit;
 
 	headerPutUint32(h, RPMTAG_INSTALLTIME, &tid, 1);
@@ -504,6 +605,10 @@ exit:
     /* Clean up. */
     headerFree(h);
     rpmPubkeyFree(pubkey);
+    for (i = 0; i < subkeysCount; i++)
+	rpmPubkeyFree(subkeys[i]);
+    free(subkeys);
+
     rpmKeyringFree(keyring);
     rpmtxnEnd(txn);
     return rc;
@@ -608,7 +713,7 @@ void rpmtsEmpty(rpmts ts)
     tsmem->orderCount = 0;
     /* The pool cannot be emptied, there might be references to its contents */
     tsmem->pool = rpmstrPoolFree(tsmem->pool);
-    removedHashEmpty(tsmem->removedPackages);
+    packageHashEmpty(tsmem->removedPackages);
     return;
 }
 
@@ -629,6 +734,7 @@ static void rpmtsPrintStats(rpmts ts)
     rpmtsPrintStat("total:       ", rpmtsOp(ts, RPMTS_OP_TOTAL));
     rpmtsPrintStat("check:       ", rpmtsOp(ts, RPMTS_OP_CHECK));
     rpmtsPrintStat("order:       ", rpmtsOp(ts, RPMTS_OP_ORDER));
+    rpmtsPrintStat("verify:      ", rpmtsOp(ts, RPMTS_OP_VERIFY));
     rpmtsPrintStat("fingerprint: ", rpmtsOp(ts, RPMTS_OP_FINGERPRINT));
     rpmtsPrintStat("install:     ", rpmtsOp(ts, RPMTS_OP_INSTALL));
     rpmtsPrintStat("erase:       ", rpmtsOp(ts, RPMTS_OP_ERASE));
@@ -657,7 +763,8 @@ rpmts rpmtsFree(rpmts ts)
 
     (void) rpmtsCloseDB(ts);
 
-    tsmem->removedPackages = removedHashFree(tsmem->removedPackages);
+    tsmem->removedPackages = packageHashFree(tsmem->removedPackages);
+    tsmem->installedPackages = packageHashFree(tsmem->installedPackages);
     tsmem->order = _free(tsmem->order);
     ts->members = _free(ts->members);
 
@@ -677,6 +784,8 @@ rpmts rpmtsFree(rpmts ts)
 
     ts->plugins = rpmpluginsFree(ts->plugins);
 
+    rpmtriggersFree(ts->trigs2run);
+
     if (_rpmts_stats)
 	rpmtsPrintStats(ts);
 
@@ -705,6 +814,42 @@ rpmVSFlags rpmtsSetVSFlags(rpmts ts, rpmVSFlags vsflags)
     return ovsflags;
 }
 
+rpmVSFlags rpmtsVfyFlags(rpmts ts)
+{
+    rpmVSFlags vfyflags = 0;
+    if (ts != NULL)
+	vfyflags = ts->vfyflags;
+    return vfyflags;
+}
+
+rpmVSFlags rpmtsSetVfyFlags(rpmts ts, rpmVSFlags vfyflags)
+{
+    rpmVSFlags ovfyflags = 0;
+    if (ts != NULL) {
+	ovfyflags = ts->vfyflags;
+	ts->vfyflags = vfyflags;
+    }
+    return ovfyflags;
+}
+
+int rpmtsVfyLevel(rpmts ts)
+{
+    int vfylevel = 0;
+    if (ts != NULL)
+	vfylevel = ts->vfylevel;
+    return vfylevel;
+}
+
+int rpmtsSetVfyLevel(rpmts ts, int vfylevel)
+{
+    int ovfylevel = 0;
+    if (ts != NULL) {
+	ovfylevel = ts->vfylevel;
+	ts->vfylevel = vfylevel;
+    }
+    return ovfylevel;
+}
+
 const char * rpmtsRootDir(rpmts ts)
 {
     return ts ? ts->rootDir : NULL;
@@ -915,6 +1060,26 @@ rpmstrPool rpmtsPool(rpmts ts)
     return tspool;
 }
 
+static int vfylevel_init(void)
+{
+    int vfylevel = -1;
+    char *val = rpmExpand("%{?_pkgverify_level}", NULL);
+
+    if (rstreq(val, "all"))
+	vfylevel = RPMSIG_SIGNATURE_TYPE|RPMSIG_DIGEST_TYPE;
+    else if (rstreq(val, "signature"))
+	vfylevel = RPMSIG_SIGNATURE_TYPE;
+    else if (rstreq(val, "digest"))
+	vfylevel = RPMSIG_DIGEST_TYPE;
+    else if (rstreq(val, "none"))
+	vfylevel = 0;
+    else if (!rstreq(val, ""))
+	rpmlog(RPMLOG_WARNING, _("invalid package verify level %s\n"), val);
+
+    free(val);
+    return vfylevel;
+}
+
 rpmts rpmtsCreate(void)
 {
     rpmts ts;
@@ -965,7 +1130,8 @@ rpmts rpmtsCreate(void)
     tsmem->pool = NULL;
     tsmem->delta = 5;
     tsmem->addedPackages = NULL;
-    tsmem->removedPackages = removedHashCreate(128, uintId, uintCmp, NULL, NULL);
+    tsmem->removedPackages = packageHashCreate(128, uintId, uintCmp, NULL, NULL);
+    tsmem->installedPackages = packageHashCreate(128, uintId, uintCmp, NULL, NULL);
     tsmem->orderAlloced = 0;
     tsmem->orderCount = 0;
     tsmem->order = NULL;
@@ -973,11 +1139,17 @@ rpmts rpmtsCreate(void)
 
     ts->rootDir = NULL;
     ts->keyring = NULL;
+    ts->vfyflags = rpmExpandNumeric("%{?_pkgverify_flags}");
+    ts->vfylevel = vfylevel_init();
 
     ts->nrefs = 0;
 
     ts->plugins = NULL;
 
+    ts->trigs2run = rpmtriggersCreate(10);
+
+    ts->min_writes = (rpmExpandNumeric("%{?_minimize_writes}") > 0);
+
     return rpmtsLink(ts);
 }
 
@@ -1066,6 +1238,8 @@ rpmtxn rpmtxnBegin(rpmts ts, rpmtxnFlags flags)
 	txn->lock = ts->lock;
 	txn->flags = flags;
 	txn->ts = rpmtsLink(ts);
+	if (txn->flags & RPMTXN_WRITE)
+	    rpmsqBlock(SIG_BLOCK);
     }
     
     return txn;
@@ -1075,6 +1249,8 @@ rpmtxn rpmtxnEnd(rpmtxn txn)
 {
     if (txn) {
 	rpmlockRelease(txn->lock);
+	if (txn->flags & RPMTXN_WRITE)
+	    rpmsqBlock(SIG_UNBLOCK);
 	rpmtsFree(txn->ts);
 	free(txn);
     }
