@@ -9,6 +9,7 @@
 #include <sys/types.h>
 
 #include <rpm/rpmtypes.h>
+#include <rpm/rpmte.h>
 #include <rpm/rpmps.h>
 #include <rpm/rpmsw.h>
 #include <rpm/rpmpgp.h>
@@ -19,13 +20,12 @@
 extern "C" {
 #endif
 
-extern int _rpmts_debug;
 extern int _rpmts_stats;
 
 /** \ingroup rpmts
  * Bit(s) to control rpmtsRun() operation.
  */
-typedef enum rpmtransFlags_e {
+enum rpmtransFlags_e {
     RPMTRANS_FLAG_NONE		= 0,
     RPMTRANS_FLAG_TEST		= (1 <<  0),	/*!< from --test */
     RPMTRANS_FLAG_BUILD_PROBS	= (1 <<  1),	/*!< don't process payload */
@@ -34,17 +34,10 @@ typedef enum rpmtransFlags_e {
     RPMTRANS_FLAG_NOTRIGGERS	= (1 <<  4),	/*!< from --notriggers */
     RPMTRANS_FLAG_NODOCS	= (1 <<  5),	/*!< from --excludedocs */
     RPMTRANS_FLAG_ALLFILES	= (1 <<  6),	/*!< from --allfiles */
-    RPMTRANS_FLAG_KEEPOBSOLETE	= (1 <<  7),	/*!< @todo Document. */
+    RPMTRANS_FLAG_NOPLUGINS	= (1 <<  7),	/*!< from --noplugins */
     RPMTRANS_FLAG_NOCONTEXTS	= (1 <<  8),	/*!< from --nocontexts */
-    RPMTRANS_FLAG_DIRSTASH	= (1 <<  9),	/*!< obsolete, unused */
-    RPMTRANS_FLAG_REPACKAGE	= (1 << 10),	/*!< obsolete, unused */
-
-    RPMTRANS_FLAG_PKGCOMMIT	= (1 << 11),
-    RPMTRANS_FLAG_PKGUNDO	= (1 << 12),
-    RPMTRANS_FLAG_COMMIT	= (1 << 13),
-    RPMTRANS_FLAG_UNDO		= (1 << 14),
-    RPMTRANS_FLAG_REVERSE	= (1 << 15),
-
+    RPMTRANS_FLAG_NOCAPS	= (1 <<  9),	/*!< from --nocaps */
+    /* bits 10-15 unused */
     RPMTRANS_FLAG_NOTRIGGERPREIN= (1 << 16),	/*!< from --notriggerprein */
     RPMTRANS_FLAG_NOPRE		= (1 << 17),	/*!< from --nopre */
     RPMTRANS_FLAG_NOPOST	= (1 << 18),	/*!< from --nopost */
@@ -53,22 +46,26 @@ typedef enum rpmtransFlags_e {
     RPMTRANS_FLAG_NOPREUN	= (1 << 21),	/*!< from --nopreun */
     RPMTRANS_FLAG_NOPOSTUN	= (1 << 22),	/*!< from --nopostun */
     RPMTRANS_FLAG_NOTRIGGERPOSTUN = (1 << 23),	/*!< from --notriggerpostun */
-    RPMTRANS_FLAG_NOPAYLOAD	= (1 << 24),
-    RPMTRANS_FLAG_APPLYONLY	= (1 << 25),
-
+    RPMTRANS_FLAG_NOPRETRANS	= (1 << 24),	/*!< from --nopretrans */
+    RPMTRANS_FLAG_NOPOSTTRANS	= (1 << 25),	/*!< from --noposttrans */
+    /* bit 26 unused */
     RPMTRANS_FLAG_NOMD5		= (1 << 27),	/*!< from --nomd5 */
     RPMTRANS_FLAG_NOFILEDIGEST	= (1 << 27),	/*!< from --nofiledigest (alias to --nomd5) */
-    RPMTRANS_FLAG_NOSUGGEST	= (1 << 28),	/*!< from --nosuggest */
-    RPMTRANS_FLAG_ADDINDEPS	= (1 << 29),	/*!< from --aid */
+    /* bit 28 unused */
+    RPMTRANS_FLAG_NOARTIFACTS	= (1 << 29),	/*!< from --noartifacts */
     RPMTRANS_FLAG_NOCONFIGS	= (1 << 30),	/*!< from --noconfigs */
     RPMTRANS_FLAG_DEPLOOPS	= (1 << 31)	/*!< from --deploops */
-} rpmtransFlags;
+};
+
+typedef rpmFlags rpmtransFlags;
 
 #define	_noTransScripts		\
   ( RPMTRANS_FLAG_NOPRE |	\
     RPMTRANS_FLAG_NOPOST |	\
     RPMTRANS_FLAG_NOPREUN |	\
-    RPMTRANS_FLAG_NOPOSTUN	\
+    RPMTRANS_FLAG_NOPOSTUN |	\
+    RPMTRANS_FLAG_NOPRETRANS |	\
+    RPMTRANS_FLAG_NOPOSTTRANS \
   )
 
 #define	_noTransTriggers	\
@@ -78,49 +75,79 @@ typedef enum rpmtransFlags_e {
     RPMTRANS_FLAG_NOTRIGGERPOSTUN \
   )
 
+/* Avoid unnecessary breakage for stuff referring to these unused flags */
+#define RPMTRANS_FLAG_NOPAYLOAD 0
+#define RPMTRANS_FLAG_APPLYONLY 0
+#define RPMTRANS_FLAG_KEEPOBSOLETE 0
+#define RPMTRANS_FLAG_DIRSTASH 0
+#define RPMTRANS_FLAG_REPACKAGE 0
+#define RPMTRANS_FLAG_PKGCOMMIT 0
+#define RPMTRANS_FLAG_PKGUNDO 0
+#define RPMTRANS_FLAG_COMMIT 0
+#define RPMTRANS_FLAG_UNDO 0
+#define RPMTRANS_FLAG_REVERSE 0
+#define RPMTRANS_FLAG_NOSUGGEST 0
+#define RPMTRANS_FLAG_ADDINDEPS 0
+
 /** \ingroup rpmts
  * Bit(s) to control digest and signature verification.
  */
-typedef enum rpmVSFlags_e {
+enum rpmVSFlags_e {
     RPMVSF_DEFAULT	= 0,
     RPMVSF_NOHDRCHK	= (1 <<  0),
     RPMVSF_NEEDPAYLOAD	= (1 <<  1),
     /* bit(s) 2-7 unused */
     RPMVSF_NOSHA1HEADER	= (1 <<  8),
-    RPMVSF_NOMD5HEADER	= (1 <<  9),	/* unimplemented */
+    RPMVSF_NOSHA256HEADER = (1 <<  9),
     RPMVSF_NODSAHEADER	= (1 << 10),
-    RPMVSF_NORSAHEADER	= (1 << 11),	/* unimplemented */
+    RPMVSF_NORSAHEADER	= (1 << 11),
     /* bit(s) 12-15 unused */
-    RPMVSF_NOSHA1	= (1 << 16),	/* unimplemented */
+    RPMVSF_NOPAYLOAD	= (1 << 16),
     RPMVSF_NOMD5	= (1 << 17),
     RPMVSF_NODSA	= (1 << 18),
     RPMVSF_NORSA	= (1 << 19)
     /* bit(s) 16-31 unused */
-} rpmVSFlags;
+};
 
-#define	_RPMVSF_NODIGESTS	\
+typedef rpmFlags rpmVSFlags;
+
+#define	RPMVSF_MASK_NODIGESTS	\
   ( RPMVSF_NOSHA1HEADER |	\
-    RPMVSF_NOMD5HEADER |	\
-    RPMVSF_NOSHA1 |		\
+    RPMVSF_NOSHA256HEADER |	\
+    RPMVSF_NOPAYLOAD |	\
     RPMVSF_NOMD5 )
+#define	_RPMVSF_NODIGESTS	RPMVSF_MASK_NODIGESTS
 
-#define	_RPMVSF_NOSIGNATURES	\
+#define	RPMVSF_MASK_NOSIGNATURES	\
   ( RPMVSF_NODSAHEADER |	\
     RPMVSF_NORSAHEADER |	\
     RPMVSF_NODSA |		\
     RPMVSF_NORSA )
+#define	_RPMVSF_NOSIGNATURES	RPMVSF_MASK_NOSIGNATURES
 
-#define	_RPMVSF_NOHEADER	\
+#define	RPMVSF_MASK_NOHEADER	\
   ( RPMVSF_NOSHA1HEADER |	\
-    RPMVSF_NOMD5HEADER |	\
+    RPMVSF_NOSHA256HEADER |	\
     RPMVSF_NODSAHEADER |	\
     RPMVSF_NORSAHEADER )
+#define	_RPMVSF_NOHEADER	RPMVSF_MASK_NOHEADER
 
-#define	_RPMVSF_NOPAYLOAD	\
-  ( RPMVSF_NOSHA1 |		\
-    RPMVSF_NOMD5 |		\
+#define	RPMVSF_MASK_NOPAYLOAD	\
+  ( RPMVSF_NOMD5 |		\
+    RPMVSF_NOPAYLOAD |		\
     RPMVSF_NODSA |		\
     RPMVSF_NORSA )
+#define	_RPMVSF_NOPAYLOAD	RPMVSF_MASK_NOPAYLOAD
+
+enum {
+    RPMSIG_NONE_TYPE		= 0,
+    RPMSIG_DIGEST_TYPE		= (1 << 0),
+    RPMSIG_SIGNATURE_TYPE	= (1 << 1),
+    RPMSIG_OTHER_TYPE		= (1 << 2),
+};
+
+#define RPMSIG_VERIFIABLE_TYPE (RPMSIG_DIGEST_TYPE|RPMSIG_SIGNATURE_TYPE)
+#define RPMSIG_UNVERIFIED_TYPE 	(1 << 30)
 
 /** \ingroup rpmts
  * Indices for timestamps.
@@ -142,9 +169,16 @@ typedef	enum rpmtsOpX_e {
     RPMTS_OP_DBGET		= 14,
     RPMTS_OP_DBPUT		= 15,
     RPMTS_OP_DBDEL		= 16,
-    RPMTS_OP_MAX		= 17
+    RPMTS_OP_VERIFY		= 17,
+    RPMTS_OP_MAX		= 18
 } rpmtsOpX;
 
+enum rpmtxnFlags_e {
+    RPMTXN_READ		= (1 << 0),
+    RPMTXN_WRITE	= (1 << 1),
+};
+typedef rpmFlags rpmtxnFlags;
+
 /** \ingroup rpmts
  * Perform dependency resolution on the transaction set.
  *
@@ -188,28 +222,18 @@ int rpmtsOrder(rpmts ts);
  *    - setup the rpm verify signature flags via rpmtsSetVSFlags().
  *       
  * @param ts		transaction set
- * @param okProbs	previously known problems (or NULL)
+ * @param okProbs	unused
  * @param ignoreSet	bits to filter problem types
  * @return		0 on success, -1 on error, >0 with newProbs set
  */
 int rpmtsRun(rpmts ts, rpmps okProbs, rpmprobFilterFlags ignoreSet);
 
-/** \ingroup rpmts
- * Unreference a transaction instance.
- * @param ts		transaction set
- * @param msg
- * @return		NULL always
- */
-rpmts rpmtsUnlink (rpmts ts,
-		const char * msg);
-
 /** \ingroup rpmts
  * Reference a transaction set instance.
  * @param ts		transaction set
- * @param msg
  * @return		new transaction set reference
  */
-rpmts rpmtsLink (rpmts ts, const char * msg);
+rpmts rpmtsLink (rpmts ts);
 
 /** \ingroup rpmts
  * Close the database used by the transaction.
@@ -268,21 +292,22 @@ int rpmtsVerifyDB(rpmts ts);
 /** \ingroup rpmts
  * Return transaction database iterator.
  * @param ts		transaction set
- * @param rpmtag	rpm tag
+ * @param rpmtag	database index tag
  * @param keyp		key data (NULL for sequential access)
  * @param keylen	key data length (0 will use strlen(keyp))
  * @return		NULL on failure
  */
-rpmdbMatchIterator rpmtsInitIterator(const rpmts ts, rpmTag rpmtag,
+rpmdbMatchIterator rpmtsInitIterator(const rpmts ts, rpmDbiTagVal rpmtag,
 			const void * keyp, size_t keylen);
 
 /** \ingroup rpmts
- * Retrieve pubkey from rpm database.
- * @param ts		rpm transaction
- * @param dig		OpenPGP packet container
- * @return		RPMRC_OK on success, RPMRC_NOKEY if not found
+ * Import a header into the rpmdb
+ * @param txn		transaction handle
+ * @param h		header
+ * @param flags		(unused)
+ * @return              RPMRC_OK/RPMRC_FAIL
  */
-rpmRC rpmtsFindPubkey(rpmts ts, pgpDig dig) RPM_GNUC_DEPRECATED;
+rpmRC rpmtsImportHeader(rpmtxn txn, Header h, rpmFlags flags);
 
 /** \ingroup rpmts
  * Import public key packet(s).
@@ -326,7 +351,7 @@ int rpmtsSetSolveCallback(rpmts ts,
 /** \ingroup rpmts
  * Return current transaction set problems.
  * @param ts		transaction set
- * @return		current problem set (or NULL)
+ * @return		current problem set (or NULL if no problems)
  */
 rpmps rpmtsProblems(rpmts ts);
 
@@ -370,6 +395,36 @@ rpmVSFlags rpmtsVSFlags(rpmts ts);
  */
 rpmVSFlags rpmtsSetVSFlags(rpmts ts, rpmVSFlags vsflags);
 
+/** \ingroup rpmts
+ * Get package verify flag(s).
+ * @param ts		transaction set
+ * @return		verify signatures flags
+ */
+rpmVSFlags rpmtsVfyFlags(rpmts ts);
+
+/** \ingroup rpmts
+ * Set package verify flag(s).
+ * @param ts		transaction set
+ * @param vfyflags	new package verify flags
+ * @return		old package verify flags
+ */
+rpmVSFlags rpmtsSetVfyFlags(rpmts ts, rpmVSFlags vfyflags);
+
+/** \ingroup rpmts
+ * Get enforced package verify level
+ * @param ts		transaction set
+ * @return		package verify level
+ */
+int rpmtsVfyLevel(rpmts ts);
+
+/** \ingroup rpmts
+ * Set enforced package verify level
+ * @param ts		transaction set
+ * @param vfylevel	new package verify level
+ * @return		old package verify level
+ */
+int rpmtsSetVfyLevel(rpmts ts, int vfylevel);
+
 /** \ingroup rpmts
  * Get transaction rootDir, i.e. path to chroot(2).
  * @param ts		transaction set
@@ -385,20 +440,6 @@ const char * rpmtsRootDir(rpmts ts);
  */
 int rpmtsSetRootDir(rpmts ts, const char * rootDir);
 
-/** \ingroup rpmts
- * Get transaction currDir, i.e. current directory before chroot(2).
- * @param ts		transaction set
- * @return		transaction currDir
- */
-const char * rpmtsCurrDir(rpmts ts);
-
-/** \ingroup rpmts
- * Set transaction currDir, i.e. current directory before chroot(2).
- * @param ts		transaction set
- * @param currDir	new transaction currDir (or NULL)
- */
-void rpmtsSetCurrDir(rpmts ts, const char * currDir);
-
 /** \ingroup rpmts
  * Get transaction script file handle, i.e. stdout/stderr on scriptlet execution
  * @param ts		transaction set
@@ -413,28 +454,6 @@ FD_t rpmtsScriptFd(rpmts ts);
  */
 void rpmtsSetScriptFd(rpmts ts, FD_t scriptFd);
 
-/** \ingroup rpmts
- * Get selinuxEnabled flag, i.e. is SE linux enabled?
- * @param ts		transaction set
- * @return		selinuxEnabled flag
- */
-int rpmtsSELinuxEnabled(rpmts ts);
-
-/** \ingroup rpmts
- * Get chrootDone flag, i.e. has chroot(2) been performed?
- * @param ts		transaction set
- * @return		chrootDone flag
- */
-int rpmtsChrootDone(rpmts ts);
-
-/** \ingroup rpmts
- * Set chrootDone flag, i.e. has chroot(2) been performed?
- * @param ts		transaction set
- * @param chrootDone	new chrootDone flag
- * @return		previous chrootDone flag
- */
-int rpmtsSetChrootDone(rpmts ts, int chrootDone);
-
 /** \ingroup rpmts
  * Get transaction id, i.e. transaction time stamp.
  * @param ts		transaction set
@@ -506,21 +525,6 @@ rpmtransFlags rpmtsFlags(rpmts ts);
  */
 rpmtransFlags rpmtsSetFlags(rpmts ts, rpmtransFlags transFlags);
 
-/** \ingroup rpmts
- * Get spec control structure from transaction set.
- * @param ts		transaction set
- * @return		spec control structure
- */
-rpmSpec rpmtsSpec(rpmts ts);
-
-/** \ingroup rpmts
- * Set a spec control structure in transaction set.
- * @param ts		transaction set
- * @param spec		new spec control structure
- * @return		previous spec control structure
- */
-rpmSpec rpmtsSetSpec(rpmts ts, rpmSpec spec);
-
 /** \ingroup rpmts
  * Retrieve color bits of transaction set.
  * @param ts		transaction set
@@ -529,7 +533,7 @@ rpmSpec rpmtsSetSpec(rpmts ts, rpmSpec spec);
 rpm_color_t rpmtsColor(rpmts ts);
 
 /** \ingroup rpmts
- * Retrieve prefered file color
+ * Retrieve preferred file color
  * @param ts		transaction set
  * @return		color bits
  */
@@ -544,7 +548,7 @@ rpm_color_t rpmtsPrefColor(rpmts ts);
 rpm_color_t rpmtsSetColor(rpmts ts, rpm_color_t color);
 
 /** \ingroup rpmts
- * Set prefered file color
+ * Set preferred file color
  * @param ts		transaction set
  * @param color		new color bits
  * @return		previous color bits
@@ -559,6 +563,13 @@ rpm_color_t rpmtsSetPrefColor(rpmts ts, rpm_color_t color);
  */
 rpmop rpmtsOp(rpmts ts, rpmtsOpX opx);
 
+/** \ingroup rpmts
+ * Get the plugins associated with a transaction set
+ * @param ts		transaction set
+ * @return		plugins
+ */
+rpmPlugins rpmtsPlugins(rpmts ts);
+
 /** \ingroup rpmts
  * Set transaction notify callback function and argument.
  *
@@ -597,6 +608,16 @@ int rpmtsAddInstallElement(rpmts ts, Header h,
 		const fnpyKey key, int upgrade,
 		rpmRelocation * relocs);
 
+/** \ingroup rpmts
+ * Add package to be reinstalled to transaction set.
+ *
+ * @param ts		transaction set
+ * @param h		header
+ * @param key		package retrieval key (e.g. file name)
+ * @return		0 on success
+ */
+int rpmtsAddReinstallElement(rpmts ts, Header h, const fnpyKey key);
+
 /** \ingroup rpmts
  * Add package to be erased to transaction set.
  * @param ts		transaction set
@@ -606,6 +627,43 @@ int rpmtsAddInstallElement(rpmts ts, Header h,
  */
 int rpmtsAddEraseElement(rpmts ts, Header h, int dboffset);
 
+/** \ingroup rpmts
+ * Create a transaction (lock) handle
+ * @param ts		transaction set
+ * @param flags		flags
+ * @return		transaction handle
+ */
+rpmtxn rpmtxnBegin(rpmts ts, rpmtxnFlags flags);
+
+/** \ingroup rpmts
+ * Destroy transaction (lock) handle
+ * @param txn		transaction handle
+ * @return		NULL always
+ */
+rpmtxn rpmtxnEnd(rpmtxn txn);
+
+/** \ingroup rpmte
+ * Destroy transaction element iterator.
+ * @param tsi		transaction element iterator
+ * @return		NULL always
+ */
+rpmtsi rpmtsiFree(rpmtsi tsi);
+
+/** \ingroup rpmte
+ * Create transaction element iterator.
+ * @param ts		transaction set
+ * @return		transaction element iterator
+ */
+rpmtsi rpmtsiInit(rpmts ts);
+
+/** \ingroup rpmte
+ * Return next transaction element of type.
+ * @param tsi		transaction element iterator
+ * @param types		transaction element type selector (0 for any)
+ * @return		next transaction element of type, NULL on termination
+ */
+rpmte rpmtsiNext(rpmtsi tsi, rpmElementTypes types);
+
 #ifdef __cplusplus
 }
 #endif
