@@ -4,14 +4,14 @@
 #include <rpm/rpmtag.h>
 #include <rpm/rpmstring.h>
 #include <rpm/rpmts.h>	/* XXX rpmtsCreate/rpmtsFree */
+#include <rpm/rpmver.h>
 
 #include "header-py.h"
 #include "rpmds-py.h"
 #include "rpmfd-py.h"
 #include "rpmfi-py.h"
 #include "rpmtd-py.h"
-
-#include "debug.h"
+#include "rpmver-py.h"
 
 /** \ingroup python
  * \class Rpm
@@ -88,18 +88,18 @@
  * 	hdr = ts.hdrFromFdno(fdno)
  *	os.close(fdno)
  *	if hdr[rpm.RPMTAG_SOURCEPACKAGE]:
- *	   print "header is from a source package"
+ *	   print("header is from a source package")
  *	else:
- *	   print "header is from a binary package"
+ *	   print("header is from a binary package")
  * \endcode
  *
  * The Python interface to the header data is quite elegant.  It
  * presents the data in a dictionary form.  We'll take the header we
  * just loaded and access the data within it:
  * \code
- * 	print hdr[rpm.RPMTAG_NAME]
- * 	print hdr[rpm.RPMTAG_VERSION]
- * 	print hdr[rpm.RPMTAG_RELEASE]
+ * 	print(hdr[rpm.RPMTAG_NAME])
+ * 	print(hdr[rpm.RPMTAG_VERSION])
+ * 	print(hdr[rpm.RPMTAG_RELEASE])
  * \endcode
  * in the case of our "foo-1.0-1.i386.rpm" package, this code would
  * output:
@@ -111,9 +111,9 @@
  *
  * You make also access the header data by string name:
  * \code
- * 	print hdr['name']
- * 	print hdr['version']
- * 	print hdr['release']
+ * 	print(hdr['name'])
+ * 	print(hdr['version'])
+ * 	print(hdr['release'])
  * \endcode
  *
  * This method of access is a teensy bit slower because the name must be
@@ -134,12 +134,23 @@ struct hdrObject_s {
 
 static PyObject * hdrKeyList(hdrObject * s)
 {
-    PyObject * keys = PyList_New(0);
-    HeaderIterator hi = headerInitIterator(s->h);
+    PyObject * keys;
+    HeaderIterator hi;
     rpmTagVal tag;
 
+    keys = PyList_New(0);
+    if (!keys) {
+        return NULL;
+    }
+
+    hi = headerInitIterator(s->h);
     while ((tag = headerNextTag(hi)) != RPMTAG_NOT_FOUND) {
-	PyObject *to = PyInt_FromLong(tag);
+	PyObject *to = PyLong_FromLong(tag);
+        if (!to) {
+            headerFreeIterator(hi);
+            Py_DECREF(keys);
+            return NULL;
+        }
 	PyList_Append(keys, to);
 	Py_DECREF(to);
     }
@@ -148,36 +159,24 @@ static PyObject * hdrKeyList(hdrObject * s)
     return keys;
 }
 
-static PyObject * hdrUnload(hdrObject * s, PyObject * args, PyObject *keywords)
+static PyObject * hdrAsBytes(hdrObject * s)
 {
-    char * buf;
-    PyObject * rc;
-    int len, legacy = 0;
-    Header h;
-    static char *kwlist[] = { "legacyHeader", NULL};
-
-    if (!PyArg_ParseTupleAndKeywords(args, keywords, "|i", kwlist, &legacy))
-	return NULL;
-
-    h = headerLink(s->h);
-    /* XXX this legacy switch is a hack, needs to be removed. */
-    if (legacy) {
-	h = headerCopy(s->h);	/* XXX strip region tags, etc */
-	headerFree(s->h);
-    }
-    len = headerSizeof(h, HEADER_MAGIC_NO);
-    buf = headerUnload(h);
-    h = headerFree(h);
+    PyObject *res = NULL;
+    unsigned int len = 0;
+    char *buf = headerExport(s->h, &len);
 
     if (buf == NULL || len == 0) {
 	PyErr_SetString(pyrpmError, "can't unload bad header\n");
-	return NULL;
+    } else {
+	res = PyBytes_FromStringAndSize(buf, len);
     }
+    free(buf);
+    return res;
+}
 
-    rc = PyBytes_FromStringAndSize(buf, len);
-    buf = _free(buf);
-
-    return rc;
+static PyObject * hdrUnload(hdrObject * s)
+{
+    return hdrAsBytes(s);
 }
 
 static PyObject * hdrExpandFilelist(hdrObject * s)
@@ -211,7 +210,6 @@ static PyObject * hdrFullFilelist(hdrObject * s)
     if (headerGet(h, RPMTAG_FILENAMES, fileNames, HEADERGET_EXT)) {
 	rpmtdSetTag(fileNames, RPMTAG_OLDFILENAMES);
 	headerPut(h, fileNames, HEADERPUT_DEFAULT);
-	rpmtdFreeData(fileNames);
     }
     rpmtdFree(fileNames);
 
@@ -220,7 +218,7 @@ static PyObject * hdrFullFilelist(hdrObject * s)
 
 static PyObject * hdrFormat(hdrObject * s, PyObject * args, PyObject * kwds)
 {
-    char * fmt;
+    const char * fmt;
     char * r;
     errmsg_t err;
     PyObject * result;
@@ -235,8 +233,8 @@ static PyObject * hdrFormat(hdrObject * s, PyObject * args, PyObject * kwds)
 	return NULL;
     }
 
-    result = Py_BuildValue("s", r);
-    r = _free(r);
+    result = utf8FromString(r);
+    free(r);
 
     return result;
 }
@@ -294,8 +292,7 @@ static PyObject * hdrWrite(hdrObject *s, PyObject *args, PyObject *kwds)
  */
 static PyObject * hdr_fiFromHeader(PyObject * s, PyObject * args, PyObject * kwds)
 {
-    return PyObject_Call((PyObject *) &rpmfi_Type,
-			 Py_BuildValue("(O)", s), NULL);
+    return PyObject_CallFunctionObjArgs((PyObject *) &rpmfi_Type, s, NULL);
 }
 
 /* Backwards compatibility. Flags argument is just a dummy and discarded. */
@@ -308,14 +305,14 @@ static PyObject * hdr_dsFromHeader(PyObject * s, PyObject * args, PyObject * kwd
             tagNumFromPyObject, &tag, &flags))
         return NULL;
 
-    return PyObject_Call((PyObject *) &rpmds_Type,
-			 Py_BuildValue("(Oi)", s, tag), NULL);
+    return PyObject_CallFunction((PyObject *) &rpmds_Type,
+                                 "(Oi)", s, tag);
 }
 
 static PyObject * hdr_dsOfHeader(PyObject * s)
 {
-    return PyObject_Call((PyObject *) &rpmds_Type,
-			Py_BuildValue("(Oi)", s, RPMTAG_NEVR), NULL);
+    return PyObject_CallFunction((PyObject *) &rpmds_Type,
+                                 "(Oi)", s, RPMTAG_NEVR);
 }
 
 static long hdr_hash(PyObject * h)
@@ -323,32 +320,45 @@ static long hdr_hash(PyObject * h)
     return (long) h;
 }
 
+static PyObject * hdr_reduce(hdrObject *s)
+{
+    PyObject *res = NULL;
+    PyObject *blob = hdrAsBytes(s);
+    if (blob) {
+	res = Py_BuildValue("O(O)", Py_TYPE(s), blob);
+	Py_DECREF(blob);
+    }
+    return res;
+}
+
 static struct PyMethodDef hdr_methods[] = {
     {"keys",		(PyCFunction) hdrKeyList,	METH_NOARGS,
-	NULL },
-    {"unload",		(PyCFunction) hdrUnload,	METH_VARARGS|METH_KEYWORDS,
-	NULL },
+     "hdr.keys() -- Return a list of the header's rpm tags (int RPMTAG_*)." },
+    {"unload",		(PyCFunction) hdrUnload,	METH_NOARGS,
+     "hdr.unload() -- Return binary representation\nof the header." },
     {"expandFilelist",	(PyCFunction) hdrExpandFilelist,METH_NOARGS,
-	NULL },
+     "DEPRECATED -- Use hdr.convert() instead." },
     {"compressFilelist",(PyCFunction) hdrCompressFilelist,METH_NOARGS,
-	NULL },
+     "DEPRECATED -- Use hdr.convert() instead." },
     {"fullFilelist",	(PyCFunction) hdrFullFilelist,	METH_NOARGS,
-	NULL },
+     "DEPRECATED -- Obsolete method."},
     {"convert",		(PyCFunction) hdrConvert,	METH_VARARGS|METH_KEYWORDS,
-	NULL },
+     "hdr.convert(op=-1) -- Convert header - See HEADERCONV_*\nfor possible values of op."},
     {"format",		(PyCFunction) hdrFormat,	METH_VARARGS|METH_KEYWORDS,
-	NULL },
+     "hdr.format(format) -- Expand a query string with the header data.\n\nSee rpm -q for syntax." },
     {"sprintf",		(PyCFunction) hdrFormat,	METH_VARARGS|METH_KEYWORDS,
-	NULL },
+     "Alias for .format()." },
     {"isSource",	(PyCFunction)hdrIsSource,	METH_NOARGS, 
-	NULL },
+     "hdr.isSource() -- Return if header describes a source package." },
     {"write",		(PyCFunction)hdrWrite,		METH_VARARGS|METH_KEYWORDS,
-	NULL },
+     "hdr.write(file, magic=True) -- Write header to file." },
     {"dsOfHeader",	(PyCFunction)hdr_dsOfHeader,	METH_NOARGS,
-	NULL},
+     "hdr.dsOfHeader() -- Return dependency set with the header's NEVR."},
     {"dsFromHeader",	(PyCFunction)hdr_dsFromHeader,	METH_VARARGS|METH_KEYWORDS,
-	NULL},
+     "hdr.dsFromHeader(to=RPMTAG_REQUIRENAME, flags=None)\nGet dependency set from header. to must be one of the NAME tags\nbelonging to a dependency:\n'Providename', 'Requirename', 'Obsoletename', 'Conflictname',\n'Triggername', 'Recommendname', 'Suggestname', 'Supplementname',\n'Enhancename' or one of the corresponding RPMTAG_*NAME constants." },
     {"fiFromHeader",	(PyCFunction)hdr_fiFromHeader,	METH_VARARGS|METH_KEYWORDS,
+     "hdr.fiFromHeader() -- Return rpm.fi object containing the file\nmeta data from the header.\n\nDEPRECATED - Use rpm.files(hdr) instead."},
+    {"__reduce__",	(PyCFunction)hdr_reduce,	METH_NOARGS,
 	NULL},
 
     {NULL,		NULL}		/* sentinel */
@@ -368,8 +378,6 @@ static PyObject *hdr_new(PyTypeObject *subtype, PyObject *args, PyObject *kwds)
 
     if (obj == NULL) {
 	h = headerNew();
-    } else if (PyCObject_Check(obj)) {
-	h = PyCObject_AsVoidPtr(obj);
     } else if (hdrObject_Check(obj)) {
 	h = headerCopy(((hdrObject*) obj)->h);
     } else if (PyBytes_Check(obj)) {
@@ -406,13 +414,21 @@ static PyObject * hdr_iternext(hdrObject *s)
     if (s->hi == NULL) s->hi = headerInitIterator(s->h);
 
     if ((tag = headerNextTag(s->hi)) != RPMTAG_NOT_FOUND) {
-	res = PyInt_FromLong(tag);
+	res = PyLong_FromLong(tag);
     } else {
 	s->hi = headerFreeIterator(s->hi);
     }
     return res;
 }
 
+PyObject * utf8FromString(const char *s)
+{
+    /* We return all strings as surrogate-escaped utf-8 */
+    if (s != NULL)
+	return PyUnicode_DecodeUTF8(s, strlen(s), "surrogateescape");
+    Py_RETURN_NONE;
+}
+
 int utf8FromPyObject(PyObject *item, PyObject **str)
 {
     PyObject *res = NULL;
@@ -433,9 +449,9 @@ int tagNumFromPyObject (PyObject *item, rpmTagVal *tagp)
     rpmTagVal tag = RPMTAG_NOT_FOUND;
     PyObject *str = NULL;
 
-    if (PyInt_Check(item)) {
+    if (PyLong_Check(item)) {
 	/* XXX we should probably validate tag numbers too */
-	tag = PyInt_AsLong(item);
+	tag = PyLong_AsLong(item);
     } else if (utf8FromPyObject(item, &str)) {
 	tag = rpmTagGetValue(PyBytes_AsString(str));
 	Py_DECREF(str);
@@ -457,9 +473,13 @@ static PyObject * hdrGetTag(Header h, rpmTagVal tag)
     PyObject *res = NULL;
     struct rpmtd_s td;
 
-    /* rpmtd_AsPyObj() knows how to handle empty containers and all */
     (void) headerGet(h, tag, &td, HEADERGET_EXT);
-    res = rpmtd_AsPyobj(&td);
+    if (rpmtdGetFlags(&td) & RPMTD_INVALID) {
+	PyErr_SetString(pyrpmError, "invalid header data");
+    } else {
+	/* rpmtd_AsPyobj() knows how to handle empty containers and all */
+	res = rpmtd_AsPyobj(&td);
+    }
     rpmtdFreeData(&td);
     return res;
 }
@@ -470,7 +490,7 @@ static int validItem(rpmTagClass tclass, PyObject *item)
 
     switch (tclass) {
     case RPM_NUMERIC_CLASS:
-	rc = (PyLong_Check(item) || PyInt_Check(item));
+	rc = (PyLong_Check(item) || PyLong_Check(item));
 	break;
     case RPM_STRING_CLASS:
 	rc = (PyBytes_Check(item) || PyUnicode_Check(item));
@@ -495,6 +515,8 @@ static int validData(rpmTagVal tag, rpmTagType type, rpmTagReturnType retype, Py
     } else if (retype == RPM_ARRAY_RETURN_TYPE && PyList_Check(value)) {
 	/* python lists can contain arbitrary objects, validate each item */
 	Py_ssize_t len = PyList_Size(value);
+	if (len == 0)
+	    valid = 0;
 	for (Py_ssize_t i = 0; i < len; i++) {
 	    PyObject *item = PyList_GetItem(value, i);
 	    if (!validItem(tclass, item)) {
@@ -527,20 +549,20 @@ static int hdrAppendItem(Header h, rpmTagVal tag, rpmTagType type, PyObject *ite
 	rc = headerPutBin(h, tag, val, len);
 	} break;
     case RPM_INT64_TYPE: {
-	uint64_t val = PyInt_AsUnsignedLongLongMask(item);
+	uint64_t val = PyLong_AsUnsignedLongLongMask(item);
 	rc = headerPutUint64(h, tag, &val, 1);
 	} break;
     case RPM_INT32_TYPE: {
-	uint32_t val = PyInt_AsUnsignedLongMask(item);
+	uint32_t val = PyLong_AsUnsignedLongMask(item);
 	rc = headerPutUint32(h, tag, &val, 1);
 	} break;
     case RPM_INT16_TYPE: {
-	uint16_t val = PyInt_AsUnsignedLongMask(item);
+	uint16_t val = PyLong_AsUnsignedLongMask(item);
 	rc = headerPutUint16(h, tag, &val, 1);
 	} break;
     case RPM_INT8_TYPE:
     case RPM_CHAR_TYPE: {
-	uint8_t val = PyInt_AsUnsignedLongMask(item);
+	uint8_t val = PyLong_AsUnsignedLongMask(item);
 	rc = headerPutUint8(h, tag, &val, 1);
 	} break;
     default:
@@ -576,7 +598,7 @@ static int hdrPutTag(Header h, rpmTagVal tag, PyObject *value)
 	    rc = hdrAppendItem(h, tag, type, item);
 	}
     } else {
-	PyErr_SetString(PyExc_RuntimeError, "cant happen, right?");
+	PyErr_SetString(PyExc_RuntimeError, "can't happen, right?");
     }
 
     return rc;
@@ -650,7 +672,56 @@ static PySequenceMethods hdr_as_sequence = {
 };
 
 static char hdr_doc[] =
-"";
+  "A header object represents an RPM package header.\n"
+  "\n"
+  "All RPM packages have headers that provide metadata for the package.\n"
+  "Header objects can be returned by database queries or loaded from a\n"
+  "binary package on disk.\n"
+  "\n"
+  "The ts.hdrFromFdno() function returns the package header from a\n"
+  "package on disk, verifying package signatures and digests of the\n"
+  "package while reading.\n"
+  "\n"
+  "Note: The older method rpm.headerFromPackage() which has been replaced\n"
+  "by ts.hdrFromFdno() used to return a (hdr, isSource) tuple.\n"
+  "\n"
+  "If you need to distinguish source/binary headers, do:\n"
+  "\n"
+  "	import os, rpm\n"
+  "\n"
+  "	ts = rpm.TransactionSet()\n"
+  "	fdno = os.open('/tmp/foo-1.0-1.i386.rpm', os.O_RDONLY)\n"
+  "	hdr = ts.hdrFromFdno(fdno)\n"
+  "	os.close(fdno)\n"
+  "	if hdr[rpm.RPMTAG_SOURCEPACKAGE]:\n"
+  "	   print('header is from a source package')\n"
+  "	else:\n"
+  "	   print('header is from a binary package')\n"
+  "\n"
+  "The Python interface to the header data is quite elegant.  It\n"
+  "presents the data in a dictionary form.  We'll take the header we\n"
+  "just loaded and access the data within it:\n"
+  "\n"
+  "	print(hdr[rpm.RPMTAG_NAME])\n"
+  "	print(hdr[rpm.RPMTAG_VERSION])\n"
+  "	print(hdr[rpm.RPMTAG_RELEASE])\n"
+  "\n"
+  "in the case of our 'foo-1.0-1.i386.rpm' package, this code would\n"
+  "output:\n"
+  "	foo\n"
+  "	1.0\n"
+  "	1\n"
+  "\n"
+  "You make also access the header data by string name:\n"
+  "\n"
+  "	print(hdr['name'])\n"
+  "	print(hdr['version'])\n"
+  "	print(hdr['release'])\n"
+  "\n"
+  "This method of access is a teensy bit slower because the name must be\n"
+  "translated into the tag number dynamically. You also must make sure\n"
+  "the strings in header lookups don't get translated, or the lookups\n"
+  "will fail.\n";
 
 PyTypeObject hdr_Type = {
 	PyVarObject_HEAD_INIT(&PyType_Type, 0)
@@ -699,8 +770,7 @@ PyObject * hdr_Wrap(PyTypeObject *subtype, Header h)
 {
     hdrObject * hdr = (hdrObject *)subtype->tp_alloc(subtype, 0);
     if (hdr == NULL) return NULL;
-
-    hdr->h = headerLink(h);
+    hdr->h = h;
     return (PyObject *) hdr;
 }
 
@@ -820,36 +890,22 @@ PyObject * versionCompare (PyObject * self, PyObject * args, PyObject * kwds)
     return Py_BuildValue("i", rpmVersionCompare(h1->h, h2->h));
 }
 
-static int compare_values(const char *str1, const char *str2)
-{
-    if (!str1 && !str2)
-	return 0;
-    else if (str1 && !str2)
-	return 1;
-    else if (!str1 && str2)
-	return -1;
-    return rpmvercmp(str1, str2);
-}
-
 PyObject * labelCompare (PyObject * self, PyObject * args)
 {
-    char *v1, *r1, *v2, *r2;
-    const char *e1, *e2;
-    int rc;
+    PyObject *rco = NULL;
+    rpmver rv1 = NULL;
+    rpmver rv2 = NULL;
 
-    if (!PyArg_ParseTuple(args, "(zzz)(zzz)",
-			&e1, &v1, &r1, &e2, &v2, &r2))
+    if (!PyArg_ParseTuple(args, "O&O&",
+			    verFromPyObject, &rv1, verFromPyObject, &rv2)) {
 	return NULL;
-
-    if (e1 == NULL)	e1 = "0";
-    if (e2 == NULL)	e2 = "0";
-
-    rc = compare_values(e1, e2);
-    if (!rc) {
-	rc = compare_values(v1, v2);
-	if (!rc)
-	    rc = compare_values(r1, r2);
     }
-    return Py_BuildValue("i", rc);
+
+    rco = Py_BuildValue("i", rpmverCmp(rv1, rv2));
+
+    rpmverFree(rv1);
+    rpmverFree(rv2);
+
+    return rco;
 }
 
