@@ -26,6 +26,7 @@
 
 #include <stdlib.h>
 #include <string.h>
+#include <dirent.h>
 #include <pwd.h>
 #include <assert.h>
 #include <sys/stat.h>		/* S_ISDIR */
@@ -75,16 +76,6 @@ typedef struct {
     int (*gl_stat)(const char *, struct stat *);
 } glob_t;
 
-#define	NAMLEN(_d)	NLENGTH(_d)
-
-#if (defined POSIX || defined WINDOWS32) && !defined __GNU_LIBRARY__
-/* Posix does not require that the d_ino field be present, and some
-   systems do not provide it. */
-#define REAL_DIR_ENTRY(dp) 1
-#else
-#define REAL_DIR_ENTRY(dp) (dp->d_ino != 0)
-#endif				/* POSIX */
-
 #include <errno.h>
 #ifndef __set_errno
 #define __set_errno(val) errno = (val)
@@ -124,30 +115,15 @@ static inline const char *next_brace_sub(const char *begin)
     unsigned int depth = 0;
     const char *cp = begin;
 
-    while (1) {
-	if (depth == 0) {
-	    if (*cp != ',' && *cp != '}' && *cp != '\0') {
-		if (*cp == '{')
-		    ++depth;
-		++cp;
-		continue;
-	    }
-	} else {
-	    while (*cp != '\0' && (*cp != '}' || depth > 0)) {
-		if (*cp == '}')
-		    --depth;
-		++cp;
-	    }
-	    if (*cp == '\0')
-		/* An incorrectly terminated brace expression.  */
-		return NULL;
+    while (*cp != '\0') {
+	if ((*cp == '}' && depth-- == 0) || (*cp == ',' && depth == 0))
+	    break;
 
-	    continue;
-	}
-	break;
+	if (*cp++ == '{')
+	    depth++;
     }
 
-    return cp;
+    return *cp != '\0' ? cp : NULL;
 }
 
 static int __glob_pattern_p(const char *pattern, int quote);
@@ -337,7 +313,7 @@ glob(const char *pattern, int flags,
 	    else {
 		char *newp;
 		size_t home_len = strlen(home_dir);
-		newp = (char *) alloca(home_len + dirlen);
+		newp = (char *) alloca(home_len + dirlen + 1);
 		mempcpy(mempcpy(newp, home_dir, home_len),
 			&dirname[1], dirlen);
 		dirname = newp;
@@ -352,7 +328,7 @@ glob(const char *pattern, int flags,
 		user_name = dirname + 1;
 	    else {
 		char *newp;
-		newp = (char *) alloca(end_name - dirname);
+		newp = (char *) alloca(end_name - dirname + 1);
 		*((char *) mempcpy(newp, dirname + 1, end_name - dirname))
 		    = '\0';
 		user_name = newp;
@@ -649,7 +625,7 @@ static int prefix_array(const char *dirname, char **array, size_t n)
 static int __glob_pattern_p(const char *pattern, int quote)
 {
     register const char *p;
-    int open = 0;
+    int openBrackets = 0;
 
     for (p = pattern; *p != '\0'; ++p)
 	switch (*p) {
@@ -663,11 +639,11 @@ static int __glob_pattern_p(const char *pattern, int quote)
 	    break;
 
 	case '[':
-	    open = 1;
+	    openBrackets = 1;
 	    break;
 
 	case ']':
-	    if (open)
+	    if (openBrackets)
 		return 1;
 	    break;
 	}
@@ -755,8 +731,6 @@ glob_in_dir(const char *pattern, const char *directory, int flags,
 					: readdir((DIR *) stream));
 		    if (d == NULL)
 			break;
-		    if (!REAL_DIR_ENTRY(d))
-			continue;
 
 #ifdef HAVE_STRUCT_DIRENT_D_TYPE
 		    /* If we shall match only directories use the information
@@ -771,7 +745,7 @@ glob_in_dir(const char *pattern, const char *directory, int flags,
 		    if (fnmatch(pattern, name, fnm_flags) == 0) {
 			struct globlink *new = (struct globlink *)
 			    alloca(sizeof(struct globlink));
-			len = NAMLEN(d);
+			len = strlen(d->d_name);
 			new->name = (char *) xmalloc(len + 1);
 			*((char *) mempcpy(new->name, name, len))
 			    = '\0';
@@ -844,6 +818,8 @@ int rpmGlob(const char * patterns, int * argcPtr, ARGV_t * argvPtr)
     int i, j;
     int rc;
 
+    gflags |= GLOB_BRACE;
+
     if (home != NULL && strlen(home) > 0) 
 	gflags |= GLOB_TILDE;
 
@@ -966,5 +942,32 @@ exit:
 
 int rpmIsGlob(const char * pattern, int quote)
 {
-    return __glob_pattern_p(pattern, quote);
+    if (!__glob_pattern_p(pattern, quote)) {
+
+	const char *begin;
+	const char *next;
+	const char *rest;
+
+	begin = strchr(pattern, '{');
+	if (begin == NULL)
+	    return 0;
+	/*
+	 * Find the first sub-pattern and at the same time find the
+	 *  rest after the closing brace.
+	 */
+	next = next_brace_sub(begin + 1);
+	if (next == NULL)
+	    return 0;
+
+	/* Now find the end of the whole brace expression.  */
+	rest = next;
+	while (*rest != '}') {
+	    rest = next_brace_sub(rest + 1);
+	    if (rest == NULL)
+		return 0;
+	}
+	/* Now we can be sure that brace expression is well-foermed. */
+    }
+
+    return 1;
 }
