@@ -8,6 +8,7 @@
 #include <errno.h>
 #include <sys/wait.h>
 #include <popt.h>
+#include <fcntl.h>
 
 #include <rpm/rpmlib.h>			/* RPMSIGTAG & related */
 #include <rpm/rpmmacro.h>
@@ -16,16 +17,15 @@
 #include <rpm/rpmfileutil.h>	/* rpmMkTemp() */
 #include <rpm/rpmlog.h>
 #include <rpm/rpmstring.h>
+#include <rpmio/rpmio_internal.h>
 
 #include "lib/rpmlead.h"
 #include "lib/signature.h"
+#include "lib/rpmvs.h"
+#include "sign/rpmsignfiles.h"
 
 #include "debug.h"
 
-#if !defined(__GLIBC__) && !defined(__APPLE__)
-char ** environ = NULL;
-#endif
-
 typedef struct sigTarget_s {
     FD_t fd;
     const char *fileName;
@@ -56,7 +56,7 @@ static int manageFile(FD_t *fdp, const char *fn, int flags)
 
     /* open a file and set *fdp */
     if (*fdp == NULL && fn != NULL) {
-	switch(flags & O_ACCMODE) {
+	switch (flags & O_ACCMODE) {
 	    case O_WRONLY:
 		fmode = "w.ufdio";
 		break;
@@ -88,6 +88,10 @@ static int manageFile(FD_t *fdp, const char *fn, int flags)
 
 /**
  * Copy header+payload, calculating digest(s) on the fly.
+ * @param sfdp source file
+ * @param sfnp source path
+ * @param tfdp destination file
+ * @param tfnp destination path
  */
 static int copyFile(FD_t *sfdp, const char *sfnp,
 		FD_t *tfdp, const char *tfnp)
@@ -96,11 +100,6 @@ static int copyFile(FD_t *sfdp, const char *sfnp,
     ssize_t count;
     int rc = 1;
 
-    if (manageFile(sfdp, sfnp, O_RDONLY))
-	goto exit;
-    if (manageFile(tfdp, tfnp, O_WRONLY|O_CREAT|O_TRUNC))
-	goto exit;
-
     while ((count = Fread(buf, sizeof(buf[0]), sizeof(buf), *sfdp)) > 0)
     {
 	if (Fwrite(buf, sizeof(buf[0]), count, *tfdp) != count) {
@@ -121,8 +120,6 @@ static int copyFile(FD_t *sfdp, const char *sfnp,
     rc = 0;
 
 exit:
-    if (*sfdp)	(void) closeFile(sfdp);
-    if (*tfdp)	(void) closeFile(tfdp);
     return rc;
 }
 
@@ -130,14 +127,13 @@ exit:
  * Validate generated signature and insert to header if it looks sane.
  * NSS doesn't support everything GPG does. Basic tests to see if the 
  * generated signature is something we can use.
- * Return 0 on success, 1 on failure.
+ * Return generated signature tag data on success, NULL on failure.
  */
-static int putSignature(Header sigh, int ishdr, uint8_t *pkt, size_t pktlen)
+static rpmtd makeSigTag(Header sigh, int ishdr, uint8_t *pkt, size_t pktlen)
 {
     pgpDigParams sigp = NULL;
     rpmTagVal sigtag;
-    struct rpmtd_s sigtd;
-    int rc = 1; /* assume failure */
+    rpmtd sigtd = NULL;
     unsigned int hash_algo;
     unsigned int pubkey_algo;
 
@@ -167,62 +163,53 @@ static int putSignature(Header sigh, int ishdr, uint8_t *pkt, size_t pktlen)
 	break;
     }
 
-    /* Looks sane, insert into header */
-    rpmtdReset(&sigtd);
-    sigtd.count = pktlen;
-    sigtd.data = pkt;
-    sigtd.type = RPM_BIN_TYPE;
-    sigtd.tag = sigtag;
-
-    /* Argh, reversed return codes */
-    rc = (headerPut(sigh, &sigtd, HEADERPUT_DEFAULT) == 0);
+    /* Looks sane, create the tag data */
+    sigtd = rpmtdNew();
+    sigtd->count = pktlen;
+    sigtd->data = memcpy(xmalloc(pktlen), pkt, pktlen);;
+    sigtd->type = RPM_BIN_TYPE;
+    sigtd->tag = sigtag;
+    sigtd->flags |= RPMTD_ALLOCED;
 
 exit:
     pgpDigParamsFree(sigp);
-    return rc;
+    return sigtd;
 }
 
-static int runGPG(sigTarget sigt, const char *sigfile, const char * passPhrase)
+static int runGPG(sigTarget sigt, const char *sigfile)
 {
     int pid = 0, status;
-    int inpipe[2];
-    int inpipe2[2];
-    FILE * fpipe = NULL;
+    int pipefd[2];
+    FILE *fpipe = NULL;
     unsigned char buf[BUFSIZ];
     ssize_t count;
     ssize_t wantCount;
     rpm_loff_t size;
     int rc = 1; /* assume failure */
 
-    inpipe[0] = inpipe[1] = 0;
-    if (pipe(inpipe) < 0) {
-	rpmlog(RPMLOG_ERR, _("Couldn't create pipe for signing: %m"));
-	goto exit;
+    if (pipe(pipefd) < 0) {
+        rpmlog(RPMLOG_ERR, _("Could not create pipe for signing: %m\n"));
+        goto exit;
     }
 
-    inpipe2[0] = inpipe2[1] = 0;
-    if (pipe(inpipe2) < 0) {
-	rpmlog(RPMLOG_ERR, _("Couldn't create pipe for signing: %m"));
-	goto exit;
-    }
-
-    addMacro(NULL, "__plaintext_filename", NULL, "-", -1);
-    addMacro(NULL, "__signature_filename", NULL, sigfile, -1);
+    rpmPushMacro(NULL, "__plaintext_filename", NULL, "-", -1);
+    rpmPushMacro(NULL, "__signature_filename", NULL, sigfile, -1);
 
     if (!(pid = fork())) {
 	char *const *av;
 	char *cmd = NULL;
-	const char *gpg_path = rpmExpand("%{?_gpg_path}", NULL);
+	const char *tty = ttyname(STDIN_FILENO);
+	const char *gpg_path = NULL;
 
-	(void) dup2(inpipe[0], 3);
-	(void) close(inpipe[1]);
-
-	(void) dup2(inpipe2[0], STDIN_FILENO);
-	(void) close(inpipe2[1]);
+	if (!getenv("GPG_TTY") && (!tty || setenv("GPG_TTY", tty, 0)))
+	    rpmlog(RPMLOG_WARNING, _("Could not set GPG_TTY to stdin: %m\n"));
 
+	gpg_path = rpmExpand("%{?_gpg_path}", NULL);
 	if (gpg_path && *gpg_path != '\0')
 	    (void) setenv("GNUPGHOME", gpg_path, 1);
-	(void) setenv("LC_ALL", "C", 1);
+
+	dup2(pipefd[0], STDIN_FILENO);
+	close(pipefd[1]);
 
 	unsetenv("MALLOC_CHECK_");
 	cmd = rpmExpand("%{?__gpg_sign_cmd}", NULL);
@@ -235,34 +222,15 @@ static int runGPG(sigTarget sigt, const char *sigfile, const char * passPhrase)
 	_exit(EXIT_FAILURE);
     }
 
-    delMacro(NULL, "__plaintext_filename");
-    delMacro(NULL, "__signature_filename");
+    rpmPopMacro(NULL, "__plaintext_filename");
+    rpmPopMacro(NULL, "__signature_filename");
 
-    (void) close(inpipe[0]);
-    inpipe[0] = 0;
-    (void) close(inpipe2[0]);
-    inpipe2[0] = 0;
-
-    fpipe = fdopen(inpipe[1], "w");
+    close(pipefd[0]);
+    fpipe = fdopen(pipefd[1], "w");
     if (!fpipe) {
-	rpmlog(RPMLOG_ERR, _("fdopen failed\n"));
+	rpmlog(RPMLOG_ERR, _("Could not open pipe for writing: %m\n"));
 	goto exit;
     }
-    inpipe[1] = 0;
-
-    if (fprintf(fpipe, "%s\n", (passPhrase ? passPhrase : "")) < 0) {
-	rpmlog(RPMLOG_ERR, _("Could not write to pipe\n"));
-	goto exit;
-    }
-    (void) fclose(fpipe);
-    fpipe = NULL;
-
-    fpipe = fdopen(inpipe2[1], "w");
-    if (!fpipe) {
-	rpmlog(RPMLOG_ERR, _("fdopen failed\n"));
-	goto exit;
-    }
-    inpipe2[1] = 0;
 
     if (Fseek(sigt->fd, sigt->start, SEEK_SET) < 0) {
 	rpmlog(RPMLOG_ERR, _("Could not seek in file %s: %s\n"),
@@ -286,8 +254,13 @@ static int runGPG(sigTarget sigt, const char *sigfile, const char * passPhrase)
 		sigt->fileName, Fstrerror(sigt->fd));
 	goto exit;
     }
-    fclose(fpipe);
-    fpipe = NULL;
+
+exit:
+
+    if (fpipe)
+	fclose(fpipe);
+    if (pipefd[1])
+	close(pipefd[1]);
 
     (void) waitpid(pid, &status, 0);
     pid = 0;
@@ -296,27 +269,6 @@ static int runGPG(sigTarget sigt, const char *sigfile, const char * passPhrase)
     } else {
 	rc = 0;
     }
-
-exit:
-    if (fpipe)
-	fclose(fpipe);
-
-    if (inpipe[0])
-	close(inpipe[0]);
-
-    if (inpipe[1])
-	close(inpipe[1]);
-
-    if (inpipe2[0])
-	close(inpipe[0]);
-
-    if (inpipe2[1])
-	close(inpipe[1]);
-
-    if (pid)
-	waitpid(pid, &status, 0);
-
-
     return rc;
 }
 
@@ -326,18 +278,17 @@ exit:
  * @param ishdr		header-only signature?
  * @param sigt		signature target
  * @param passPhrase	private key pass phrase
- * @return		0 on success, 1 on failure
+ * @return		generated sigtag on success, 0 on failure
  */
-static int makeGPGSignature(Header sigh, int ishdr, sigTarget sigt,
-			    const char * passPhrase)
+static rpmtd makeGPGSignature(Header sigh, int ishdr, sigTarget sigt)
 {
     char * sigfile = rstrscat(NULL, sigt->fileName, ".sig", NULL);
     struct stat st;
     uint8_t * pkt = NULL;
     size_t pktlen = 0;
-    int rc = 1; /* assume failure */
+    rpmtd sigtd = NULL;
 
-    if (runGPG(sigt, sigfile, passPhrase))
+    if (runGPG(sigt, sigfile))
 	goto exit;
 
     if (stat(sigfile, &st)) {
@@ -352,7 +303,7 @@ static int makeGPGSignature(Header sigh, int ishdr, sigTarget sigt,
 
     {	FD_t fd;
 
-	rc = 0;
+	int rc = 0;
 	fd = Fopen(sigfile, "r.ufdio");
 	if (fd != NULL && !Ferror(fd)) {
 	    rc = Fread(pkt, sizeof(*pkt), pktlen, fd);
@@ -367,47 +318,13 @@ static int makeGPGSignature(Header sigh, int ishdr, sigTarget sigt,
     rpmlog(RPMLOG_DEBUG, "Got %zd bytes of GPG sig\n", pktlen);
 
     /* Parse the signature, change signature tag as appropriate. */
-    rc = putSignature(sigh, ishdr, pkt, pktlen);
+    sigtd = makeSigTag(sigh, ishdr, pkt, pktlen);
 exit:
     (void) unlink(sigfile);
     free(sigfile);
     free(pkt);
 
-    return rc;
-}
-
-static int rpmGenSignature(Header sigh, sigTarget sigt1, sigTarget sigt2,
-			    const char * passPhrase)
-{
-    int ret;
-
-    ret = makeGPGSignature(sigh, 0, sigt1, passPhrase);
-    if (ret)
-	goto exit;
-
-    ret = makeGPGSignature(sigh, 1, sigt2, passPhrase);
-    if (ret)
-	goto exit;
-exit:
-    return ret;
-}
-
-/**
- * Retrieve signature from header tag
- * @param sigh		signature header
- * @param sigtag	signature tag
- * @return		parsed pgp dig or NULL
- */
-static pgpDigParams getSig(Header sigh, rpmTagVal sigtag)
-{
-    struct rpmtd_s pkt;
-    pgpDigParams sig = NULL;
-
-    if (headerGet(sigh, sigtag, &pkt, HEADERGET_DEFAULT) && pkt.data != NULL) {
-	pgpPrtParams(pkt.data, pkt.count, PGPTAG_SIGNATURE, &sig);
-	rpmtdFreeData(&pkt);
-    }
-    return sig;
+    return sigtd;
 }
 
 static void deleteSigs(Header sigh)
@@ -419,42 +336,134 @@ static void deleteSigs(Header sigh)
     headerDel(sigh, RPMSIGTAG_PGP5);
 }
 
-static int sameSignature(rpmTagVal sigtag, Header h1, Header h2)
+static int haveSignature(rpmtd sigtd, Header h)
 {
-    pgpDigParams sig1 = getSig(h1, sigtag);
-    pgpDigParams sig2 = getSig(h2, sigtag);;
+    pgpDigParams sig1 = NULL;
+    pgpDigParams sig2 = NULL;
+    struct rpmtd_s oldtd;
+    int rc = 0; /* assume no */
 
-    int rc = pgpDigParamsCmp(sig1, sig2);
+    if (!headerGet(h, rpmtdTag(sigtd), &oldtd, HEADERGET_DEFAULT))
+	return rc;
 
+    pgpPrtParams(sigtd->data, sigtd->count, PGPTAG_SIGNATURE, &sig1);
+    while (rpmtdNext(&oldtd) >= 0 && rc == 0) {
+	pgpPrtParams(oldtd.data, oldtd.count, PGPTAG_SIGNATURE, &sig2);
+	if (pgpDigParamsCmp(sig1, sig2) == 0)
+	    rc = 1;
+	pgpDigParamsFree(sig2);
+    }
     pgpDigParamsFree(sig1);
-    pgpDigParamsFree(sig2);
-    return (rc == 0);
+    rpmtdFreeData(&oldtd);
+
+    return rc;
 }
 
-static int replaceSignature(Header sigh, sigTarget sigt1, sigTarget sigt2,
-			    const char *passPhrase)
+static int replaceSignature(Header sigh, sigTarget sigt_v3, sigTarget sigt_v4)
 {
-    /* Grab a copy of the header so we can compare the result */
-    Header oldsigh = headerCopy(sigh);
     int rc = -1;
+    rpmtd sigtd = NULL;
     
+    /* Make the cheaper v4 signature first */
+    if ((sigtd = makeGPGSignature(sigh, 1, sigt_v4)) == NULL)
+	goto exit;
+
+    /* See if we already have a signature by the same key and parameters */
+    if (haveSignature(sigtd, sigh)) {
+	rc = 1;
+	goto exit;
+    }
     /* Nuke all signature tags */
     deleteSigs(sigh);
 
-    /*
-     * rpmGenSignature() internals parse the actual signing result and 
-     * adds appropriate tags for DSA/RSA.
-     */
-    if (rpmGenSignature(sigh, sigt1, sigt2, passPhrase) == 0) {
-	/* Lets see what we got and whether its the same signature as before */
-	rpmTagVal sigtag = headerIsEntry(sigh, RPMSIGTAG_DSA) ?
-					RPMSIGTAG_DSA : RPMSIGTAG_RSA;
+    if (headerPut(sigh, sigtd, HEADERPUT_DEFAULT) == 0)
+	goto exit;
 
-	rc = sameSignature(sigtag, sigh, oldsigh);
+    if (sigt_v3) {
+	rpmtdFree(sigtd);
 
+	/* Assume the same signature test holds for v3 signature too */
+	if ((sigtd = makeGPGSignature(sigh, 0, sigt_v3)) == NULL)
+	    goto exit;
+
+	if (headerPut(sigh, sigtd, HEADERPUT_DEFAULT) == 0)
+	    goto exit;
     }
 
-    headerFree(oldsigh);
+    rc = 0;
+exit:
+    rpmtdFree(sigtd);
+    return rc;
+}
+
+static void unloadImmutableRegion(Header *hdrp, rpmTagVal tag)
+{
+    struct rpmtd_s td;
+    Header oh = NULL;
+
+    if (headerGet(*hdrp, tag, &td, HEADERGET_DEFAULT)) {
+	oh = headerCopyLoad(td.data);
+	rpmtdFreeData(&td);
+    } else {
+	/* XXX should we warn if the immutable region is corrupt/missing? */
+	oh = headerLink(*hdrp);
+    }
+
+    if (oh) {
+	/* Perform a copy to eliminate crud from buggy signing tools etc */
+	Header nh = headerCopy(oh);
+	headerFree(*hdrp);
+	*hdrp = headerLink(nh);
+	headerFree(nh);
+	headerFree(oh);
+    }
+}
+
+static rpmRC includeFileSignatures(Header *sigp, Header *hdrp)
+{
+#ifdef WITH_IMAEVM
+    rpmRC rc;
+    char *key = rpmExpand("%{?_file_signing_key}", NULL);
+    char *keypass = rpmExpand("%{?_file_signing_key_password}", NULL);
+
+    if (rstreq(keypass, "")) {
+	free(keypass);
+	keypass = NULL;
+    }
+
+    rc = rpmSignFiles(*sigp, *hdrp, key, keypass);
+
+    free(keypass);
+    free(key);
+    return rc;
+#else
+    rpmlog(RPMLOG_ERR, _("file signing support not built in\n"));
+    return RPMRC_FAIL;
+#endif
+}
+
+static int msgCb(struct rpmsinfo_s *sinfo, void *cbdata)
+{
+    char **msg = cbdata;
+    if (sinfo->rc && *msg == NULL)
+	*msg = rpmsinfoMsg(sinfo);
+    return (sinfo->rc != RPMRC_FAIL);
+}
+
+/* Require valid digests on entire package for signing. */
+static int checkPkg(FD_t fd, char **msg)
+{
+    int rc;
+    struct rpmvs_s *vs = rpmvsCreate(RPMSIG_DIGEST_TYPE, 0, NULL);
+    off_t offset = Ftell(fd);
+
+    Fseek(fd, 0, SEEK_SET);
+    rc = rpmpkgRead(vs, fd, NULL, NULL, msg);
+    if (!rc)
+	rc = rpmvsVerify(vs, RPMSIG_DIGEST_TYPE, msgCb, msg);
+    Fseek(fd, offset, SEEK_SET);
+
+    rpmvsFree(vs);
     return rc;
 }
 
@@ -462,25 +471,24 @@ static int replaceSignature(Header sigh, sigTarget sigt1, sigTarget sigt2,
  * Create/modify elements in signature header.
  * @param rpm		path to package
  * @param deleting	adding or deleting signature?
- * @param passPhrase	passPhrase (ignored when deleting)
+ * @param flags
  * @return		0 on success, -1 on error
  */
-static int rpmSign(const char *rpm, int deleting, const char *passPhrase)
+static int rpmSign(const char *rpm, int deleting, int flags)
 {
     FD_t fd = NULL;
     FD_t ofd = NULL;
-    rpmlead lead = NULL;
     char *trpm = NULL;
     Header sigh = NULL;
     Header h = NULL;
-    char * msg = NULL;
+    char *msg = NULL;
     int res = -1; /* assume failure */
     rpmRC rc;
     struct rpmtd_s utd;
     off_t headerStart;
     off_t sigStart;
-    struct sigTarget_s sigt1;
-    struct sigTarget_s sigt2;
+    struct sigTarget_s sigt_v3;
+    struct sigTarget_s sigt_v4;
     unsigned int origSigSize;
     int insSig = 0;
 
@@ -489,25 +497,28 @@ static int rpmSign(const char *rpm, int deleting, const char *passPhrase)
     if (manageFile(&fd, rpm, O_RDWR))
 	goto exit;
 
-    if ((rc = rpmLeadRead(fd, &lead, NULL, &msg)) != RPMRC_OK) {
+    /* Ensure package is intact before attempting to sign */
+    if ((rc = checkPkg(fd, &msg))) {
+	rpmlog(RPMLOG_ERR, "not signing corrupt package %s: %s\n", rpm, msg);
+	goto exit;
+    }
+
+    if ((rc = rpmLeadRead(fd, &msg)) != RPMRC_OK) {
 	rpmlog(RPMLOG_ERR, "%s: %s\n", rpm, msg);
-	free(msg);
 	goto exit;
     }
 
     sigStart = Ftell(fd);
-    rc = rpmReadSignature(fd, &sigh, RPMSIGTYPE_HEADERSIG, &msg);
+    rc = rpmReadSignature(fd, &sigh, &msg);
     if (rc != RPMRC_OK) {
 	rpmlog(RPMLOG_ERR, _("%s: rpmReadSignature failed: %s"), rpm,
 		    (msg && *msg ? msg : "\n"));
-	msg = _free(msg);
 	goto exit;
     }
 
     headerStart = Ftell(fd);
     if (rpmReadHeader(NULL, fd, &h, &msg) != RPMRC_OK) {
 	rpmlog(RPMLOG_ERR, _("%s: headerRead failed: %s\n"), rpm, msg);
-	msg = _free(msg);
 	goto exit;
     }
 
@@ -516,41 +527,35 @@ static int rpmSign(const char *rpm, int deleting, const char *passPhrase)
 	goto exit;
     }
 
-    /* Dump the immutable region (if present). */
-    if (headerGet(sigh, RPMTAG_HEADERSIGNATURES, &utd, HEADERGET_DEFAULT)) {
-	struct rpmtd_s copytd;
-	Header nh = headerNew();
-	Header oh = headerCopyLoad(utd.data);
-	HeaderIterator hi = headerInitIterator(oh);
-	while (headerNext(hi, &copytd)) {
-	    if (copytd.data)
-		headerPut(nh, &copytd, HEADERPUT_DEFAULT);
-	    rpmtdFreeData(&copytd);
-	}
-	headerFreeIterator(hi);
-	headerFree(oh);
-	rpmtdFreeData(&utd);
-
-	headerFree(sigh);
-	sigh = headerLink(nh);
-	headerFree(nh);
+    /* Always add V3 signatures if no payload digest present */
+    if (!(headerIsEntry(h, RPMTAG_PAYLOADDIGEST) ||
+	  headerIsEntry(h, RPMTAG_PAYLOADDIGESTALT))) {
+	flags |= RPMSIGN_FLAG_RPMV3;
     }
+
+    unloadImmutableRegion(&sigh, RPMTAG_HEADERSIGNATURES);
     origSigSize = headerSizeof(sigh, HEADER_MAGIC_YES);
 
+    if (flags & RPMSIGN_FLAG_IMA) {
+	if (includeFileSignatures(&sigh, &h))
+	    goto exit;
+    }
+
     if (deleting) {	/* Nuke all the signature tags. */
 	deleteSigs(sigh);
     } else {
 	/* Signature target containing header + payload */
-	sigt1.fd = fd;
-	sigt1.start = headerStart;
-	sigt1.fileName = rpm;
-	sigt1.size = fdSize(fd) - headerStart;
+	int v3 = (flags & RPMSIGN_FLAG_RPMV3);
+	sigt_v3.fd = fd;
+	sigt_v3.start = headerStart;
+	sigt_v3.fileName = rpm;
+	sigt_v3.size = fdSize(fd) - headerStart;
 
 	/* Signature target containing only header */
-	sigt2 = sigt1;
-	sigt2.size = headerSizeof(h, HEADER_MAGIC_YES);
+	sigt_v4 = sigt_v3;
+	sigt_v4.size = headerSizeof(h, HEADER_MAGIC_YES);
 
-	res = replaceSignature(sigh, &sigt1, &sigt2, passPhrase);
+	res = replaceSignature(sigh, v3 ? &sigt_v3 : NULL, &sigt_v4);
 	if (res != 0) {
 	    if (res == 1) {
 		rpmlog(RPMLOG_WARNING,
@@ -565,25 +570,12 @@ static int rpmSign(const char *rpm, int deleting, const char *passPhrase)
     }
 
     /* Try to make new signature smaller to have size of original signature */
-    rpmtdReset(&utd);
     if (headerGet(sigh, RPMSIGTAG_RESERVEDSPACE, &utd, HEADERGET_MINMEM)) {
-	int diff;
-	int count;
-	char *reservedSpace = NULL;
+	int diff = headerSizeof(sigh, HEADER_MAGIC_YES) - origSigSize;
 
-	count = utd.count;
-	diff = headerSizeof(sigh, HEADER_MAGIC_YES) - origSigSize;
-
-	if (diff < count) {
-	    reservedSpace = xcalloc(count - diff, sizeof(char));
-	    headerDel(sigh, RPMSIGTAG_RESERVEDSPACE);
-	    rpmtdReset(&utd);
-	    utd.tag = RPMSIGTAG_RESERVEDSPACE;
-	    utd.count = count - diff;
-	    utd.type = RPM_BIN_TYPE;
-	    utd.data = reservedSpace;
-	    headerPut(sigh, &utd, HEADERPUT_DEFAULT);
-	    free(reservedSpace);
+	if (diff > 0 && diff < utd.count) {
+	    utd.count -= diff;
+	    headerMod(sigh, &utd);
 	    insSig = 1;
 	}
     }
@@ -617,7 +609,7 @@ static int rpmSign(const char *rpm, int deleting, const char *passPhrase)
 	}
 
 	/* Write the lead/signature of the output rpm */
-	rc = rpmLeadWrite(ofd, lead);
+	rc = rpmLeadWrite(ofd, h);
 	if (rc != RPMRC_OK) {
 	    rpmlog(RPMLOG_ERR, _("%s: writeLead failed: %s\n"), trpm,
 		Fstrerror(ofd));
@@ -654,9 +646,9 @@ exit:
     if (fd)	(void) closeFile(&fd);
     if (ofd)	(void) closeFile(&ofd);
 
-    rpmFreeSignature(sigh);
+    headerFree(sigh);
     headerFree(h);
-    rpmLeadFree(lead);
+    free(msg);
 
     /* Clean up intermediate target */
     if (trpm) {
@@ -667,8 +659,7 @@ exit:
     return res;
 }
 
-int rpmPkgSign(const char *path,
-		const struct rpmSignArgs * args, const char *passPhrase)
+int rpmPkgSign(const char *path, const struct rpmSignArgs * args)
 {
     int rc;
 
@@ -676,29 +667,29 @@ int rpmPkgSign(const char *path,
 	if (args->hashalgo) {
 	    char *algo = NULL;
 	    rasprintf(&algo, "%d", args->hashalgo);
-	    addMacro(NULL, "_gpg_digest_algo", NULL, algo, RMIL_GLOBAL);
+	    rpmPushMacro(NULL, "_gpg_digest_algo", NULL, algo, RMIL_GLOBAL);
 	    free(algo);
 	}
 	if (args->keyid) {
-	    addMacro(NULL, "_gpg_name", NULL, args->keyid, RMIL_GLOBAL);
+	    rpmPushMacro(NULL, "_gpg_name", NULL, args->keyid, RMIL_GLOBAL);
 	}
     }
 
-    rc = rpmSign(path, 0, passPhrase);
+    rc = rpmSign(path, 0, args ? args->signflags : 0);
 
     if (args) {
 	if (args->hashalgo) {
-	    delMacro(NULL, "_gpg_digest_algo");
+	    rpmPopMacro(NULL, "_gpg_digest_algo");
 	}
 	if (args->keyid) {
-	    delMacro(NULL, "_gpg_name");
+	    rpmPopMacro(NULL, "_gpg_name");
 	}
     }
 
     return rc;
 }
 
-int rpmPkgDelSign(const char *path)
+int rpmPkgDelSign(const char *path, const struct rpmSignArgs * args)
 {
-    return rpmSign(path, 1, NULL);
+    return rpmSign(path, 1, 0);
 }
